package cvj2.app;

import java.awt.Color;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;

import org.apache.commons.lang3.SystemUtils;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.widgets.Display;

import cvj2.app.command.CommandMapInput;
import cvj2.app.command.DefaultCommandMapInputCreator;
import cvj2.app.gui.util.MsgBox;
import cvj2.file.fs.FileInfoComparator;
import cvj2.gui.swt.util.FontResourceManager;
import cvj2.gui.swt.widgets.gdidpanel.GridPanel.MouseScrollType;
import cvj2.migemo.Sen;
import cvj2.migemo.Sen.MigemoDictUseType;
import cvj2.util.Log;
import cvj2.util.Util;

/**
 * アプリケーションの設定。<br>
 * このクラスの getter, setter は {@link XMLEncoder}、{@link XMLDecoder} のために自動生成されたもの。<br>
 * public で final でないフィールドは画面で使用する設定だが、XMLに保存されたくないので getter, setter は存在しない。<br>
 */
public class Config implements Cloneable {
  private static Config instance;

  /** アプリケーション名 */
  public static final String APP_NAME = "ComicViewerJ2";
  /** アプリケーションのバージョン */
  public static final String APP_VERSION = "1.0.6";
  /** OS名 ("Windows" or "Linux") */
  static public final String OS_NAME;
  static {
    if (SystemUtils.IS_OS_WINDOWS) {
      OS_NAME = "Windows";
    }
    else if (SystemUtils.IS_OS_LINUX) {
      OS_NAME = "Linux";
    }
    else {
      OS_NAME = "OS";
    }
  }
  
  /* 設定ファイル名 */
  private static final String CONFIG_FILENAME = "ComicViewerJ2.xml";
  /* 設定ファイル (load()で設定される) */
  private static File configFile;
  
  private CommandMapInput cmdMapMain = new DefaultCommandMapInputCreator().createMainWindowCommandMap();
  private CommandMapInput cmdMapView = new DefaultCommandMapInputCreator().createViewWindowCommandMap();
  
  // 画面フォント
  
  private String[] fontDatas = FontResourceManager.getFontDataStrings(Display.getDefault().getSystemFont().getFontData());
  private String[] fontTabDatas = FontResourceManager.getFontDataStrings(Display.getDefault().getSystemFont().getFontData());

  // システム設定
  
  private String version = "";
  
  private boolean showSplash = true;
  private int imageLoadBackgroundThreadNum = Runtime.getRuntime().availableProcessors();
  private int zipFileCacheNum = imageLoadBackgroundThreadNum;
  private String zipFileEncoding = "MS932";
  private Sen.MigemoDictUseType senUseMigemoDict = MigemoDictUseType.ALL;
  private boolean migemoIgnoreHiraKata = true;
  private boolean enableZipFolder = true;
  private boolean enablePdfFolder = true;
  
  private String execProgramNoConfirmExts = ".txt\n" +
      ".jpg\n" +
      ".jpeg\n" +
      ".png\n" +
      ".gif\n" +
      ".bmp\n" +
      ".html\n" +
      ".htm\n" +
      ".zip\n" +
      ".pdf\n" +
      ".doc\n" +
      ".docx\n" +
      ".xls\n" +
      ".xlsx\n" +
      ".ppt\n" +
      ".pptx\n" +
      ".mp3\n" +
      ".mpg\n" +
      ".mpeg\n" +
      ".avi\n" +
      ".wmv\n" +
      "";

  private int defaultLaunchProgram = -1;
  private LaunchProgram[] launchPrograms = createDefaultLaunchProgram();
  
  /** 外部起動プログラムの設定 (XML読み書き用クラス) */
  public static class LaunchProgram implements Serializable {
    private static final long serialVersionUID = 1L;
    
    public static final String REPLACE_SELECTED_FILES = "${selected_files}";
    public static final String REPLACE_CURRENT_DIR = "${current_dir}";
    
    private String name;
    private String cmdarray;
    private int maxFileArgs;
    private boolean confirm;
    
    public String getName() { return name; }
    public void setName(String name) { this.name = name; }
    public String getCmdarray() { return cmdarray; }
    public void setCmdarray(String cmdarray) { this.cmdarray = cmdarray; }
    public boolean isConfirm() { return confirm; }
    public void setConfirm(boolean confirm) { this.confirm = confirm; }
    public int getMaxFileArgs() { return maxFileArgs; }
    public void setMaxFileArgs(int maxFileArgs) { this.maxFileArgs = maxFileArgs; }
  }
  
  private int execProgramMaxNum = 8;

  private boolean msgboxHeadTailCR = true;
  
  private int progressMonitorDialogPlusWidth = 100;
  
  private int thumnailFileCacheCount = 100000;

  private String homeDir = SystemUtils.getUserHome().getAbsolutePath();
  
  // ダイアログ画面設定
  
  private boolean propertyDialogHideReadableCheckBox = SystemUtils.IS_OS_WINDOWS;
  private boolean propertyDialogHideExecuteCheckBox = SystemUtils.IS_OS_WINDOWS;
  
  // デバッグ設定
  
  private boolean isDebugMode = false;
  private Log.Level logDebugLevel = Log.Level.ERROR;
  private boolean logStdout = false;
  private boolean logStdoutputSync = false;
  private boolean logErrWarnToStderr = true;
  private String logFilename = "";
  
  // 画像ロード設定

  private long imageLoadMemCacheBytesThumbnail = 50 * 1024 * 1024L;
  private long imageLoadMemCacheBytesView = 100 * 1024 * 1024L;
  private long imageLoadBackgroundWorkMemory = 256 * 1024 * 1024L;
  private int imageLoadBackgroundThreadShutdownTimeout = 500;
  private long imageLoadMaxFileSizeBytes = 128L * 1024 * 1024L;

  private Color imageLoadBackgroundColor = Color.LIGHT_GRAY;
  private Color imageBackgroundColor = Color.LIGHT_GRAY;
  private boolean isDrawThumbnailBackground = false;
  
  private int imageLoadMaxWidth = 4096;
  private int imageLoadMaxHeight = 4096;
  
  private int imageLoadPdfDpi = 96;
  private double imageLoadPdfFitZoomRatio = 1.0;
  private int imageLoadPdfCacheSize = 1;
  
  private String imageLoaderDefaultPlugin = "ImageIO";
  private String[][] imageLoaderPlugins = {
      { "jpg",  "AWTToolkit" },
      { "jpeg", "AWTToolkit" },
      { "png",  "AWTToolkit" },
      { "gif",  "AWTToolkit" },
  };
  
  private double imageSizeConverterChangeMethodZoomRatio = 0.7; // これ未満なら、Scaled で変換する
  
  private int imageSizeConverterAffineType = AffineTransformOp.TYPE_BICUBIC;
  private int imageSizeConverterScaledHint = Image.SCALE_AREA_AVERAGING;
  private int imageTypeConverterType = BufferedImage.TYPE_INT_RGB;
  
  private int imageLoaderAWTToolkitImageType = BufferedImage.TYPE_INT_RGB;

  private double[][] imageSharpnessConverterParamsByZoomRatios = { // 画像縮小率が[0]未満なら、[1]の値を使用する
      { 0.7, 0.3 },
      { 1.0, 0.1 },
      { 1.01, 0.0 },
      { 100.0, 0.1 },
  };
  private double[][] imageSharpnessConverterParamsByZoomRatiosThumbnail = { // 画像縮小率が[0]未満なら、[1]の値を使用する
      { 1.0, 0.1 },
      { 1.01, 0.0 },
      { 100.0, 0.1 },
  };
  
  // 動作タイミング
  
  private int fileInfoReadInitializeTryTime = 500;
  private double fileInfoReadInitializeTryItemRatio = 0.2;
  
  private int fileInfoReadBackgroundWorkTime = 100;
  private int fileInfoReadBackgroundSleepTime = 10;
  
  private int windowStatusClearTimeMillisec = 1500;
  
  // 画面部品の見栄え
  
  private int windowTabHeightPlus = 6;

  // メイン画面の動作

  private int mainWindowFileWatchInterval = 1000;
  private int mainWindowFileWatch1SecDelayFileCount = 10000;
  
  private boolean mainWindowMigemoTooltipUseSWT = SystemUtils.IS_OS_UNIX || SystemUtils.IS_OS_LINUX;
  private int mainWindowMigemoKeyTimeout = 2000;
  private int mainWindowMigemoTooltipHideDelay = 1000;
  
  private int mainWindowTabMax = 8;
  
  public boolean mainWindowHideFiles = true;
  private String mainWindowHideRegExp = "^(\\..*|System Volume Information|RECYCLER|\\$RECYCLE\\.BIN|\\$Recycle\\.Bin)$";
  
  private boolean mainWindowRenameSelectNewName = true;

  private boolean mainWindowListShowIcon = true;
  private boolean mainWindowThumbnailShowIcon = true;
  private boolean mainWindowUseOsIcon = true;
  
  private int mainWindowFocusPositionHistoryNum = 32;
  
  private boolean mainWindowEnterChildAndShowViewWindowWhenListMode = false;
  private boolean mainWindowEnterChildAndShowViewWindowWhenThumbnailMode = true;
  private boolean mainWindowEnterChildAndShowViewWindowReferLastFocus = true;

  private boolean mainWindowSortFileExtIgnoreCase = true;
  
  private FileInfoComparator.Type mainWindowSortFileEncode = FileInfoComparator.Type.CUSTOM;
  private boolean mainWindowSortCustomIgnoreCase = true;
  private boolean mainWindowSortCustomIgnoreHiraKata = true;
  private boolean mainWindowSortCustomIgnoreZenHan = true;
  private boolean mainWindowSortCustomSymbolFirst = true;
  private boolean mainWindowSortCustomcnvSjis = true;
  
  private boolean mainWindowNewTabCurDirHandOver = true;

  private boolean mainWindowNewTabModeHandOver = true;
  private boolean mainWindowNewTabListMode = false;
  
  private MouseScrollType mainWindowMouseScrollType = MouseScrollType.HALF_PAGE;
  
  // メイン画面のリスト表示モード

  private int mainWindowListWidthName = 100;
  private int mainWindowListWidthExt  = 60;
  private int mainWindowListWidthSize = 130;
  private int mainWindowListWidthDate = 150;

  private int mainWindowListMarginTop = SystemUtils.IS_OS_WINDOWS ? 3 : 1;
  private int mainWindowListMarginBottom = SystemUtils.IS_OS_WINDOWS ? 2 : 1;
  private int mainWindowListMarginLeft = 3;
  
  // メイン画面のサムネイル表示の動作
  
  private int mainWindowThumbnailPrefetchPage = 2;
  private int mainWindowThumbnailSearchDirDepth = 3;

  private int mainWindowThumbnailBackgroundLoadSleep = 100;
  
  // メイン画面のサムネイル表示モード

  private int mainWindowThumbnailSizeX = 128;
  private int mainWindowThumbnailSizeY = 128;
  private int mainWindowThumbnailMarginX = 1;
  private int mainWindowThumbnailMarginY = 1;

  private String mainWindowThumbnailSizes = "64 64\n96 96\n128 128\n192 192\n256 256\n";
  
  // 画像表示画面
  
  private boolean viewWindowEnable = true;

  private int viewWindowForwardLoadNum = Math.max(2, imageLoadBackgroundThreadNum);
  private double viewWindowForwardLoadInCacheRatio = 0.5;
  
  private int viewWindowMoveCursorColumnDelay = 100;
  private int viewWindowMoveCursorRowDelay = 300;
  
  private double viewWindowZoomInRatio = 2.0;
  private double viewWindowZoomOutRatio = 2.0;
  private double viewWindowMoveRatio = 0.2;
  private double viewWindowZoomInRatioMax = 10.0;
  private double viewWindowZoomOutRatioMin = 1.0 / 10.0;
  
  private int viewWindowScrollVisibleTime = 1000;
  
  private int viewWindowFullscreenHideCursorTime = 1000;
  
  private int viewWindowFitImageMaxLoadMagnification = 4;
  
  private boolean viewWindowShowImageInfoDebugMessage = false;
  
  
  // 設定画面では設定しない値

  private int mainWindowTabStatesFocus = -1;
  private TabState[] mainWindowTabStates = {};
  /** 画面のタブ状態 (XML読み書き用クラス) */
  static public class TabState implements Serializable {
    private static final long serialVersionUID = 1L;
    
    private String file;
    private boolean isThumbnailMode;
    
    public String getFile() { return file; }
    public void setFile(String file) { this.file = file; }
    public boolean isThumbnailMode() { return isThumbnailMode; }
    public void setThumbnailMode(boolean isThumbnailMode) { this.isThumbnailMode = isThumbnailMode; }
  }
  
  private HashMap<String, WndState> windowStates = new HashMap<String, WndState>();
  /** ウインドウ状態 (XML読み書き用クラス) */
  static public class WndState implements Serializable {
    private static final long serialVersionUID = 1L;
    
    private Rectangle bounds;
    private boolean maximized;
    
    public Rectangle getBounds() { return bounds; }
    public void setBounds(Rectangle bounds) { this.bounds = bounds; }
    public boolean isMaximized() { return maximized; }
    public void setMaximized(boolean maximized) { this.maximized = maximized; }
  }
  
  // ロードするたびに初期状態に戻す値 (load() でセットする)
  
  /** フルスクリーン表示するか */
  public boolean viewWindowFullscreen = false;
  /** 画像情報を表示するか */
  public boolean viewWindowShowImageInfo = false;


  /*
   * デフォルトの起動プログラム
   */
  private LaunchProgram[] createDefaultLaunchProgram() {
    ArrayList<LaunchProgram> list = new ArrayList<LaunchProgram>();
    
    // Windows の場合
    if (SystemUtils.IS_OS_WINDOWS) {
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "メモ帳で開く";
        l.cmdarray = "notepad.exe\n" + LaunchProgram.REPLACE_SELECTED_FILES + "\n";
        l.confirm = false;
        l.maxFileArgs = 1;
        list.add(l);
      }
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "カレントディレクトリでコマンドプロンプトを開く";
        l.cmdarray = "cmd.exe\n" + "/c\n" + "start\n" + "cmd.exe\n";
        l.confirm = false;
        l.maxFileArgs = 0;
        list.add(l);
      }
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "コマンドプロンプトの中で実行する";
        l.cmdarray = "cmd.exe\n" + "/c\n" + "start\n" + "cmd.exe\n" + "/c\n" + LaunchProgram.REPLACE_SELECTED_FILES + "\n";
        l.confirm = true;
        l.maxFileArgs = 1;
        list.add(l);
      }
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "秀丸で開く (サンプル)";
        l.cmdarray = "c:\\Program Files\\Hidemaru\\Hidemaru.exe\n" + LaunchProgram.REPLACE_SELECTED_FILES + "\n";
        l.confirm = false;
        l.maxFileArgs = -1;
        list.add(l);
      }
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "cyglaunchで開く (サンプル)";
        l.cmdarray = "c:\\Program Files\\teraterm\\cyglaunch.exe\n" + "-nocd\n" + "-nols\n" + "-d\n" + LaunchProgram.REPLACE_CURRENT_DIR + "\n";
        l.confirm = false;
        l.maxFileArgs = 0;
        list.add(l);
      }
    }

    // Linux の場合
    if (SystemUtils.IS_OS_LINUX) {
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "geditで開く";
        l.cmdarray = "gedit\n" + LaunchProgram.REPLACE_SELECTED_FILES + "\n";
        l.confirm = false;
        l.maxFileArgs = -1;
        list.add(l);
      }
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "カレントディレクトリでGnome端末を開く";
        l.cmdarray = "gnome-terminal\n" + "--geometry=100x30\n" + "--hide-menubar\n";
        l.confirm = false;
        l.maxFileArgs = 0;
        list.add(l);
      }
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "Gnome端末の中で実行する";
        l.cmdarray = "gnome-terminal\n" + "-x\n" + LaunchProgram.REPLACE_SELECTED_FILES + "\n";
        l.confirm = true;
        l.maxFileArgs = 1;
        list.add(l);
      }
      {
        LaunchProgram l = new LaunchProgram();
        l.name = "jeditで開く (サンプル)";
        l.cmdarray = "jedit\n" + LaunchProgram.REPLACE_SELECTED_FILES + "\n";
        l.confirm = false;
        l.maxFileArgs = -1;
        list.add(l);
      }
    }
    return list.toArray(new LaunchProgram[0]);
  }
  
  /**
   * clone()
   */
  @SuppressWarnings("unchecked")
  @Override
  public Config clone() {
    try {
      Config obj = (Config) super.clone();
      
      obj.cmdMapMain = (CommandMapInput) this.cmdMapMain.clone();
      obj.cmdMapView = (CommandMapInput) this.cmdMapView.clone();
      obj.launchPrograms = (LaunchProgram[]) Util.deepClone(this.launchPrograms);
      obj.windowStates = (HashMap<String, WndState>) Util.deepClone(this.windowStates);
      obj.fontDatas = (String[]) Util.deepClone(fontDatas);
      obj.fontTabDatas = (String[]) Util.deepClone(fontTabDatas);
      obj.mainWindowTabStates = (TabState[]) Util.deepClone(mainWindowTabStates);
      obj.imageLoaderPlugins = (String[][]) Util.deepClone(imageLoaderPlugins);
      obj.imageSharpnessConverterParamsByZoomRatios = (double[][]) Util.deepClone(imageSharpnessConverterParamsByZoomRatios);
      obj.imageSharpnessConverterParamsByZoomRatiosThumbnail = (double[][]) Util.deepClone(imageSharpnessConverterParamsByZoomRatiosThumbnail);
      
      return obj;
    }
    catch (CloneNotSupportedException e) {
      Log.e(e);
      throw new RuntimeException(e);
    }
  }
  
  /**
   * コンストラクタ
   * @deprecated XML読み書きのみで使用する
   */
  public Config() {  }
  
  /*
   * 内部使用
   * @param dummy 未使用
   */
  private Config(boolean dummy) { }

  /**
   * デフォルト設定のインスタンスを取得する
   */
  public static Config createDefault() {
    Config c = new Config(true);
    c.version = APP_VERSION; // バージョンはここで上書きしないと、XMLファイルに書き込まれない。
    return c;
  }
  
  /**
   * フォントの取得
   */
  public Font getFont() {
    return FontResourceManager.instance().getFont(fontDatas);
  }

  /**
   * フォントの取得
   */
  public Font getFontTab() {
    return FontResourceManager.instance().getFont(fontTabDatas);
  }

  /**
   * インスタンスの取得
   */
  public static Config get() {
    // ロードされていないときはデフォルトをかえす
    if (instance == null) return createDefault();
    return instance;
  }
  
  /**
   * インスタンス内容の上書き
   */
  public static void set(Config config) {
    instance = config;
  }

  /**
   * 設定ファイルロード
   * (エラーが起きたら、メソッドの中でメッセージボックスを表示してから終了する)
   * 
   * @return 正常に設定ファイルを初期化できたら、true 
   *          アプリケーションを起動できない場合は、false
   */
  public static boolean load(File configDir) {
    // 新規作成か
    boolean newCreate = false;
    if (!configDir.exists()) {
      if (!MsgBox.okCancel("設定ディレクトリが存在しません", "新しい設定ディレクトリを作成してもよろしいですか？\n" + configDir)) return false;
      configDir.mkdirs();
      newCreate = true;
    }

    // 存在確認
    if (!configDir.isDirectory()) {
      MsgBox.error("エラー", "アプリケーションの設定ディレクトリを作れませんでした\n" + configDir);
      return false;
    }
    
    configFile = new File(configDir, CONFIG_FILENAME);

    // 設定ファイル存在確認
    if (!configFile.exists()) {
      if (!newCreate) {
        String msg = "設定ファイルを作成してもよろしいですか？\n" + 
                     "  設定ディレクトリ: " + configFile.getParent() + "\n" + 
                     "  設定ファイル: " + configFile.getName();
        if (!MsgBox.okCancel("設定ファイルが存在しません", msg)) return false;
      }
      instance = createDefault();
      return true;
    }
    
    // ロード
    try {
      instance = (Config) Util.readXml(configFile);
      
      // バージョン確認
      if (!APP_VERSION.equals(instance.getVersion())) {
        String msg = "この" + APP_NAME + "のバージョンは " + APP_VERSION + " ですが、\n" +
                     "設定ファイルが作られたのは " + instance.getVersion() + " です。\n" +
                     "新しい設定のデフォルトを使用してもよろしいですか？ (すべて更新)\n" +
                     "(この確認は最初の1回だけ表示されます)\n" +
                     "(もし問題が発生したら、" + APP_NAME + "を終了してから設定ファイルを消してもう一度起動してみてください)";
        int ret = MsgBox.showMsgBoxDialog("アプリケーションが新しいバージョンになりました", SWT.ICON_WARNING, 
                new String[] { "全て更新", "キーボード入力設定のみ置き換える", "現在の設定をそのまま使用する", "起動中止" }, 
                null, 2, msg);
        if (ret == 0) {
          instance = createDefault();
        }
        else if (ret == 1) {
          instance.cmdMapMain = new DefaultCommandMapInputCreator().createMainWindowCommandMap();
          instance.cmdMapView = new DefaultCommandMapInputCreator().createViewWindowCommandMap();
          instance.setVersion(APP_VERSION);
        }
        else if (ret == 2) {
          instance.setVersion(APP_VERSION);
        }
        else {
          return false;
        }
      }

      // 初期状態に戻す
      instance.viewWindowFullscreen = false;
      instance.viewWindowShowImageInfo = false;
      
      return true;
    }
    catch (Exception e) {
      Log.e(e);
      if (!MsgBox.okCancel("設定ファイル読み込みエラー", 
          "ファイル読み込みに失敗しました\n" +
          "アプリケーションがバージョンアップして古い設定ファイルを読めなかったのかもしれません\n" + 
          "デフォルト設定を使用しますか？")) return false;
      instance = createDefault();
      return true;
    }
  }

  /**
   * 設定ファイルセーブ
   */
  public static void save() {
    Log.start();
    try {
      File tmpFile = File.createTempFile(CONFIG_FILENAME, null, configFile.getParentFile());
      Util.writeXml(tmpFile, instance);
      configFile.delete();
      tmpFile.renameTo(configFile);
    }
    catch (IOException e) {
      Log.e(e);
      MsgBox.error("設定ファイル書き込みエラー", "ファイル書き込み失敗\n" + configFile + "\n\n" + e.getMessage());
    }
  }

  /** Eclipse のアウトラインで表示する目印のためのダミー */
  void EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE() {}
  // これ以降は自動生成
  
  public CommandMapInput getCmdMapMain() {
    return cmdMapMain;
  }

  public void setCmdMapMain(CommandMapInput cmdMapMain) {
    this.cmdMapMain = cmdMapMain;
  }

  public CommandMapInput getCmdMapView() {
    return cmdMapView;
  }

  public void setCmdMapView(CommandMapInput cmdMapView) {
    this.cmdMapView = cmdMapView;
  }

  public String[] getFontDatas() {
    return fontDatas;
  }

  public void setFontDatas(String[] fontDatas) {
    this.fontDatas = fontDatas;
  }

  public boolean isDebugMode() {
    return isDebugMode;
  }

  public void setDebugMode(boolean isDebugMode) {
    this.isDebugMode = isDebugMode;
  }

  public int getZipFileCacheNum() {
    return zipFileCacheNum;
  }

  public void setZipFileCacheNum(int zipFileCacheNum) {
    this.zipFileCacheNum = zipFileCacheNum;
  }

  public String getZipFileEncoding() {
    return zipFileEncoding;
  }

  public void setZipFileEncoding(String zipFileEncoding) {
    this.zipFileEncoding = zipFileEncoding;
  }

  public boolean isLogStdout() {
    return logStdout;
  }

  public void setLogStdout(boolean logStdout) {
    this.logStdout = logStdout;
  }

  public String getLogFilename() {
    return logFilename;
  }

  public void setLogFilename(String logFilename) {
    this.logFilename = logFilename;
  }

  public int getImageLoadBackgroundThreadNum() {
    return imageLoadBackgroundThreadNum;
  }

  public void setImageLoadBackgroundThreadNum(int imageLoadBackgroundThreadNum) {
    this.imageLoadBackgroundThreadNum = imageLoadBackgroundThreadNum;
  }

  public long getImageLoadMemCacheBytesThumbnail() {
    return imageLoadMemCacheBytesThumbnail;
  }

  public void setImageLoadMemCacheBytesThumbnail(long imageLoadMemCacheBytesThumbnail) {
    this.imageLoadMemCacheBytesThumbnail = imageLoadMemCacheBytesThumbnail;
  }

  public long getImageLoadMemCacheBytesView() {
    return imageLoadMemCacheBytesView;
  }

  public void setImageLoadMemCacheBytesView(long imageLoadMemCacheBytesView) {
    this.imageLoadMemCacheBytesView = imageLoadMemCacheBytesView;
  }

  public long getImageLoadBackgroundWorkMemory() {
    return imageLoadBackgroundWorkMemory;
  }

  public void setImageLoadBackgroundWorkMemory(long imageLoadBackgroundWorkMemory) {
    this.imageLoadBackgroundWorkMemory = imageLoadBackgroundWorkMemory;
  }

  public int getImageLoadBackgroundThreadShutdownTimeout() {
    return imageLoadBackgroundThreadShutdownTimeout;
  }

  public void setImageLoadBackgroundThreadShutdownTimeout(int imageLoadBackgroundThreadShutdownTimeout) {
    this.imageLoadBackgroundThreadShutdownTimeout = imageLoadBackgroundThreadShutdownTimeout;
  }

  public long getImageLoadMaxFileSizeBytes() {
    return imageLoadMaxFileSizeBytes;
  }

  public void setImageLoadMaxFileSizeBytes(long imageLoadMaxFileSizeBytes) {
    this.imageLoadMaxFileSizeBytes = imageLoadMaxFileSizeBytes;
  }

  public int getImageLoadMaxWidth() {
    return imageLoadMaxWidth;
  }

  public void setImageLoadMaxWidth(int imageLoadMaxWidth) {
    this.imageLoadMaxWidth = imageLoadMaxWidth;
  }

  public int getImageLoadMaxHeight() {
    return imageLoadMaxHeight;
  }

  public void setImageLoadMaxHeight(int imageLoadMaxHeight) {
    this.imageLoadMaxHeight = imageLoadMaxHeight;
  }

  public String getImageLoaderDefaultPlugin() {
    return imageLoaderDefaultPlugin;
  }

  public void setImageLoaderDefaultPlugin(String imageLoaderDefaultPlugin) {
    this.imageLoaderDefaultPlugin = imageLoaderDefaultPlugin;
  }

  public String[][] getImageLoaderPlugins() {
    return imageLoaderPlugins;
  }

  public void setImageLoaderPlugins(String[][] imageLoaderPlugins) {
    this.imageLoaderPlugins = imageLoaderPlugins;
  }

  public double getImageSizeConverterChangeMethodZoomRatio() {
    return imageSizeConverterChangeMethodZoomRatio;
  }

  public void setImageSizeConverterChangeMethodZoomRatio(double imageSizeConverterChangeMethodZoomRatio) {
    this.imageSizeConverterChangeMethodZoomRatio = imageSizeConverterChangeMethodZoomRatio;
  }

  public int getImageSizeConverterAffineType() {
    return imageSizeConverterAffineType;
  }

  public void setImageSizeConverterAffineType(int imageSizeConverterAffineType) {
    this.imageSizeConverterAffineType = imageSizeConverterAffineType;
  }

  public int getImageSizeConverterScaledHint() {
    return imageSizeConverterScaledHint;
  }

  public void setImageSizeConverterScaledHint(int imageSizeConverterScaledHint) {
    this.imageSizeConverterScaledHint = imageSizeConverterScaledHint;
  }

  public int getImageTypeConverterType() {
    return imageTypeConverterType;
  }

  public void setImageTypeConverterType(int imageTypeConverterType) {
    this.imageTypeConverterType = imageTypeConverterType;
  }

  public int getImageLoaderAWTToolkitImageType() {
    return imageLoaderAWTToolkitImageType;
  }

  public void setImageLoaderAWTToolkitImageType(int imageLoaderAWTToolkitImageType) {
    this.imageLoaderAWTToolkitImageType = imageLoaderAWTToolkitImageType;
  }

  public double[][] getImageSharpnessConverterParamsByZoomRatios() {
    return imageSharpnessConverterParamsByZoomRatios;
  }

  public void setImageSharpnessConverterParamsByZoomRatios(double[][] imageSharpnessConverterParamsByZoomRatios) {
    this.imageSharpnessConverterParamsByZoomRatios = imageSharpnessConverterParamsByZoomRatios;
  }
  
  public double[][] getImageSharpnessConverterParamsByZoomRatiosThumbnail() {
    return imageSharpnessConverterParamsByZoomRatiosThumbnail;
  }

  public void setImageSharpnessConverterParamsByZoomRatiosThumbnail(
      double[][] imageSharpnessConverterParamsByZoomRatiosThumbnail) {
    this.imageSharpnessConverterParamsByZoomRatiosThumbnail = imageSharpnessConverterParamsByZoomRatiosThumbnail;
  }

  public int getFileInfoReadInitializeTryTime() {
    return fileInfoReadInitializeTryTime;
  }

  public void setFileInfoReadInitializeTryTime(int fileInfoReadInitializeTryTime) {
    this.fileInfoReadInitializeTryTime = fileInfoReadInitializeTryTime;
  }

  public double getFileInfoReadInitializeTryItemRatio() {
    return fileInfoReadInitializeTryItemRatio;
  }

  public void setFileInfoReadInitializeTryItemRatio(double fileInfoReadInitializeTryItemRatio) {
    this.fileInfoReadInitializeTryItemRatio = fileInfoReadInitializeTryItemRatio;
  }

  public int getFileInfoReadBackgroundWorkTime() {
    return fileInfoReadBackgroundWorkTime;
  }

  public void setFileInfoReadBackgroundWorkTime(int fileInfoReadBackgroundWorkTime) {
    this.fileInfoReadBackgroundWorkTime = fileInfoReadBackgroundWorkTime;
  }

  public int getFileInfoReadBackgroundSleepTime() {
    return fileInfoReadBackgroundSleepTime;
  }

  public void setFileInfoReadBackgroundSleepTime(int fileInfoReadBackgroundSleepTime) {
    this.fileInfoReadBackgroundSleepTime = fileInfoReadBackgroundSleepTime;
  }

  public int getWindowStatusClearTimeMillisec() {
    return windowStatusClearTimeMillisec;
  }

  public void setWindowStatusClearTimeMillisec(int windowStatusClearTimeMillisec) {
    this.windowStatusClearTimeMillisec = windowStatusClearTimeMillisec;
  }

  public int getMainWindowMigemoTooltipHideDelay() {
    return mainWindowMigemoTooltipHideDelay;
  }

  public void setMainWindowMigemoTooltipHideDelay(int migemoTooltipHideDelay) {
    this.mainWindowMigemoTooltipHideDelay = migemoTooltipHideDelay;
  }

  public int getWindowTabHeightPlus() {
    return windowTabHeightPlus;
  }

  public void setWindowTabHeightPlus(int windowTabHeightPlus) {
    this.windowTabHeightPlus = windowTabHeightPlus;
  }

  public int getMainWindowFocusPositionHistoryNum() {
    return mainWindowFocusPositionHistoryNum;
  }

  public void setMainWindowFocusPositionHistoryNum(int mainWindowFocusPositionHistoryNum) {
    this.mainWindowFocusPositionHistoryNum = mainWindowFocusPositionHistoryNum;
  }

  public boolean isMainWindowEnterChildAndShowViewWindowWhenListMode() {
    return mainWindowEnterChildAndShowViewWindowWhenListMode;
  }

  public void setMainWindowEnterChildAndShowViewWindowWhenListMode(
      boolean mainWindowEnterChildAndShowViewWindowWhenListMode) {
    this.mainWindowEnterChildAndShowViewWindowWhenListMode = mainWindowEnterChildAndShowViewWindowWhenListMode;
  }

  public boolean isMainWindowEnterChildAndShowViewWindowWhenThumbnailMode() {
    return mainWindowEnterChildAndShowViewWindowWhenThumbnailMode;
  }

  public void setMainWindowEnterChildAndShowViewWindowWhenThumbnailMode(
      boolean mainWindowEnterChildAndShowViewWindowWhenThumbnailMode) {
    this.mainWindowEnterChildAndShowViewWindowWhenThumbnailMode = mainWindowEnterChildAndShowViewWindowWhenThumbnailMode;
  }

  public boolean isMainWindowEnterChildAndShowViewWindowReferLastFocus() {
    return mainWindowEnterChildAndShowViewWindowReferLastFocus;
  }

  public void setMainWindowEnterChildAndShowViewWindowReferLastFocus(
      boolean mainWindowEnterChildAndShowViewWindowReferLastFocus) {
    this.mainWindowEnterChildAndShowViewWindowReferLastFocus = mainWindowEnterChildAndShowViewWindowReferLastFocus;
  }

  public boolean isMainWindowSortFileExtIgnoreCase() {
    return mainWindowSortFileExtIgnoreCase;
  }

  public void setMainWindowSortFileExtIgnoreCase(boolean mainWindowSortFileExtIgnoreCase) {
    this.mainWindowSortFileExtIgnoreCase = mainWindowSortFileExtIgnoreCase;
  }

  public boolean isMainWindowListShowIcon() {
    return mainWindowListShowIcon;
  }

  public void setMainWindowListShowIcon(boolean mainWindowListShowIcon) {
    this.mainWindowListShowIcon = mainWindowListShowIcon;
  }

  public int getMainWindowListWidthName() {
    return mainWindowListWidthName;
  }

  public void setMainWindowListWidthName(int mainWindowListWidthName) {
    this.mainWindowListWidthName = mainWindowListWidthName;
  }

  public int getMainWindowListWidthExt() {
    return mainWindowListWidthExt;
  }

  public void setMainWindowListWidthExt(int mainWindowListWidthExt) {
    this.mainWindowListWidthExt = mainWindowListWidthExt;
  }

  public int getMainWindowListWidthSize() {
    return mainWindowListWidthSize;
  }

  public void setMainWindowListWidthSize(int mainWindowListWidthSize) {
    this.mainWindowListWidthSize = mainWindowListWidthSize;
  }

  public int getMainWindowListWidthDate() {
    return mainWindowListWidthDate;
  }

  public void setMainWindowListWidthDate(int mainWindowListWidthDate) {
    this.mainWindowListWidthDate = mainWindowListWidthDate;
  }

  public int getMainWindowListMarginLeft() {
    return mainWindowListMarginLeft;
  }

  public void setMainWindowListMarginLeft(int mainWindowListMarginLeft) {
    this.mainWindowListMarginLeft = mainWindowListMarginLeft;
  }

  public int getMainWindowThumbnailPrefetchPage() {
    return mainWindowThumbnailPrefetchPage;
  }

  public void setMainWindowThumbnailPrefetchPage(int mainWindowThumbnailPrefetchPage) {
    this.mainWindowThumbnailPrefetchPage = mainWindowThumbnailPrefetchPage;
  }

  public int getMainWindowThumbnailSearchDirDepth() {
    return mainWindowThumbnailSearchDirDepth;
  }

  public void setMainWindowThumbnailSearchDirDepth(int mainWindowThumbnailSearchDirDepth) {
    this.mainWindowThumbnailSearchDirDepth = mainWindowThumbnailSearchDirDepth;
  }

  public boolean isMainWindowThumbnailShowIcon() {
    return mainWindowThumbnailShowIcon;
  }

  public void setMainWindowThumbnailShowIcon(boolean mainWindowThumbnailShowIcon) {
    this.mainWindowThumbnailShowIcon = mainWindowThumbnailShowIcon;
  }

  public int getMainWindowThumbnailSizeX() {
    return mainWindowThumbnailSizeX;
  }

  public void setMainWindowThumbnailSizeX(int mainWindowThumbnailSizeX) {
    this.mainWindowThumbnailSizeX = mainWindowThumbnailSizeX;
  }

  public int getMainWindowThumbnailSizeY() {
    return mainWindowThumbnailSizeY;
  }

  public void setMainWindowThumbnailSizeY(int mainWindowThumbnailSizeY) {
    this.mainWindowThumbnailSizeY = mainWindowThumbnailSizeY;
  }

  public int getMainWindowThumbnailMarginX() {
    return mainWindowThumbnailMarginX;
  }

  public void setMainWindowThumbnailMarginX(int mainWindowThumbnailMarginX) {
    this.mainWindowThumbnailMarginX = mainWindowThumbnailMarginX;
  }

  public int getMainWindowThumbnailMarginY() {
    return mainWindowThumbnailMarginY;
  }

  public void setMainWindowThumbnailMarginY(int mainWindowThumbnailMarginY) {
    this.mainWindowThumbnailMarginY = mainWindowThumbnailMarginY;
  }

  public int getViewWindowForwardLoadNum() {
    return viewWindowForwardLoadNum;
  }

  public void setViewWindowForwardLoadNum(int viewWindowForwardLoadNum) {
    this.viewWindowForwardLoadNum = viewWindowForwardLoadNum;
  }

  public double getViewWindowForwardLoadInCacheRatio() {
    return viewWindowForwardLoadInCacheRatio;
  }

  public void setViewWindowForwardLoadInCacheRatio(double viewWindowForwardLoadInCacheRatio) {
    this.viewWindowForwardLoadInCacheRatio = viewWindowForwardLoadInCacheRatio;
  }

  public double getViewWindowZoomInRatio() {
    return viewWindowZoomInRatio;
  }

  public void setViewWindowZoomInRatio(double viewWindowZoomInRatio) {
    this.viewWindowZoomInRatio = viewWindowZoomInRatio;
  }

  public double getViewWindowZoomOutRatio() {
    return viewWindowZoomOutRatio;
  }

  public void setViewWindowZoomOutRatio(double viewWindowZoomOutRatio) {
    this.viewWindowZoomOutRatio = viewWindowZoomOutRatio;
  }

  public double getViewWindowMoveRatio() {
    return viewWindowMoveRatio;
  }

  public void setViewWindowMoveRatio(double viewWindowMoveRatio) {
    this.viewWindowMoveRatio = viewWindowMoveRatio;
  }

  public double getViewWindowZoomInRatioMax() {
    return viewWindowZoomInRatioMax;
  }

  public void setViewWindowZoomInRatioMax(double viewWindowZoomInRatioMax) {
    this.viewWindowZoomInRatioMax = viewWindowZoomInRatioMax;
  }

  public double getViewWindowZoomOutRatioMin() {
    return viewWindowZoomOutRatioMin;
  }

  public void setViewWindowZoomOutRatioMin(double viewWindowZoomOutRatioMin) {
    this.viewWindowZoomOutRatioMin = viewWindowZoomOutRatioMin;
  }

  public int getViewWindowScrollVisibleTime() {
    return viewWindowScrollVisibleTime;
  }

  public void setViewWindowScrollVisibleTime(int viewWindowScrollVisibleTime) {
    this.viewWindowScrollVisibleTime = viewWindowScrollVisibleTime;
  }

  public int getViewWindowFullscreenHideCursorTime() {
    return viewWindowFullscreenHideCursorTime;
  }

  public void setViewWindowFullscreenHideCursorTime(int viewWindowFullscreenHideCursorTime) {
    this.viewWindowFullscreenHideCursorTime = viewWindowFullscreenHideCursorTime;
  }

  public int getViewWindowFitImageMaxLoadMagnification() {
    return viewWindowFitImageMaxLoadMagnification;
  }

  public void setViewWindowFitImageMaxLoadMagnification(int viewWindowFitImageMaxLoadMagnification) {
    this.viewWindowFitImageMaxLoadMagnification = viewWindowFitImageMaxLoadMagnification;
  }

  public boolean isViewWindowShowImageInfoDebugMessage() {
    return viewWindowShowImageInfoDebugMessage;
  }

  public void setViewWindowShowImageInfoDebugMessage(boolean viewWindowShowImageInfoDebugMessage) {
    this.viewWindowShowImageInfoDebugMessage = viewWindowShowImageInfoDebugMessage;
  }

  public int getMainWindowThumbnailBackgroundLoadSleep() {
    return mainWindowThumbnailBackgroundLoadSleep;
  }

  public void setMainWindowThumbnailBackgroundLoadSleep(int mainWindowThumbnailBackgroundLoadSleep) {
    this.mainWindowThumbnailBackgroundLoadSleep = mainWindowThumbnailBackgroundLoadSleep;
  }

  public int getMainWindowTabMax() {
    return mainWindowTabMax;
  }

  public void setMainWindowTabMax(int mainWindowTabMax) {
    this.mainWindowTabMax = mainWindowTabMax;
  }
  
  public boolean isLogStdoutputSync() {
    return logStdoutputSync;
  }

  public void setLogStdoutputSync(boolean logStdoutputSync) {
    this.logStdoutputSync = logStdoutputSync;
  }

  public boolean isLogErrWarnToStderr() {
    return logErrWarnToStderr;
  }

  public void setLogErrWarnToStderr(boolean logErrWarnToStderr) {
    this.logErrWarnToStderr = logErrWarnToStderr;
  }

  public boolean isMainWindowNewTabCurDirHandOver() {
    return mainWindowNewTabCurDirHandOver;
  }

  public void setMainWindowNewTabCurDirHandOver(boolean mainWindowNewTabCurDirHandOver) {
    this.mainWindowNewTabCurDirHandOver = mainWindowNewTabCurDirHandOver;
  }

  public boolean isMainWindowNewTabModeHandOver() {
    return mainWindowNewTabModeHandOver;
  }

  public void setMainWindowNewTabModeHandOver(boolean mainWindowNewTabModeHandOver) {
    this.mainWindowNewTabModeHandOver = mainWindowNewTabModeHandOver;
  }

  public boolean isMainWindowNewTabListMode() {
    return mainWindowNewTabListMode;
  }

  public void setMainWindowNewTabListMode(boolean mainWindowNewTabListMode) {
    this.mainWindowNewTabListMode = mainWindowNewTabListMode;
  }

  public TabState[] getMainWindowTabStates() {
    return mainWindowTabStates;
  }

  public void setMainWindowTabStates(TabState[] mainWindowTabStates) {
    this.mainWindowTabStates = mainWindowTabStates;
  }

  public boolean isMainWindowMigemoTooltipUseSWT() {
    return mainWindowMigemoTooltipUseSWT;
  }

  public void setMainWindowMigemoTooltipUseSWT(boolean mainWindowMigemoTooltipUseSWT) {
    this.mainWindowMigemoTooltipUseSWT = mainWindowMigemoTooltipUseSWT;
  }

  public Sen.MigemoDictUseType getSenUseMigemoDict() {
    return senUseMigemoDict;
  }

  public void setSenUseMigemoDict(Sen.MigemoDictUseType senUseMigemoDict) {
    this.senUseMigemoDict = senUseMigemoDict;
  }

  public int getMainWindowTabStatesFocus() {
    return mainWindowTabStatesFocus;
  }

  public void setMainWindowTabStatesFocus(int mainWindowTabStatesFocus) {
    this.mainWindowTabStatesFocus = mainWindowTabStatesFocus;
  }

  public int getExecProgramMaxNum() {
    return execProgramMaxNum;
  }

  public void setExecProgramMaxNum(int execProgramMaxNum) {
    this.execProgramMaxNum = execProgramMaxNum;
  }

  public boolean isMsgboxHeadTailCR() {
    return msgboxHeadTailCR;
  }

  public void setMsgboxHeadTailCR(boolean msgboxHeadTailCR) {
    this.msgboxHeadTailCR = msgboxHeadTailCR;
  }

  public int getProgressMonitorDialogPlusWidth() {
    return progressMonitorDialogPlusWidth;
  }

  public void setProgressMonitorDialogPlusWidth(int progressMonitorDialogPlusWidth) {
    this.progressMonitorDialogPlusWidth = progressMonitorDialogPlusWidth;
  }

  public boolean isPropertyDialogHideReadableCheckBox() {
    return propertyDialogHideReadableCheckBox;
  }

  public void setPropertyDialogHideReadableCheckBox(boolean propertyDialogHideReadableCheckBox) {
    this.propertyDialogHideReadableCheckBox = propertyDialogHideReadableCheckBox;
  }

  public boolean isPropertyDialogHideExecuteCheckBox() {
    return propertyDialogHideExecuteCheckBox;
  }

  public void setPropertyDialogHideExecuteCheckBox(boolean propertyDialogHideExecuteCheckBox) {
    this.propertyDialogHideExecuteCheckBox = propertyDialogHideExecuteCheckBox;
  }

  public String getVersion() {
    return version;
  }

  public void setVersion(String version) {
    this.version = version;
  }

  public int getThumnailFileCacheCount() {
    return thumnailFileCacheCount;
  }

  public void setThumnailFileCacheCount(int thumnailFileCacheCount) {
    this.thumnailFileCacheCount = thumnailFileCacheCount;
  }

  public HashMap<String, WndState> getWindowStates() {
    return windowStates;
  }

  public void setWindowStates(HashMap<String, WndState> windowStates) {
    this.windowStates = windowStates;
  }

  public FileInfoComparator.Type getMainWindowSortFileEncode() {
    return mainWindowSortFileEncode;
  }

  public void setMainWindowSortFileEncode(FileInfoComparator.Type mainWindowSortFileEncode) {
    this.mainWindowSortFileEncode = mainWindowSortFileEncode;
  }

  public Log.Level getLogDebugLevel() {
    return logDebugLevel;
  }

  public void setLogDebugLevel(Log.Level logDebugLevel) {
    this.logDebugLevel = logDebugLevel;
  }

  public int getImageLoadPdfDpi() {
    return imageLoadPdfDpi;
  }

  public void setImageLoadPdfDpi(int imageLoadPdfDpi) {
    this.imageLoadPdfDpi = imageLoadPdfDpi;
  }

  public double getImageLoadPdfFitZoomRatio() {
    return imageLoadPdfFitZoomRatio;
  }

  public void setImageLoadPdfFitZoomRatio(double imageLoadPdfFitZoomRatio) {
    this.imageLoadPdfFitZoomRatio = imageLoadPdfFitZoomRatio;
  }

  public int getMainWindowFileWatchInterval() {
    return mainWindowFileWatchInterval;
  }

  public void setMainWindowFileWatchInterval(int mainWindowFileWatchInterval) {
    this.mainWindowFileWatchInterval = mainWindowFileWatchInterval;
  }

  public int getMainWindowFileWatch1SecDelayFileCount() {
    return mainWindowFileWatch1SecDelayFileCount;
  }

  public void setMainWindowFileWatch1SecDelayFileCount(int mainWindowFileWatch1SecDelayFileCount) {
    this.mainWindowFileWatch1SecDelayFileCount = mainWindowFileWatch1SecDelayFileCount;
  }

  public Color getImageLoadBackgroundColor() {
    return imageLoadBackgroundColor;
  }

  public void setImageLoadBackgroundColor(Color imageLoadBackgroundColor) {
    this.imageLoadBackgroundColor = imageLoadBackgroundColor;
  }

  public boolean isMainWindowSortCustomIgnoreCase() {
    return mainWindowSortCustomIgnoreCase;
  }

  public void setMainWindowSortCustomIgnoreCase(boolean mainWindowSortCustomIgnoreCase) {
    this.mainWindowSortCustomIgnoreCase = mainWindowSortCustomIgnoreCase;
  }

  public boolean isMainWindowSortCustomIgnoreHiraKata() {
    return mainWindowSortCustomIgnoreHiraKata;
  }

  public void setMainWindowSortCustomIgnoreHiraKata(boolean mainWindowSortCustomIgnoreHiraKata) {
    this.mainWindowSortCustomIgnoreHiraKata = mainWindowSortCustomIgnoreHiraKata;
  }

  public boolean isMainWindowSortCustomIgnoreZenHan() {
    return mainWindowSortCustomIgnoreZenHan;
  }

  public void setMainWindowSortCustomIgnoreZenHan(boolean mainWindowSortCustomIgnoreZenHan) {
    this.mainWindowSortCustomIgnoreZenHan = mainWindowSortCustomIgnoreZenHan;
  }

  public boolean isMainWindowSortCustomSymbolFirst() {
    return mainWindowSortCustomSymbolFirst;
  }

  public void setMainWindowSortCustomSymbolFirst(boolean mainWindowSortCustomSymbolFirst) {
    this.mainWindowSortCustomSymbolFirst = mainWindowSortCustomSymbolFirst;
  }

  public boolean isMainWindowSortCustomcnvSjis() {
    return mainWindowSortCustomcnvSjis;
  }

  public void setMainWindowSortCustomcnvSjis(boolean mainWindowSortCustomcnvSjis) {
    this.mainWindowSortCustomcnvSjis = mainWindowSortCustomcnvSjis;
  }

  public String getHomeDir() {
    return homeDir;
  }

  public void setHomeDir(String homeDir) {
    this.homeDir = homeDir;
  }

  public int getViewWindowMoveCursorColumnDelay() {
    return viewWindowMoveCursorColumnDelay;
  }

  public void setViewWindowMoveCursorColumnDelay(int viewWindowMoveCursorColumnDelay) {
    this.viewWindowMoveCursorColumnDelay = viewWindowMoveCursorColumnDelay;
  }

  public int getViewWindowMoveCursorRowDelay() {
    return viewWindowMoveCursorRowDelay;
  }

  public void setViewWindowMoveCursorRowDelay(int viewWindowMoveCursorRowDelay) {
    this.viewWindowMoveCursorRowDelay = viewWindowMoveCursorRowDelay;
  }

  public String[] getFontTabDatas() {
    return fontTabDatas;
  }

  public void setFontTabDatas(String[] fontTabDatas) {
    this.fontTabDatas = fontTabDatas;
  }

  public int getImageLoadPdfCacheSize() {
    return imageLoadPdfCacheSize;
  }

  public void setImageLoadPdfCacheSize(int imageLoadPdfCacheSize) {
    this.imageLoadPdfCacheSize = imageLoadPdfCacheSize;
  }

  public int getMainWindowListMarginTop() {
    return mainWindowListMarginTop;
  }

  public void setMainWindowListMarginTop(int mainWindowListMarginTop) {
    this.mainWindowListMarginTop = mainWindowListMarginTop;
  }

  public int getMainWindowListMarginBottom() {
    return mainWindowListMarginBottom;
  }

  public void setMainWindowListMarginBottom(int mainWindowListMarginBottom) {
    this.mainWindowListMarginBottom = mainWindowListMarginBottom;
  }

  public boolean isShowSplash() {
    return showSplash;
  }

  public void setShowSplash(boolean showSplash) {
    this.showSplash = showSplash;
  }

  public Color getImageBackgroundColor() {
    return imageBackgroundColor;
  }

  public void setImageBackgroundColor(Color imageBackgroundColor) {
    this.imageBackgroundColor = imageBackgroundColor;
  }

  public boolean isDrawThumbnailBackground() {
    return isDrawThumbnailBackground;
  }

  public void setDrawThumbnailBackground(boolean isDrawThumbnailBackground) {
    this.isDrawThumbnailBackground = isDrawThumbnailBackground;
  }

  public int getMainWindowMigemoKeyTimeout() {
    return mainWindowMigemoKeyTimeout;
  }

  public void setMainWindowMigemoKeyTimeout(int mainWindowMigemoKeyTimeout) {
    this.mainWindowMigemoKeyTimeout = mainWindowMigemoKeyTimeout;
  }

  public boolean isEnablePdfFolder() {
    return enablePdfFolder;
  }

  public void setEnablePdfFolder(boolean enablePdfFolder) {
    this.enablePdfFolder = enablePdfFolder;
  }

  public boolean isEnableZipFolder() {
    return enableZipFolder;
  }

  public void setEnableZipFolder(boolean enableZipFolder) {
    this.enableZipFolder = enableZipFolder;
  }

  public boolean isMainWindowRenameSelectNewName() {
    return mainWindowRenameSelectNewName;
  }

  public void setMainWindowRenameSelectNewName(boolean mainWindowRenameSelectNewName) {
    this.mainWindowRenameSelectNewName = mainWindowRenameSelectNewName;
  }

  public boolean isMainWindowUseOsIcon() {
    return mainWindowUseOsIcon;
  }

  public void setMainWindowUseOsIcon(boolean mainWindowUseOsIcon) {
    this.mainWindowUseOsIcon = mainWindowUseOsIcon;
  }

  public String getExecProgramNoConfirmExts() {
    return execProgramNoConfirmExts;
  }

  public void setExecProgramNoConfirmExts(String execProgramNoConfirmExts) {
    this.execProgramNoConfirmExts = execProgramNoConfirmExts;
  }

  public String getMainWindowHideRegExp() {
    return mainWindowHideRegExp;
  }

  public void setMainWindowHideRegExp(String mainWindowHideRegExp) {
    this.mainWindowHideRegExp = mainWindowHideRegExp;
  }

  public int getDefaultLaunchProgram() {
    return defaultLaunchProgram;
  }

  public void setDefaultLaunchProgram(int defaultLaunchProgram) {
    this.defaultLaunchProgram = defaultLaunchProgram;
  }

  public LaunchProgram[] getLaunchPrograms() {
    return launchPrograms;
  }

  public void setLaunchPrograms(LaunchProgram[] launchPrograms) {
    this.launchPrograms = launchPrograms;
  }

  public String getMainWindowThumbnailSizes() {
    return mainWindowThumbnailSizes;
  }

  public void setMainWindowThumbnailSizes(String mainWindowThumbnailSizes) {
    this.mainWindowThumbnailSizes = mainWindowThumbnailSizes;
  }

  public boolean isViewWindowEnable() {
    return viewWindowEnable;
  }

  public void setViewWindowEnable(boolean viewWindowEnable) {
    this.viewWindowEnable = viewWindowEnable;
  }

  public MouseScrollType getMainWindowMouseScrollType() {
    return mainWindowMouseScrollType;
  }

  public void setMainWindowMouseScrollType(MouseScrollType mainWindowMouseScrollType) {
    this.mainWindowMouseScrollType = mainWindowMouseScrollType;
  }

  public boolean isMainWindowHideFiles() {
    return mainWindowHideFiles;
  }

  public void setMainWindowHideFiles(boolean mainWindowHideFiles) {
    this.mainWindowHideFiles = mainWindowHideFiles;
  }

  public boolean isMigemoIgnoreHiraKata() {
    return migemoIgnoreHiraKata;
  }

  public void setMigemoIgnoreHiraKata(boolean migemoIgnoreHiraKata) {
    this.migemoIgnoreHiraKata = migemoIgnoreHiraKata;
  }
}
