package cvj2.app;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.lang3.time.StopWatch;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.wb.swt.ResourceManager;
import org.eclipse.wb.swt.SWTResourceManager;

import cvj2.app.file.ApplicationLock;
import cvj2.app.file.imgload.BackgroundImageLoader;
import cvj2.app.file.imgload.CachedImage;
import cvj2.app.file.thumbnail.ThumbnailCache;
import cvj2.app.file.thumbnail.fs.FileCacheManager;
import cvj2.app.gui.data.SystemMonitor;
import cvj2.app.gui.util.GlobalErrorHandler;
import cvj2.app.gui.util.MsgBox;
import cvj2.app.gui.widgets.ListPanelToolTip;
import cvj2.app.gui.window.WindowSelector;
import cvj2.app.gui.window.main.MainWindow;
import cvj2.app.gui.window.splash.SplashWindow;
import cvj2.file.fs.FileInfo;
import cvj2.file.fs.FileInfoComparator;
import cvj2.file.fs.FileInfoFactory;
import cvj2.file.fs.FileInfoList;
import cvj2.file.fs.FileInfoPDFManager;
import cvj2.file.fs.FileInfoZipManager;
import cvj2.gui.swt.migemo.SWTMigemoKeyHandler;
import cvj2.gui.swt.util.FontResourceManager;
import cvj2.gui.swt.util.NullCursor;
import cvj2.img.ImageLocation;
import cvj2.img.ZoomLevel;
import cvj2.img.converter.ImageTypeConverter;
import cvj2.img.loader.ImageLoader;
import cvj2.img.loader.ImageLoader2;
import cvj2.img.loader.ImageLoaderPluginAWTToolkit;
import cvj2.migemo.Migemo;
import cvj2.migemo.MigemoKeyHandler;
import cvj2.migemo.MigemoReverseDict;
import cvj2.migemo.Sen;
import cvj2.util.Log;

/**
 * Main
 */
public class Main {
  // Migemo キー入力はアプリケーション全体で共有する
  private static SWTMigemoKeyHandler MIGEMO_KEY = new SWTMigemoKeyHandler();
  // クリップボードはアプリケーション全体で共有する
  volatile private static Clipboard CLIPBOARD;
  // シャットダウン中
  private static boolean isShutdown = false;
  // スプラッシュウインドウ
  private static SplashWindow splash;

  private Main() {}
  
  /**
   * main
   */
  public static void main(String[] args) {
    try {
      // コマンドライン引数解析
      if (!ParseArg.parse(args)) return;
      
      // 設定ファイル読み込み
      if (!Config.load(ParseArg.configDir)) return;
      
      splash = new SplashWindow();
      if (Config.get().isShowSplash()) splash.showSplashWindow();
      
      // ログ出力開始
      setupLogByConfig();
      if (Config.get().getLogFilename().length() != 0) {
        Log.fileOpen(new File(ParseArg.configDir, Config.get().getLogFilename()));
      }
      Log.start((Object[])args);
      printSystemEnv();

      setSplashMessage("アプリケーション初期化中");

      // ロックファイル初期化
      ApplicationLock.instance().initialize();
      
      // 設定ファイルに従ってアプリケーションの初期化
      setupByConfig();

      // Migemo,Sen 初期化
      setSplashMessage("Migemoを初期化しています");
      Migemo.initialize();
      MigemoReverseDict.instance().initialize();

      // サムネイルキャッシュファイル初期化
      setSplashMessage("サムネイルキャッシュを初期化しています");
      ThumbnailCache.instance().initialize(ApplicationLock.instance().isFirstApp());
      
      // システムモニタ初期化
      SystemMonitor.instance().initialize();
      
      // スレッドの例外ハンドラ登録
      Thread.setDefaultUncaughtExceptionHandler(new ThreadErrorHandler());

      // 画面を開く
      setSplashMessage("画面を開いています");
      Log.d("mainWindow open");
      WindowSelector.instance().initialOpen();
      Log.d("mainWindow closed");
      
      // 設定ファイルセーブ
      if (ApplicationLock.instance().isFirstApp()) {
        Config.save();
      }

      // サムネイルキャッシュのシャットダウン
      ThumbnailCache.instance().shutdown();
      
      if (!isShutdown) throw new RuntimeException("シャットダウン不良");
    }
    catch (Throwable e) {
      Log.e(e);
      MsgBox.error("アプリケーションエラー", "重大なエラーが発生しました\n\n" + e.getMessage());
    }
    finally {
      if (CLIPBOARD != null && !CLIPBOARD.isDisposed()) CLIPBOARD.dispose();
      SWTResourceManager.dispose();
      ResourceManager.dispose();
      FontResourceManager.instance().dispose();
      NullCursor.dispose();
      FileInfoPDFManager.pdfCache.setLimit(0);
      Log.close();
    }
  }

  /**
   * システム環境をログに出力
   */
  private static void printSystemEnv() {
    // 環境変数
    Map<String,String> env = System.getenv();
    Log.d("env.size = ", env.size());
    String[] envKeys = env.keySet().toArray(new String[0]);
    Arrays.sort(envKeys);
    for (String key : envKeys) {
      Log.d("\t", key, "=", env.get(key));
    }
    
    // プロパティ
    Properties prop = System.getProperties();
    Log.d("System properties.size = ", prop.size());
    ArrayList<String> propKeys = new ArrayList<String>();
    for (Object key : prop.keySet()) {
      propKeys.add("" + key);
    }
    Collections.sort(propKeys);
    for (String key : propKeys) {
      Log.d("\t", key, "=", prop.get(key));
    }
  }

  /**
   * ログの再設定
   */
  public static void setupLogByConfig() {
    Config c = Config.get();
    Log.printLevel = c.getLogDebugLevel();
    Log.stdoutput = c.isLogStdout();
    Log.stdoutputSync = c.isLogStdoutputSync();
    Log.errWarnToStderr = c.isLogErrWarnToStderr();
  }

  /**
   * 設定内容に従って、アプリケーションを再設定する
   */
  public static void setupByConfig() {
    Log.start();
    Config c = Config.get();

    // ログ設定
    setupLogByConfig();
    
    // 画面フォント

    MainWindow mainWindow = WindowSelector.instance().getMainWindow();
    if (mainWindow != null) {
      mainWindow.setFont(c.getFont());
      mainWindow.getTabFolder().setFont(c.getFontTab());
    }
    
    // システム設定
    
    GlobalErrorHandler.isDebug = c.isDebugMode();
    BackgroundImageLoader.instance().setup(
        c.getImageLoadBackgroundThreadNum(), 
        c.getImageLoadBackgroundWorkMemory(), 
        c.getImageLoadMemCacheBytesThumbnail(), 
        c.getImageLoadMemCacheBytesView());
    FileInfoFactory.enableZipFolder = c.isEnableZipFolder();
    FileInfoFactory.enablePdfFolder = c.isEnablePdfFolder();
    FileInfoZipManager.encoding = c.getZipFileEncoding();
    FileInfoZipManager.cache.setLimit(c.getZipFileCacheNum());
    FileInfoPDFManager.infoCache.setLimit(c.getImageLoadPdfCacheSize() * 2);
    FileInfoPDFManager.pdfCache.setLimit(c.getImageLoadPdfCacheSize());
    Sen.useMigemoDictType = c.getSenUseMigemoDict();
    FileCacheManager.instance().setCacheLimit(c.getThumnailFileCacheCount());
    FileInfoList.hideNameRegExp = c.getMainWindowHideRegExp();
    FileInfoList.hideFiles = c.isMainWindowHideFiles();
    
    // 画像ロード設定
    
    assert ImageLoader.getAllPlugins().containsKey(c.getImageLoaderDefaultPlugin());
    ImageLoader.defaultPlugin = c.getImageLoaderDefaultPlugin();
    ImageLoader.setExtPluginNameMap(c.getImageLoaderPlugins());
    
    ImageTypeConverter.backgroundColor = c.getImageLoadBackgroundColor();
    ImageLoaderPluginAWTToolkit.outputBufferedImageType = c.getImageLoaderAWTToolkitImageType();
    
    ImageLoader2.backgroundColor = c.getImageLoadBackgroundColor();
    ImageLoader2.type = c.getImageTypeConverterType();

    // 内部動作

    FileInfoComparator.customCmp.init(
        c.isMainWindowSortCustomIgnoreCase(),
        c.isMainWindowSortCustomIgnoreHiraKata(),
        c.isMainWindowSortCustomIgnoreZenHan(),
        c.isMainWindowSortCustomSymbolFirst(),
        null,
        c.isMainWindowSortCustomcnvSjis());
    FileInfo.cmp = new FileInfoComparator(c.getMainWindowSortFileEncode());
    
    // 動作タイミング
    
    ListPanelToolTip.autoHideDelay = c.getMainWindowMigemoTooltipHideDelay();
    MigemoKeyHandler.setInputTimeoutMilli(c.getMainWindowMigemoKeyTimeout());

    // 画像表示画面
    
    ZoomLevel.zoomInRatio = c.getViewWindowZoomInRatio();
    ZoomLevel.zoomOutRatio = c.getViewWindowZoomOutRatio();
    ImageLocation.moveRatio = c.getViewWindowMoveRatio();
  }
  
  /**
   * アプリケーションアイコン取得
   */
  public static Image getAppIcon() {
    return SWTResourceManager.getImage(MainWindow.class, "ComicViewerJ2.png");
  }
  
  /**
   * クリップボードインスタンス取得
   */
  public static Clipboard getClipboard() {
    return CLIPBOARD;
  }

  /**
   * Migemo入力インスタンス取得
   */
  public static SWTMigemoKeyHandler getMigemoKey() {
    return MIGEMO_KEY;
  }
  
  /**
   * 画面が出てから初期化する必要のある処理
   */
  public static void initializeByMainWindow() {
    Log.start();
    if (CLIPBOARD == null) {
      CLIPBOARD = new Clipboard(Display.getDefault());
    }
  }
  
  /**
   * スプラッシュ画面にメッセージを出力
   */
  public static void setSplashMessage(String msg) {
    Log.start(msg);
    splash.setMessage(msg);
    
    while (!splash.isDisposed()) {
      if (!splash.getDisplay().readAndDispatch()) break;
    }
  }
  
  /**
   * スプラッシュ画面を閉じる
   */
  public static void closeSplash() {
    splash.closeSplashWindow();
    splash = null;
  }
  
  /**
   * 画面が出ているうちに行う必要のある終了処理
   */
  public static void shutdownByMainWindow(Shell shell) {
    if (isShutdown) return;
    isShutdown = true;
    Log.start();
    Config conf = Config.get();
    
    // 終わるまで待つ
    BackgroundImageLoader.instance().shutdown();
    StopWatch sw = new StopWatch();
    sw.start();
    while (true) {
      while (!shell.isDisposed()) {
        if (!shell.getDisplay().readAndDispatch()) break;
      }
      try {
        Thread.sleep(100);
      }
      catch (InterruptedException e) {
        Log.e(e);
      }
      int imageLoadWaitNum = BackgroundImageLoader.instance().getWaitNum();
      int imageLoading = BackgroundImageLoader.instance().getRunNum();
      Log.d("imageLoadWaitNum = ", imageLoadWaitNum, ", imageLoading = ", imageLoading);
      if (imageLoadWaitNum + imageLoading == 0) break;
      if (sw.getTime() > conf.getImageLoadBackgroundThreadShutdownTimeout()) break;
    }
    
    System.gc();
    if (CachedImage.instanceCount.get() != 0) {
      Log.e("CachedImage instanceCount = ", CachedImage.instanceCount);
      GlobalErrorHandler.instance().setDebugMessage("CachedImage instanceCount = " + CachedImage.instanceCount, null);
    }
    Log.d("end");
  }

  /*
   * キャッチされない例外のハンドラー
   */
  static class ThreadErrorHandler implements Thread.UncaughtExceptionHandler {
    @Override
    public void uncaughtException(Thread t, final Throwable e) {
      Log.e("スレッドでキャッチされなかったエラー", t);
      Log.e(e);
      GlobalErrorHandler.instance().setFatalMessage("プログラム内部エラー", e);
    }
  }
}
