package cvj2.app.command;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import cvj2.util.Log;

/**
 * Command を処理する抽象クラス。<br>
 * 派生クラスでメソッドを宣言すると、名前から検索して呼ばれるようになる<br>
 * <br>
 * 例) <br>
 *   以下のようなクラスを作成すれば、<br>
 *   execCmd() に Command.FILE_EXIT を渡すと<br>
 *   handle_FILE_EXIT() が呼ばれるようになる。<br>
 * <br>
 * <code><pre>
 *   public class XXXCommandHandler extends CommandHandler {
 *     public void handle_FILE_EXIT() {
 *        System.exit(0);
 *     }
 *   }
 * </pre></code>
 */
public abstract class CommandHandler {
  /**
   * コマンド実行
   */
  public void execCmd(Command command) {
    Log.start(command);
    
    try {
      getMethod(command).invoke(this, new Object[0]);
    }
    catch (IllegalArgumentException e) {
      throw new RuntimeException(e);
    }
    catch (IllegalAccessException e) {
      throw new RuntimeException(e);
    }
    catch (InvocationTargetException e) {
      throw new RuntimeException(e);
    }
  }

  /*
   * メソッド取得 
   */
  private Method getMethod(Command command) {
    try {
      return this.getClass().getMethod("handle_" + command, new Class[0]);
    }
    catch (SecurityException e) {
      throw new RuntimeException(e);
    }
    catch (NoSuchMethodException e) {
      throw new RuntimeException(e);
    }
  }
}
