package cvj2.app.file;

import java.util.Observer;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.commons.lang3.time.StopWatch;
import org.eclipse.swt.widgets.Shell;

import cvj2.app.Config;
import cvj2.app.gui.util.GlobalErrorHandler;
import cvj2.app.gui.util.MsgBox;
import cvj2.file.fs.FileInfoList;
import cvj2.gui.swt.util.WaitCursor;
import cvj2.util.Log;

/**
 * FileInfoList の情報を(バックグラウンド/フォアグラウンドで)読み込む。<br>
 * メイン画面で、ユーザーがファイルをソートしようとした時のためにあらかじめファイル情報をバックグラウンドで読み込んでおくためのクラス。<br>
 * まだファイル情報が読み終わっていないときにソートされた時のために、時間制限つきの読み込みメソッドを用意する。<br>
 */
public class FileInfoReader {
  static private FileInfoReader instance = new FileInfoReader();
  
  // バックグラウンドに依頼する処理
  // 新しいリストがキューに投入されると、現在の処理を捨てて新しいリストを処理し始める
  // ※) バックグラウンド処理を中断するには、FileInfoList.NullList を入れる
  static private LinkedBlockingQueue<BgItem> workQueue = new LinkedBlockingQueue<BgItem>();
  static private volatile BgItem canceledItem;

  /**
   * インスタンス取得
   */
  static public FileInfoReader instance() {
    return instance;
  }
  
  /**
   * コンストラクタ
   */
  private FileInfoReader() {
    Thread t = new BgThread();
    t.setName("FileInfoReader");
    t.setDaemon(true);
    t.start();
  }
  
  /**
   * ファイル情報の読み込みを試す (時間がかかるときはメッセージボックスを出して確認する)
   * @return 全て読めたら、true
   */
  public boolean tryReadFg(Shell shell, final FileInfoList list) {
    // ファイルの情報を読み込む (時間制限付)
    int remainMs = WaitCursor.start(shell, new WaitCursor.Work<Integer>() {
      @Override public Integer run() {
        return readFg(
            list, 
            Config.get().getFileInfoReadInitializeTryTime(), 
            Config.get().getFileInfoReadInitializeTryItemRatio());
      }
    });
    resumeBg();
    
    // まだ読み込み途中なら、
    if (remainMs > 0) {
      boolean ret = MsgBox.okCancel("ソート",
          String.format("名前や拡張子以外で並べ替えを行うにはディレクトリ内の全てのファイルの詳しい属性が必要です。\n" + 
                        "ファイルの並べ替えに必要な情報の取得に、あと %d 秒くらいかかりそうです。\n\n" + 
                        "続けますか？", remainMs / 1000));
      if (!ret) return false;
    }
    
    // 最後まで読み込む
    WaitCursor.start(shell, new WaitCursor.Work<Object>() {
      @Override public Object run() {
        readAllFg(list);
        return null;
      }
    });
    resumeBg();
    return true;
  }
  
  /**
   * ファイル情報の全初期化
   */
  private void readAllFg(FileInfoList list) {
    cancelBg();
    int ret = readFg(list, Integer.MAX_VALUE, 1.0);
    assert ret == 0;
  }
  
  /**
   * ファイル情報の初期化を試みる (制限時間つき)
   * 
   * @param list ファイルリスト
   * @param tryTimeMs 制限時間 (ms)
   * @param tryItemRato 制限時間内に処理できたファイル情報の割合がこれ以上なら、制限時間を越えても全てのファイル情報を初期化する
   * @return 残りの情報を解析するのに必要そうな時間 (ms)
   *        全ての情報が初期化できたら、0
   */
  private int readFg(FileInfoList list, int tryTimeMs, double tryItemRato) {
    cancelBg();
    StopWatch sw = new StopWatch();
    sw.start();
    
    // 未初期化数
    int notInitCount = 0;
    for (int i = 0; i < list.size(); i++) {
      if (!list.get(i).isAttributeInitialized()) notInitCount++;
    }
    if (notInitCount <= 0) return 0;
    
    // 解析済み数 (ゼロ割を防ぐため、1に初期化)
    int initCount = 1;
    for (int i = 0; i < list.size(); i++) {
      // 制限時間チェック
      if (sw.getTime() > tryTimeMs) {
        // 解析済み数が制限達しなければ、終了
        double finishRatio = (double)initCount / notInitCount;
        if (finishRatio < tryItemRato) {
          return (int)(sw.getTime() / finishRatio);
        }
        // 制限数を越えて入れば、最後まで解析する
        tryTimeMs = Integer.MAX_VALUE;
      }
      
      if (!list.get(i).isAttributeInitialized()) {
        list.get(i).lastModified();
        initCount++;
      }
    }
    
    return 0;
  }
  
  /**
   * バックグラウンドで解析を依頼する
   * @param observer 解析状況が更新されたら、通知される (呼ばれる引数はnull)
   */
  public void readBg(FileInfoList list, Observer observer) {
    try {
      workQueue.put(new BgItem(list, observer));
    }
    catch (InterruptedException e) {
      Log.e(e);
      throw new RuntimeException(e);
    }
  }

  /**
   * バックグラウンド処理を中断する
   */
  public void cancelBg() {
    try {
      workQueue.put(new BgItem(FileInfoList.NULL_LIST, null));
    }
    catch (InterruptedException e) {
      Log.e(e);
      throw new RuntimeException(e);
    }
  }

  /**
   * 中断されていた処理を再開する
   */
  public void resumeBg() {
    BgItem i = canceledItem;
    if (i == null) return;
    try {
      workQueue.put(i);
    }
    catch (InterruptedException e) {
      Log.e(e);
      throw new RuntimeException(e);
    }
  }
  
  static private class BgItem {
    FileInfoList list;
    Observer observer;
    public BgItem(FileInfoList list, Observer observer) {
      this.list = list;
      this.observer = observer;
    }
  }
  
  /*
   * バックグラウンド読み込み処理
   */
  static private class BgThread extends Thread {
    Config config = Config.get();
    StopWatch sw = new StopWatch();

    @Override
    public void run() {
      try {
        while (true) {
          read();
        }
      }
      catch (Throwable e) {
        Log.e(e);
        GlobalErrorHandler.instance().setFatalMessage("ファイル情報の読み込み中にプログラム内部エラーが発生しました", e);
      }
    }

    private void read() throws InterruptedException {
      sw.reset();
      sw.start();
      BgItem item = workQueue.take();
      FileInfoList list = item.list;
      for (int i = 0; i < list.size(); i++) {
        if (workQueue.peek() != null) {
          canceledItem = item;
          break;
        }
        list.get(i).lastModified();
        
        if (sw.getTime() > config.getFileInfoReadBackgroundWorkTime()) {
          if (item.observer != null) item.observer.update(null, null);
          sleep(config.getFileInfoReadBackgroundSleepTime());
          sw.reset();
          sw.start();
        }
      }
      if (item.observer != null) item.observer.update(null, null);
    }
  }
}
