package cvj2.app.file.imgload;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.atomic.AtomicLong;

import org.eclipse.swt.widgets.Display;

import cvj2.app.Config;
import cvj2.file.fs.FileInfo;
import cvj2.util.Log;
import cvj2.util.MemorySizeMutex;
import cvj2.util.SizeCacheMap;

/**
 * 画像のバックグラウンド読み込みクラス。
 */
public class BackgroundImageLoader {
  final Object LOCK_OBJ = new String("LOCK_OBJ");
  
  private static BackgroundImageLoader instance = new BackgroundImageLoader();
  
  volatile boolean isShutdown = false;
  
  // 実行中 or 実行待ち スレッド
  // (追加はこのクラスから行う。削除はスレッド自身が行う)
  Set<BackgroundImageLoaderThread> threads = new HashSet<BackgroundImageLoaderThread>();
  // 最大スレッド数 (スレッド数がこの数より大きいと、スレッドは自殺する)
  volatile int maxThreads;

  // ワークメモリー制限
  // (タスクが参照する)
  MemorySizeMutex workMemory = new MemorySizeMutex(Config.get().getImageLoadBackgroundWorkMemory()); 
  
  // 実行中 or 実行待ち タスク 
  // (wait -> run への移動、run からの消滅はスレッドが行う)
  TreeSet<BackgroundImageLoaderTask> waitTasks = new TreeSet<BackgroundImageLoaderTask>();
  TreeSet<BackgroundImageLoaderTask> runTasks = new TreeSet<BackgroundImageLoaderTask>();
  // 実行待ちタスクのキュー (中身は waitTask と同期する)
  Queue<BackgroundImageLoaderTask> waitQueue = new LinkedList<BackgroundImageLoaderTask>();

  // キャッシュされているロード済み画像 
  // (登録はタスクが行う)
  SizeCacheMap<FileInfo, CachedImage> thumbCache = new SizeCacheMap<FileInfo, CachedImage>(0, new RmCallback(), true);
  SizeCacheMap<FileInfo, CachedImage> viewCache = new SizeCacheMap<FileInfo, CachedImage>(0, new RmCallback(), true);
  private class RmCallback implements SizeCacheMap.RemoveCallback<FileInfo, CachedImage> {
    @Override
    public void removed(final FileInfo key, final CachedImage val, final long size) {
      Log.start(key, val);
      Display.getDefault().asyncExec(new Runnable() { // デッドロックが発生するので、async でないといけない
        @Override public void run() {
          val.dispose();
        }
      });
    }
  }
  final AtomicLong processedPixelBytes = new AtomicLong();
  final AtomicLong processedFileBytes = new AtomicLong();
  final AtomicLong processedFileNum = new AtomicLong();
  
  /**
   * 統計情報取得
   * @return [実行待ちタスク数, 実行中タスク数, 
   *        サムネイルキャッシュ数, 画像キャッシュ数, 
   *        使用中作業メモリ, 
   *        累積処理ピクセルバイト数, 累積読み込みファイルバイト, 累積読み込みファイル数 ]
   */
  public long[] getStat() {
    synchronized (LOCK_OBJ) {
      return new long[] {
          waitTasks.size(), runTasks.size(),
          thumbCache.size(), viewCache.size(),
          workMemory.getUsedMemory(),
          processedPixelBytes.get(), processedFileBytes.get(), processedFileNum.get(),
      };
    }
  }
  
  /**
   * ファイル読み込み通知インターフェイス
   */
  public interface Notify {
    /** 状態変更通知 */
    public void stateChanged(FileInfo f);
  }
  
  private BackgroundImageLoader() {}
  
  /**
   * インスタンス取得
   */
  public static BackgroundImageLoader instance() {
    return instance;
  }
  
  /**
   * スレッド数などを再設定する
   */
  public void setup(int threadNum, long workMemorySize, long thumbCacheSize, long viewImageCacheSize) {
    Log.start(threadNum);
    assert !isShutdown;
    
    synchronized (LOCK_OBJ) {
      // 足りないスレッドを起動
      maxThreads = threadNum;
      while (threads.size() < maxThreads) {
        BackgroundImageLoaderThread t = new BackgroundImageLoaderThread();
        t.start();
        threads.add(t);
      }
    }
    
    // 作業メモリーサイズ設定
    workMemory.setLimit(workMemorySize);
    
    // キャッシュ容量設定
    thumbCache.setLimit(thumbCacheSize);
    viewCache.setLimit(viewImageCacheSize);
  }
  
  /**
   * シャットダウンする
   */
  public void shutdown() {
    Log.start();
    assert !isShutdown;
    
    synchronized (LOCK_OBJ) {
      isShutdown = true;
      
      // スレッド数を0にする
      maxThreads = 0;
      
      // キャッシュクリア
      thumbCache.setLimit(0);
      viewCache.setLimit(0);
      
      // 全てのタスクをキャンセルする
      cancelAllTasks();
      
      // 眠っているスレッドを起こす (実行待ちのタスクが無く、maxThread が 0 なので終了するはず)
      LOCK_OBJ.notifyAll();
    }
  }
  
  /**
   * シャットダウンしたかどうか
   */
  public boolean isShutdown() {
    return isShutdown;
  }

  /**
   * キャッシュをクリア
   */
  public void clearCache() {
    thumbCache.clear();
    viewCache.clear();
  }

  /** サムネイルキャッシュが使用しているメモリ数を取得 */
  public long getObjectBytesThumbnail() {
    // Key や CachedImage の付随情報は無視する
    return thumbCache.getTotalBytes();
  }

  /** 画像キャッシュが使用しているメモリ数を取得 */
  public long getObjectBytesView() {
    // Key や CachedImage の付随情報は無視する
    return viewCache.getTotalBytes();
  }
  
  /**
   * イメージが読み込み中か、読み込み待ちかどうか調べる
   */
  public boolean existsTask(ImageLoadProperty prop) {
    if (existsTask(runTasks, prop)) return true;
    if (existsTask(waitTasks, prop)) return true;
    return false;
  }
  
  /**
   * サムネイルが読み込み中か、読み込み待ちかどうか調べる
   */
  public boolean existsThumbnailTask(FileInfo f) {
    assert f != null;
    return existsTask(ImageLoadProperty.createForThumbnail(f, false));
  }
  
  /**
   * サムネイルが読み込み中か調べる
   */
  public boolean existsThumbnailRunTask(FileInfo f) {
    assert f != null;
    return existsTask(runTasks, ImageLoadProperty.createForThumbnail(f, false));
  }

  private boolean existsTask(TreeSet<BackgroundImageLoaderTask> tasks, ImageLoadProperty prop) {
    BackgroundImageLoaderTask tmpTask = new BackgroundImageLoaderTask(prop, null, null);
    synchronized (LOCK_OBJ) {
      if (tasks.contains(tmpTask)) return true;
    }
    return false;
  }
  
  /**
   * イメージが読み込み済みであるかどうか調べる (サムネイル)
   */
  public boolean existsInThumbCache(FileInfo f) {
    assert f != null;
    return thumbCache.exists(f);
  }
  /**
   * 画像サイズなどは気にしないでキャッシュを取得する (一枚画像)
   */
  public CachedImage findViewCache(FileInfo f) {
    assert f != null;
    return viewCache.get(f);
  }

  /**
   * 全タスクの情報を取得
   */
  public Set<ImageLoadProperty> getAllTaskInfo() {
    Set<ImageLoadProperty> props = new HashSet<ImageLoadProperty>();
    synchronized (LOCK_OBJ) {
      for (BackgroundImageLoaderTask t : waitTasks) {
        if (!t.isCanceled()) props.add(t.getLoadProperty());
      }
      for (BackgroundImageLoaderTask t : runTasks) {
        if (!t.isCanceled()) props.add(t.getLoadProperty());
      }
    }
    return props;
  }

  /**
   * 実行待ち数
   */
  public int getWaitNum() {
    synchronized (LOCK_OBJ) {
      assert waitTasks.size() == waitQueue.size();
      int num = 0;
      for (BackgroundImageLoaderTask t : waitTasks) {
        if (!t.isCanceled()) num++;
      }
      return num;
    }
  }

  /**
   * 実行中の数
   */
  public int getRunNum() {
    synchronized (LOCK_OBJ) {
      return runTasks.size();
    }
  }
  
  /**
   * イメージを取得する (サムネイル)
   */
  public CachedImage getThumbnail(FileInfo fileOrDir, Notify notify, boolean forceAddToThumbnailCache) {
    Log.start(fileOrDir);
    return getImageOrStartTask(thumbCache, ImageLoadProperty.createForThumbnail(fileOrDir, forceAddToThumbnailCache), notify);
  }
  
  /**
   * イメージを取得する (一枚画像)
   * @param notify 読み込み後通知
   * @return 既にキャッシュにあれば、それをかえす
   */
  public CachedImage getViewImage(ImageLoadProperty prop, Notify notify) {
    Log.start(prop);
    return getImageOrStartTask(viewCache, prop, notify);
  }
  
  /*
   * 画像読み込み (キャッシュにあればそれをかえす)
   */
  private CachedImage getImageOrStartTask(SizeCacheMap<FileInfo, CachedImage> cache, ImageLoadProperty prop, Notify notify) {
    if (isShutdown) return null;

    // キャッシュにあれば、それをかえす
    if (!prop.isForceAddToThumbnailCache()) {
      CachedImage img = cache.get(prop.getFileOrDir());
      if (img != null && img.equalsProperty(prop)) return img;
    }

    synchronized (LOCK_OBJ) {
      // タスクに入っていれば何もしない
      if (existsTask(prop)) return null;
  
      // タスク起動
      BackgroundImageLoaderTask task = new BackgroundImageLoaderTask(prop, notify, cache);
      waitTasks.add(task);
      waitQueue.add(task);
      Log.d("new task: ", task);
      assert waitTasks.size() == waitQueue.size();
      
      // 眠っているスレッドをどれか一つ起こす
      LOCK_OBJ.notify();

      return null;
    }
  }
  
  /**
   * 実行待ちのタスクを全てキャンセルする
   */
  public void cancelWaitTasks() {
    Log.start();
    
    synchronized (LOCK_OBJ) {
      assert waitTasks.size() == waitQueue.size();
      invokeCancel(waitTasks, null);
    }
  }

  /**
   * 全てのタスクをキャンセルする
   */
  public void cancelAllTasks() {
    Log.start();
    
    synchronized (LOCK_OBJ) {
      assert waitTasks.size() == waitQueue.size();
      invokeCancel(waitTasks, null);
      invokeCancel(runTasks, null);
    }
  }
  
  /**
   * タスクをキャンセルする
   * @param loadProp キャンセルするタスク。null なら、全て
   */
  public void cancelTask(ImageLoadProperty loadProp) {
    assert loadProp != null;
    Log.start(loadProp);
    
    synchronized (LOCK_OBJ) {
      assert waitTasks.size() == waitQueue.size();
      invokeCancel(waitTasks, loadProp);
      invokeCancel(runTasks, loadProp);
    }
  }
  
  /**
   * タスクのキャンセル (task.cancel() を呼ぶだけ)
   * @param loadProp キャンセル対象。null の場合は全て
   */
  private void invokeCancel(TreeSet<BackgroundImageLoaderTask> tasks, ImageLoadProperty loadProp) {
    for (BackgroundImageLoaderTask t : tasks) {
      if (loadProp != null && !loadProp.equals(t.getLoadProperty())) continue;
      Log.d("cancel: ", t);
      t.cancel();
    }
  }
}
