package cvj2.app.file.imgload;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.text.DecimalFormat;
import java.util.concurrent.atomic.AtomicInteger;

import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.widgets.Display;
import org.eclipse.wb.swt.SWTResourceManager;

import cvj2.app.Config;
import cvj2.app.file.imgload.BackgroundImageLoader.Notify;
import cvj2.app.file.thumbnail.ThumbnailCache;
import cvj2.app.gui.util.GlobalErrorHandler;
import cvj2.app.gui.widgets.ListPanel;
import cvj2.file.fs.FileInfo;
import cvj2.file.fs.FileInfoList;
import cvj2.img.ImageInfo;
import cvj2.img.ImageUtil;
import cvj2.img.SWTImageConverter;
import cvj2.img.converter.ImageConverter;
import cvj2.img.converter.ImageConverterRotate;
import cvj2.img.converter.ImageSharpnessConverter;
import cvj2.img.converter.ImageSizeConverter;
import cvj2.img.converter.ImageSizeConverterAffineTransform;
import cvj2.img.converter.ImageSizeConverterScaledInstance;
import cvj2.img.converter.ImageTypeConverter;
import cvj2.img.loader.ImageLoader2;
import cvj2.util.Log;
import cvj2.util.SizeCacheMap;

/**
 * 画像読み込みタスク。<br>
 * <pre>
 * {@link BackgroundImageLoader} で作成され、{@link BackgroundImageLoaderThread} から起動される
 * 
 * 画像の読み込み処理は以下の順で行われる
 * 1) ディレクトリが指定されたら、読み込み対象の画像ファイルを見つける
 * 2) 1枚 or 2枚の画像をファイルから読み込む
 * 3) 2枚の画像だったら、1つの画像に結合する
 * 4) 使用する色数を統一する
 * 5) 回転
 * 6) 拡大縮小
 * 7) 輪郭強調
 * 8) AWT の BufferedImage から、SWT の ImageData に変換する
 * 9) ImageData を SWT の Image に変換する
 * </pre>
 */
public class BackgroundImageLoaderTask implements Runnable, Comparable<BackgroundImageLoaderTask> {
  /** 実行中のタスク数 (統計情報用) */
  static public final AtomicInteger runningCount = new AtomicInteger();
  static private long maxTaskId = 0;

  private long taskId;
  volatile private boolean isCanceled = false;

  private ImageLoadProperty loadProperty;
  private ImageLoadResult loadResult = new ImageLoadResult();
  private Notify notify;
  private SizeCacheMap<FileInfo, CachedImage> cache;

  /**
   * コンストラクタ
   */
  public BackgroundImageLoaderTask(ImageLoadProperty prop, Notify notify, SizeCacheMap<FileInfo, CachedImage> cache) {
    this.taskId = maxTaskId++;
    this.loadProperty = prop;
    this.notify = notify;
    this.cache = cache;
  }

  /**
   * ロード情報を取得
   */
  public ImageLoadProperty getLoadProperty() {
    return loadProperty;
  }
  
  /**
   * タスクの実行
   */
  @Override
  public void run() {
    runningCount.incrementAndGet();
    try {
      // ロード実行
      load();
    }
    catch (InterruptedException e) {
      // エラー処理は画像ロードスレッドに任せる
      throw new RuntimeException(e);
    }
    finally {
      runningCount.decrementAndGet();
    }
  }

  /**
   * ロード実行
   */
  private void load() throws InterruptedException {
    Log.start(this);
    Config conf = Config.get();
    BackgroundImageLoader loader = BackgroundImageLoader.instance();
    FileInfo fileOrDir = loadProperty.getFileOrDir();
    boolean memoryGot = false;

    try {
      if (!checkImageLoadCancel()) return;
      
      // 読み込み開始を通知する
      notify.stateChanged(fileOrDir);

      // サムネイルキャッシュに見つかったら、戻る
      if (findThumbnail()) return;
      
      // ロードできる画像を探す
      FileInfo imgFile = findImage(fileOrDir, conf.getMainWindowThumbnailSearchDirDepth());
      if (imgFile == null) throw new ImageNotFoundException();
      if (!checkImageLoadCancel()) return;

      // 必要メモリー容量計算
      long memoryReserve = getNecessaryWorkMemory(imgFile, loadProperty.getFile2());
      loadResult.convertDebugMessage.append("画像ロード時に必要な作業メモリー: " + DecimalFormat.getNumberInstance().format(memoryReserve) + " バイト"); 
      if (!checkImageLoadCancel()) return;
      
      // メモリー確保
      loader.workMemory.get(memoryReserve);
      memoryGot = true;
      if (memoryReserve > loader.workMemory.getLimit()) {
        // MemorySizeMutex はオーバーしていても同時に1つだけは通過するので改めてチェックして警告を出す
        Log.w("work memory over: ", memoryReserve, " ", fileOrDir);
      }
      if (!checkImageLoadCancel()) return;

      // 画像読み込み
      Integer dpi = loadProperty.getZoomRatio() == null ? null : conf.getImageLoadPdfDpi();
      Dimension loadSize = loadProperty.getTargetSize() == null ? null : new Dimension(
          (int)(loadProperty.getTargetSize().width * conf.getImageLoadPdfFitZoomRatio()), 
          (int)(loadProperty.getTargetSize().height * conf.getImageLoadPdfFitZoomRatio()));
      if (loadProperty.isThumbnail()) {
        // PDF画像のサムネイルの場合は、原寸で読み込んでから縮小する
        dpi = conf.getImageLoadPdfDpi();
        loadSize = null;
      }
      ImageLoader2 loader2 = new ImageLoader2();
      BufferedImage bi = loader2.load(imgFile, loadProperty.getFile2(), dpi, loadSize);
      loadResult.originalImageInfo = new ImageInfo(bi);
      loadResult.originalImageInfos = loader2.getImageInfos();
      if (!checkImageLoadCancel()) return;

      // 画像形式変換
      bi = createTypeConverter().cnv(bi);
      if (!checkImageLoadCancel()) return;

      // 回転
      bi = createRotateConverter().cnv(bi);
      if (!checkImageLoadCancel()) return;
      
      // サイズ変更
      bi = createSizeConverter(bi).cnv(bi);
      if (!checkImageLoadCancel()) return;

      // 輪郭強調
      bi = createSharpnessConverter(bi, loadResult.originalImageInfo).cnv(bi);
      if (!checkImageLoadCancel()) return;
      
      // AWT から SWT に変換
      ImageData id = SWTImageConverter.cnv(bi);
      bi = null;
      if (!checkImageLoadCancel()) return;

      // イメージのリソースに変換
      Image img = new Image(Display.getDefault(), id);
      CachedImage ci = new CachedImage(loadProperty, img, false, loadResult); 
      
      // キャッシュサイズチェック
      if (cache.getLimit() > 0 && ci.getObjectBytes() > cache.getLimit()) {
        String msg = "現在の表示設定ではこの画像をメモリキャッシュに収めることができません。\n" + 
            "  " + loadProperty.getFileOrDir() + "\n\n" + 
            "この画像に必要なメモリ " + ci.getObjectBytes() + " バイト\n" + 
            "キャッシュメモリ " + cache.getLimit() + " バイト\n\n" +
            "キャッシュメモリ設定、最大読み込み画像サイズ設定、最大画像拡大率設定 を見直してみてください";
        GlobalErrorHandler.instance().setErrorMessage(msg, null);
      }
      
      Log.d("load successful: ", this);
 
      if (!ci.ref()) throw new RuntimeException("program internal error");
      try {
        // 通知
        cache.put(fileOrDir, ci);
        notify.stateChanged(fileOrDir);
        
        // サムネイルキャッシュに保存する (ディレクトリの画像の場合)
        if (!fileOrDir.equals(imgFile) || loadProperty.isForceAddToThumbnailCache()) {
          saveThumbnail(ci);
        }
      }
      finally {
        ci.unref();
      }

      // サムネイル強制登録の場合は、キャッシュに登録した後で通知する
      if (loadProperty.isForceAddToThumbnailCache()) {
        notify.stateChanged(fileOrDir);
      }
      
      // 統計情報
      setStat(imgFile, ci);
    }
    catch (IOException e) {
      // ロード失敗
      Log.w("load failed: ", this);
      Log.w(e);
      Image img = SWTResourceManager.getImage(ListPanel.class, "image_broken.png");
      loadResult.loadFailMessage = "画像のロードに失敗しました";
      loadResult.originalImageInfo = new ImageInfo(BufferedImage.TYPE_CUSTOM, img.getBounds().width, img.getBounds().height);
      cache.put(fileOrDir, new CachedImage(loadProperty, img, true, loadResult));
      notify.stateChanged(fileOrDir);
    }
    catch (ImageNotFoundException e) {
      // 画像が見つからない
      Log.d("image not found: ", this);
      Image img = SWTResourceManager.getImage(ListPanel.class, "folder.png");
      loadResult.loadFailMessage = "";
      loadResult.originalImageInfo = new ImageInfo(BufferedImage.TYPE_CUSTOM, img.getBounds().width, img.getBounds().height);
      cache.put(fileOrDir, new CachedImage(loadProperty, img, true, loadResult));
      notify.stateChanged(fileOrDir);
    }
    catch (TooLargeImageException e) {
      // 画像が大きすぎる
      Log.d("too large image: ", this);
      Image img = SWTResourceManager.getImage(ListPanel.class, "image_toolarge.png");
      loadResult.loadFailMessage = e.getMessage();
      loadResult.originalImageInfo = new ImageInfo(BufferedImage.TYPE_CUSTOM, img.getBounds().width, img.getBounds().height);
      cache.put(fileOrDir, new CachedImage(loadProperty, img, true, loadResult));
      notify.stateChanged(fileOrDir);
    }
    catch (OutOfMemoryError e) {
      System.gc();
      Log.e("out of memory error: ", this);
      Log.e(e);
      // OutOfMemoryError
      Image img = SWTResourceManager.getImage(ListPanel.class, "image_toolarge.png");
      loadResult.loadFailMessage = "画像ロード中にメモリー不足のシステムエラーが起きました\n(かなり深刻な状態です。システム設定を見直してみてください。)";
      loadResult.originalImageInfo = new ImageInfo(BufferedImage.TYPE_CUSTOM, img.getBounds().width, img.getBounds().height);
      cache.put(fileOrDir, new CachedImage(loadProperty, img, true, loadResult));
      notify.stateChanged(fileOrDir);
      GlobalErrorHandler.instance().setFatalMessage("画像ロード中にメモリー不足のシステムエラーが起きました\n(かなり深刻な状態です。システム設定を見直してみてください。)", e);
    }
    finally {
      // 使用メモリー解放
      if (memoryGot) loader.workMemory.release();
    }
  }

  /*
   * 画像読み込みに必要な作業メモリ量
   */
  private long getNecessaryWorkMemory(FileInfo file1, FileInfo file2) throws IOException, TooLargeImageException {
    Dimension size1 = getImageUseMemory(file1);
    Dimension size2 = getImageUseMemory(file2);

    // 画像サイズが不明なので、推定値をかえす
    if (size1 == null || size2 == null) {
      return guessWorkMemory(file1, size1) + guessWorkMemory(file2, size2);
    }
    
    // 横に並べた画像の、ピクセル数 × 32bit × 2枚
    long imgSize = Math.max(size1.height, size2.height) * (size1.width + size2.width) * 4 * 2;
    
    if (loadProperty.getZoomRatio() != null) {
      // ズームする場合は、倍率をかける
      double z = loadProperty.getZoomRatio();
      if (z < 1.0) return imgSize;
      return (long)(z * imgSize);
    }
    else {
      // サイズ指定の場合は、それを使用する
      Dimension size = loadProperty.getTargetSize();
      int sizeFromTarget = size.width * size.height * 4 * 2;
      return Math.max(sizeFromTarget, imgSize);
    }
  }
  // 画像サイズ
  private Dimension getImageUseMemory(FileInfo f) throws IOException, TooLargeImageException {
    if (f == null) return new Dimension();
    
    Config conf = Config.get();

    // PDF画像サイズ取得
    Dimension pdfSize = f.getImageSize(conf.getImageLoadPdfDpi());
    if (pdfSize != null) {
      if (loadProperty.getTargetSize() == null) return pdfSize;
      pdfSize = ImageUtil.calcImageFitSize(pdfSize, loadProperty.getTargetSize());
      pdfSize = new Dimension((int)(pdfSize.width * conf.getImageLoadPdfFitZoomRatio()), (int)(pdfSize.height * conf.getImageLoadPdfFitZoomRatio()));
      return pdfSize;
    }
    
    // ファイルサイズチェック
    if (f.length() > conf.getImageLoadMaxFileSizeBytes()) throw new TooLargeImageException("ファイルが大きすぎます " + f.length() + " バイト");
    
    // 画像サイズ取得
    InputStream in = f.openStream();
    if (in == null) return null;
    in.mark(4096);
    try {
      Dimension size = ImageUtil.readImageInfo(in);
      if (size == null) return null;
      long max = conf.getImageLoadMaxHeight() * conf.getImageLoadMaxWidth(); // 32bitピクセルと仮定
      if (size.width * size.height > max) throw new TooLargeImageException("画像が大きすぎます " + size.width + " x " + size.height + "");
      return size;
    }
    finally {
      in.close();
    }
  }
  // 画像サイズ or ファイルサイズから推定
  private long guessWorkMemory(FileInfo file, Dimension size) {
    if (size != null) {
      return size.width * size.height * 4 * 2;
    }
    return file.length() * 50;
  }

  /*
   * キャンセルチェック
   */
  private boolean checkImageLoadCancel() {
    if (!isCanceled) return true;
    Log.d("load canceled: ", this);
    
    // キャンセルされたことを通知する
    notify.stateChanged(loadProperty.getFileOrDir());
    
    return false;
  }

  /*
   * タイプ変換
   */
  private ImageConverter createTypeConverter() {
    loadResult.convertDebugMessage.append("画像変換: " + ImageUtil.getBufferedImageTypeString(Config.get().getImageTypeConverterType()));
    return new ImageTypeConverter(Config.get().getImageTypeConverterType());
  }
  
  /*
   * 回転
   */
  private ImageConverterRotate createRotateConverter() {
    loadResult.convertDebugMessage.append("画像回転: " + loadProperty.getRotate());
    return new ImageConverterRotate(loadProperty.getRotate());
  }
  
  /*
   * サイズ変換
   */
  private ImageSizeConverter createSizeConverter(BufferedImage bi) {
    Config conf = Config.get();
    
    Dimension targetSize = loadProperty.getTargetSize();
    Double zoomRatio = loadProperty.getZoomRatio();
    ImageSizeConverterAffineTransform cnv;
    
    if (targetSize != null) cnv = new ImageSizeConverterAffineTransform(targetSize, 0);
    else                    cnv = new ImageSizeConverterAffineTransform(zoomRatio, 0);
    double ratio = cnv.getZoomRatio(bi);

    String comment = "???"; 
    try {
      if (ratio >= conf.getImageSizeConverterChangeMethodZoomRatio()) {
        comment = "Affine: " + conf.getImageSizeConverterAffineType();
        if (zoomRatio != null) {
          return new ImageSizeConverterAffineTransform(zoomRatio, conf.getImageSizeConverterAffineType());
        }
        else {
          return new ImageSizeConverterAffineTransform(targetSize, conf.getImageSizeConverterAffineType());
        }
      }
      else {
        comment = "Scaled: " + conf.getImageSizeConverterScaledHint();
        if (zoomRatio != null) {
          return new ImageSizeConverterScaledInstance(zoomRatio, conf.getImageSizeConverterScaledHint());
        }
        else {
          return new ImageSizeConverterScaledInstance(targetSize, conf.getImageSizeConverterScaledHint());
        }
      }
    }
    finally {
      loadResult.convertDebugMessage.append("サイズ変換 " + comment);
    }
  }

  /*
   * 輪郭強調
   */
  private ImageSharpnessConverter createSharpnessConverter(BufferedImage bi, ImageInfo orgImageInfo) {
    Config conf = Config.get();
    double[][] params = conf.getImageSharpnessConverterParamsByZoomRatios();
    if (loadProperty.isThumbnail()) params = conf.getImageSharpnessConverterParamsByZoomRatiosThumbnail();
    double param = params[params.length - 1][1];
    
    double zoomRatio = (double)bi.getWidth() / (double)orgImageInfo.width;
    for (int i = 0; i < params.length; i++) {
      if (zoomRatio < params[i][0]) {
        param = params[i][1];
        break;
      }
    }
    loadResult.convertDebugMessage.append("輪郭強調: " + param);
    
    return new ImageSharpnessConverter((float)param);
  }
  
  /*
   * ロードできる画像ファイルを探す
   */
  private FileInfo findImage(FileInfo f, int depth) {
    if (!f.isDirectory()) {
      // 画像ファイルが見つかった
      if (f.isImageFile()) return f;
      return null;
    }
    
    if (depth <= 0) return null;

    // 子ファイルの先頭を探す
    FileInfoList childs = f.getList();
    if (childs.size() == 0) return null;
    return findImage(childs.get(0), depth - 1);
  }

  /*
   * サムネイルキャッシュから探す
   */
  private boolean findThumbnail() {
    if (!loadProperty.isThumbnail()) return false;
    
    // キャッシュを探す
    ImageData id = ThumbnailCache.instance().find(loadProperty.getFileOrDir(), loadProperty.getTargetSize());
    if (id == null) return false;
    
    // イメージのリソースに変換
    Image img = new Image(Display.getDefault(), id);
    CachedImage ci = new CachedImage(loadProperty, img, false, loadResult); 
    
    Log.d("found in thumbnail cache: ", this);

    // 通知
    cache.put(loadProperty.getFileOrDir(), ci);
    notify.stateChanged(loadProperty.getFileOrDir());
    
    return true;
  }
  
  /*
   * サムネイルキャッシュに登録する
   */
  private void saveThumbnail(CachedImage ci) {
    assert loadProperty.isThumbnail();
    assert ci.getRefCount() > 0;
    
    ImageData id = ci.getImage().getImageData();
    ThumbnailCache.instance().put(loadProperty.getFileOrDir(), loadProperty.getTargetSize(), id);
  }
  
  private class TooLargeImageException extends Exception {
    private static final long serialVersionUID = 1L;
    public TooLargeImageException(String message) {
      super(message);
    }
  }
  private class ImageNotFoundException extends Exception {
    private static final long serialVersionUID = 1L;
  }

  /*
   * 統計情報セット
   */
  private void setStat(FileInfo imgFile, CachedImage ci) {
    long pixel = ci.getObjectBytes() + (loadResult.originalImageInfo.width * loadResult.originalImageInfo.height * 4);
      
    long fileLen = imgFile.length();
    if (loadProperty.getFile2() != null) {
      fileLen += loadProperty.getFile2().length();
    }
    
    BackgroundImageLoader.instance().processedPixelBytes.addAndGet(pixel);
    BackgroundImageLoader.instance().processedFileBytes.addAndGet(fileLen);
    BackgroundImageLoader.instance().processedFileNum.incrementAndGet();
  }
  
  /**
   * タスクをキャンセルする
   */
  public void cancel() {
    isCanceled = true;
  }
  
  /**
   * キャンセルされたか
   */
  public boolean isCanceled() {
    return isCanceled;
  }

  /** TreeSet に格納するためにオーバーライド */
  @Override
  public int compareTo(BackgroundImageLoaderTask other) {
    return this.loadProperty.compareTo(other.loadProperty);
  }
  
  /** オーバーライド */
  @Override
  public boolean equals(Object obj) {
    if (obj == null) return false;
    return this.compareTo((BackgroundImageLoaderTask) obj) == 0;
  }
  
  /** オーバーライド */
  @Override
  public int hashCode() {
    return loadProperty.hashCode();
  }

  /** オーバーライド */
  @Override
  public String toString() {
    return "ID=" + taskId + "|" + loadProperty;
  }
}
