package cvj2.app.file.imgload;

import cvj2.app.gui.util.GlobalErrorHandler;
import cvj2.util.Log;

/**
 * 画像読み込みスレッド
 */
public class BackgroundImageLoaderThread extends Thread {
  static private int threadId = 0;
  
  /**
   * コンストラクタ
   */
  BackgroundImageLoaderThread() {
    String name = "ImgLd-" + threadId++;
    Log.start(name);
    setName(name);
    setDaemon(true);
    setPriority(NORM_PRIORITY - 1);
  }

  /**
   * 処理の本体
   */
  @Override
  public void run() {
    try {
      Log.d("start thread");
      
      while (true) {
        // 実行待ちタスクを取得 (取得できるまでブロックする)
        BackgroundImageLoaderTask task = getWaitTask();
        if (task == null) break;

        // タスク実行
        try {
          Log.d("start task: ", task);
          task.run();
          Log.d("end task: ", task);
        }
        catch (Throwable e) {
          Log.e("abend task: ", task);
          Log.e(e);
          GlobalErrorHandler.instance().setFatalMessage("画像ロード中にプログラム内部エラーが発生しました", e);
        }
        finally {
          // 実行中リストから削除
          removeFromRunTaskList(task);
        }
      }
      Log.d("end thread");
    }
    catch (Throwable e) {
      Log.d("abend thread");
      Log.e(e);
      GlobalErrorHandler.instance().setFatalMessage("バックグラウンド画像読み込みスレッドでプログラム内部エラーが発生しました", e);
    }
    finally {
      // 眠っている他のスレッドを起こす
      synchronized (BackgroundImageLoader.instance().LOCK_OBJ) {
        BackgroundImageLoader.instance().LOCK_OBJ.notify();
      }
    }
  }

  /**
   * 実行待ちタスクを取得 (取得できるまでブロックする)
   * @return シャットダウンされていたり、スレッド実行数がオーバーしていたら、null
   */
  private BackgroundImageLoaderTask getWaitTask() throws InterruptedException {
    BackgroundImageLoader loader = BackgroundImageLoader.instance();
    
    synchronized (loader.LOCK_OBJ) {
      // タスクが取得できるようになるまでループする
      while (loader.waitQueue.size() == 0) {
        // シャットダウンされていたら、終了
        if (loader.isShutdown) return null;
        
        // キューにタスクが入るまで眠る
        loader.LOCK_OBJ.wait();
        
        // スレッドを終了するかどうかチェックする
        if (!checkMaxThredNum()) return null;
      }
      
      // 実行待ちキューからタスクを取得
      assert loader.waitTasks.size() == loader.waitQueue.size();
      BackgroundImageLoaderTask task = loader.waitQueue.remove();
      assert task != null;
      loader.waitTasks.remove(task);
      assert loader.waitTasks.size() == loader.waitQueue.size();
      
      // 実行中リストに登録
      loader.runTasks.add(task);
      
      return task;
    }
  }

  // 実行中リストから削除
  private void removeFromRunTaskList(BackgroundImageLoaderTask task) {
    BackgroundImageLoader loader = BackgroundImageLoader.instance();
    
    synchronized (loader.LOCK_OBJ) {
      boolean b = loader.runTasks.remove(task);
      assert b;
    }
  }

  /** 
   * スレッド数のチェック
   * @return このスレッドが終了したなら、false
   */
  private boolean checkMaxThredNum() {
    BackgroundImageLoader loader = BackgroundImageLoader.instance();
    
    // 最大スレッド数チェック
    if (loader.maxThreads >= loader.threads.size()) return true;

    Log.d("スレッド最大数が縮小していたので、このスレッドを終了します");
    boolean b = loader.threads.remove(this);
    assert b;
    return false;
  }
}
