package cvj2.app.file.imgload;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.util.concurrent.atomic.AtomicInteger;

import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;

import cvj2.app.gui.util.GlobalErrorHandler;
import cvj2.img.ImageInfo;
import cvj2.util.Log;
import cvj2.util.SizeCountable;

/**
 * メモリ内にキャッシュされた読み込み済み画像
 */
public class CachedImage implements SizeCountable {
  /** 存在するインスタンス数 (デバッグ用) */
  static public final AtomicInteger instanceCount = new AtomicInteger();

  private ImageLoadProperty loadProp;
  private Image image;
  private ImageInfo info;
  private boolean systemResource;
  private ImageLoadResult loadResult;
  
  private int refCount = 0;
  private boolean reserveDispose = false;

  /**
   * コンストラクタ
   * @param loadProp 画像ロード条件
   * @param image 画像
   * @param systemResource システムリソースなので、dispose() を呼ばなくてもよいかどうか
   * @param loadResult 画像ロード結果情報
   */
  public CachedImage(ImageLoadProperty loadProp, Image image, boolean systemResource, ImageLoadResult loadResult) {
    instanceCount.incrementAndGet();
    this.loadProp = loadProp;
    this.image = image;
    this.systemResource = systemResource;
    this.loadResult = loadResult;
    
    if (loadResult.originalImageInfos[0] == null) {
      loadResult.originalImageInfos[0] = loadResult.originalImageInfo;
    }    

    // ※) Image.getBounds() はここで呼ばなければならない。
    // このメソッドはSWT内部のグローバルロックを取得しているため、
    // SizedCacheMap のロックをUIスレッドとバックグラウンドスレッドで取り合ってデッドロックする。
    //
    // おそらく、
    // 1) UIスレッドがSWTのロックを持っている状態で画面描写などしており、
    // 2) バックグラウンドスレッドが SizedCacheMap に put() し、(※ SizedCacheMap のロックを取得)
    // 3) その内部で getObjectBytes() 経由で Image.getBounds() が呼ばれると、SWTのロックを取得しようとして待ちになり、
    // 5) UIスレッドが SizedCacheMap にアクセスしようとしてデッドロックになる。
    // 
    // (一度 getBounds() を呼んでおけば、2度目以降はグローバルロックを取得しないらしいが、
    //  今後の実装でどうなるかわからないので、一応保存しておく。)
    // 
    // ※) その後、SizedCachedMap を修正したので、以前の実装でも問題なくなった。
    Rectangle rect = image.getBounds();
    this.info = new ImageInfo(BufferedImage.TYPE_CUSTOM, rect.width, rect.height);
  }

  /**
   * ロードした画像。<br>
   * ※) このメソッドを呼ぶ前に ref() して、呼んだ後に unref() すること!!!
   */
  public Image getImage() {
    assert refCount > 0;
    return image;
  }

  /**
   * 画像の読み込み条件
   */
  public ImageLoadProperty getLoadProperty() {
    return loadProp;
  }
  
  /**
   * 画像サイズ
   */
  public Dimension getSize() {
    return info.getSize();
  }
  
  /**
   * 読み込み結果情報
   */
  public ImageLoadResult getLoadResult() {
    return loadResult;
  }
  
  /**
   * dispose() しなくてもいいか。<br>
   * @return 内部のデフォルトイメージの場合は、true
   */
  public boolean isSystemResource() {
    return systemResource;
  }
  
  /**
   * 画像のファイルと読み込み条件が一致するかどうかチェックする
   */
  public boolean equalsProperty(ImageLoadProperty prop) {
    return this.loadProp.equals(prop);
  }
  
  /**
   * 画像のメモリー占有バイト数
   */
  @Override
  public long getObjectBytes() {
    if (systemResource) return 0;
    // 32bitと仮定
    return info.width * info.height * 4;
  }

  /** 参照カウント取得 */
  synchronized public int getRefCount() {
    return refCount;
  }
  
  /**
   * 参照カウントを増やす。<br>
   * ※) 参照している間は、dispose されない
   */
  synchronized public boolean ref() {
    if (reserveDispose) return false;
    refCount++;
    return true;
  }
  
  /**
   * 参照カウントを減らす
   */
  synchronized public void unref() {
    refCount--;
    if (refCount == 0 && reserveDispose) {
      Log.d("reserveDispose = ", reserveDispose, ", ", toString());
      dispose();
    }
  }
  
  /**
   * リソースの破棄
   */
  synchronized void dispose() {
    Log.start(toString());
    reserveDispose = true;
    if (refCount != 0) {
      Log.d("refCount = ", refCount, ", ", toString());
      return;
    }
    instanceCount.decrementAndGet();
    if (image != null && !image.isDisposed() && !systemResource) image.dispose();
    image = null;
  }

  /** オーバーライド */
  @Override
  protected void finalize() throws Throwable {
    if (image != null || refCount > 0) {
      Log.e("not disposed: ", toString());
      Log.e(new Exception("スタックトレースを表示します"));
      GlobalErrorHandler.instance().setFatalMessage("画像リソース破棄中に、プログラム内部エラーが発生しました", null);
    }
  }

  /** オーバーライド */
  @Override
  public String toString() {
    return "" + loadProp + "|" + (systemResource ? "SR":"  ") + "|" + (reserveDispose ? "RD":"  ") + 
          "|ref=" + refCount + "|orgImgInfo=" + loadResult.originalImageInfo + "|img=" + image + 
          "|bytes=" + getObjectBytes();
  }
}
