package cvj2.app.file.imgload;

import java.awt.Dimension;

import cvj2.app.Config;
import cvj2.file.fs.FileInfo;
import cvj2.img.ImageRotate;

/**
 * 画像ロード時の条件設定
 */
public class ImageLoadProperty implements Comparable<ImageLoadProperty> {
  private FileInfo fileOrDir;
  private FileInfo file2;
  private ImageRotate rotate;
  private Dimension targetSize;
  private Double zoomRatio;
  private boolean isThumbnail;
  private boolean forceAddToThumbnailCache;

  /**
   * コンストラクタ。<br>
   * targetSize か zoomRatio はどちらかが設定されていなければならない
   * @param fileOrDir 読み込み対象
   * @param rotate 回転
   * @param targetSize サイズ変換する場合のサイズ (サイズは、回転後のサイズ)
   * @param zoomRatio サイズ変換する場合の拡大縮小率
   */
  public ImageLoadProperty(FileInfo fileOrDir, FileInfo file2, ImageRotate rotate, Dimension targetSize, Double zoomRatio) {
    assert fileOrDir != null;
    assert file2 == null || (!fileOrDir.isDirectory() && fileOrDir.isImageFile() && !file2.isDirectory() && file2.isImageFile());
    assert rotate != null;
    assert targetSize != null || zoomRatio != null;
        
    this.fileOrDir = fileOrDir;
    this.file2 = file2;
    this.rotate = rotate;
    this.targetSize = targetSize;
    this.zoomRatio = zoomRatio;
  }
  
  /**
   * サムネイル用オブジェクトを生成
   */
  public static ImageLoadProperty createForThumbnail(FileInfo fileOrDir, boolean forceAddToThumbnailCache) {
    ImageLoadProperty ret = new ImageLoadProperty(
        fileOrDir, 
        null, 
        new ImageRotate(), 
        new Dimension(Config.get().getMainWindowThumbnailSizeX(), Config.get().getMainWindowThumbnailSizeY()), 
        null);
    ret.isThumbnail = true;
    ret.forceAddToThumbnailCache = forceAddToThumbnailCache;
    return ret;
  }
  
  /** 読み込み対象のファイル or ディレクトリ */
  public FileInfo getFileOrDir() {
    return fileOrDir;
  }

  /** 読み込み対象のファイル (画像表示画面で2頁表示する場合) */
  public FileInfo getFile2() {
    return file2;
  }
  
  /** 回転 */
  public ImageRotate getRotate() {
    return rotate;
  }

  /** ロード結果として欲しい画像サイズ */
  public Dimension getTargetSize() {
    return targetSize;
  }

  /** ズーム */
  public Double getZoomRatio() {
    return zoomRatio;
  }

  /** サムネイル用画像か */
  public boolean isThumbnail() {
    return isThumbnail;
  }

  /** サムネイルの一括登録用か */
  public boolean isForceAddToThumbnailCache() {
    return forceAddToThumbnailCache;
  }

  /**
   * ファイルと回転が同じかどうか
   */
  public boolean equalsFileAndRotate(FileInfo fileOrDir, FileInfo file2, ImageRotate rotate) {
    return this.compareToFile(fileOrDir, file2) == 0 && this.rotate.equals(rotate);
  }
  
  private int compareToFile(FileInfo fileOrDir, FileInfo file2) {
    int x;
    
    // ファイル名で比較
    x = this.fileOrDir.compareTo(fileOrDir);
    if (x != 0) return x;
    
    if (this.file2 == null || file2 == null) {
      if (this.file2 == null && file2 == null) return 0;
      if (file2 != null) return -1;
      if (this.file2 != null) return 1;
      assert false;
    }
    x = this.file2.compareTo(file2);
    if (x != 0) return x;
    
    return 0;
  }
  
  /** オーバーライド */
  @Override
  public int compareTo(ImageLoadProperty other) {
    int x;
    // ファイル名で比較
    x = compareToFile(other.fileOrDir, other.file2);
    if (x != 0) return x;
    
    // 回転で比較
    x = Integer.valueOf(rotate.getRotate()).compareTo(other.rotate.getRotate());
    if (x != 0) return x;
    
    // サイズで比較
    if (this.targetSize != null && other.targetSize != null) {
      // 両方存在するなら、比較する
      x = Integer.valueOf(this.targetSize.width).compareTo(other.targetSize.width);
      if (x != 0) return x;
      x = Integer.valueOf(this.targetSize.height).compareTo(other.targetSize.height);
      if (x != 0) return x;
    }
    else if (this.targetSize != null || other.targetSize != null) {
      // 片方が存在するなら違いあり
      return this.targetSize == null ? -1 : 1;
    }

    // ズームで比較
    if (this.zoomRatio != null && other.zoomRatio != null) {
      // 両方存在するなら、比較する
      x = this.zoomRatio.compareTo(other.zoomRatio);
      if (x != 0) return x;
    }
    else if (this.zoomRatio != null || other.zoomRatio != null) {
      // 片方が存在するなら違いあり
      return this.zoomRatio == null ? -1 : 1;
    }
    
    return 0;
  }

  /** オーバーライド */
  @Override
  public boolean equals(Object obj) {
    if (obj == null) return false;
    return this.compareTo((ImageLoadProperty)obj) == 0;
  }
  
  /** オーバーライド */
  @Override
  public int hashCode() {
    return fileOrDir.hashCode();
  }

  /** オーバーライド */
  @Override
  public String toString() {
    return "F=" + fileOrDir + (file2 == null ? "" : ",+1") + "|R=" + rotate.getRotate() + 
        "|SZ=" + (targetSize == null ? "null" : ("" + targetSize.width + "x" + targetSize.height)) + 
        "|Z=" + zoomRatio;
  }
}
