package cvj2.app.file.thumbnail;

import java.awt.Dimension;
import java.io.IOException;

import org.apache.commons.lang3.ObjectUtils;
import org.eclipse.swt.graphics.ImageData;

import cvj2.app.file.thumbnail.fs.FileCacheManager;
import cvj2.file.fs.FileInfo;
import cvj2.util.Log;

/**
 * サムネイルのキャッシュを永続的に保存するためのクラス
 */
public class ThumbnailCache {
  private static ThumbnailCache instance = new ThumbnailCache();
  
  private FileCacheManager loader = FileCacheManager.instance();
  
  /** キャッシュ情報 (サムネイルキャッシュを特定するための情報) */
  static public class Info {
    /** ファイル or ディレクトリ */
    public String absolutePath;
    /** サムネイル枠サイズ */
    public Dimension targetSize;

    @Override public int hashCode() {
      return ObjectUtils.hashCodeMulti(absolutePath, targetSize);
    }
    @Override public boolean equals(Object obj) {
      if (!(obj instanceof Info)) return false;
      Info other = (Info) obj;
      return this.absolutePath.equals(other.absolutePath) && this.targetSize.equals(other.targetSize);
    }
    @Override public String toString() {
      return absolutePath + " | " + targetSize;
    }
  }

  /** 内部で使用するのキャッシュ要素 (サムネイルキャッシュとして格納されている情報) */
  static public class Cache extends Info {
    /** 画像 */
    public ImageData img;
    /** ファイル or ディレクトリのサイズ */
    public long length;
    /** 最終更新日時 */
    public long lastModified;
  }
  
  private ThumbnailCache() { }
  
  /**
   * インスタンス取得
   */
  public static ThumbnailCache instance() {
    return instance;
  }
  
  /**
   * 初期化
   * @param writable 最初のアプリかどうか 
   *        最初のアプリなら、キャッシュを読み書き可能になる
   *        そうでなければ、読むだけ 
   */
  public void initialize(boolean writable) throws IOException {
    Log.start();
    loader.initialize(writable);
  }

  /**
   * 終了
   */
  public void shutdown() throws IOException {
    Log.start();
    loader.shutdown();
  }
  
  /**
   * 一覧取得
   */
  public Info[] getAll() {
    return loader.getAll();
  }

  /**
   * キャッシュを削除する
   */
  public void remove(Info info) throws IOException {
    loader.remove(info);
  }
  
  /**
   * キャッシュに保存する
   */
  public void put(FileInfo fileOrDir, Dimension targetSize, ImageData image) {
    try {
      loader.put(fileOrDir, targetSize, image);
    }
    catch (IOException e) {
      throw new RuntimeException(e);
    }
  }
  
  /**
   * 画像取得
   * @param fileOrDir サムネイルを表示する ファイル、ディレクトリ
   * @return 該当する画像があればそれをかえす
   */
  public ImageData find(FileInfo fileOrDir, Dimension size) {
    try {
      return loader.find(fileOrDir, size);
    }
    catch (IOException e) {
      throw new RuntimeException(e);
    }
  }
  
  /**
   * 存在チェック
   */
  public boolean exists(FileInfo fileOrDir, Dimension size) {
    return loader.exists(fileOrDir, size);
  }
}
