package cvj2.app.file.thumbnail.fs;

import java.awt.Dimension;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Iterator;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.ObjectUtils;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;

import cvj2.app.file.thumbnail.ThumbnailCache.Cache;
import cvj2.app.file.thumbnail.ThumbnailCache.Info;
import cvj2.file.fs.FileInfo;
import cvj2.util.Log;

/**
 * ファイルにサムネイルキャッシュを保存するクラス。<br>
 * <pre> 
 * ファイル名
 * ${アプリの設定ディレクトリ}/cache_file/XX/XXXXXXXX.bmp
 * (XXXXXXXX は、フルパス名とサイズのハッシュ)
 * 
 * ファイルフォーマット
 * [生の画像ファイル長さバイト] 画像のデータ
 * ----- ここからフッター -----
 * [4バイト]                    ファイル名文字列の長さ
 * [ファイル名の長さバイト]     フルパスの String.getBytes()
 * [8バイト]                    ファイル or ディレクトリのサイズ
 * [8バイト]                    ファイル or ディレクトリの最終更新日時
 * [4バイト]                    画像幅
 * [4バイト]                    画像高
 * [4バイト]                    フッターの開始位置
 * </pre>
 */
public class FileCacheLoader {
  static private final int DIR_MASK = 0xFF;
  static private final String DIR_FORMAT_STRING = "%02X";

  static private final int IMG_TYPE = SWT.IMAGE_PNG;
  static private final String IMG_EXT = ".png";

  
  static private final int MIN_FOOTER_SIZE = 4 + 8 + 8 + 4 + 4 + 4;

  private FileCacheLoader() {}
  
  /**
   * ファイル読み込み
   */
  protected static Cache load(FileInfo fileOrDir, Dimension size, boolean isFirstApp) throws IOException {
    Log.ut_start(fileOrDir, size);
    
    // ファイルの存在チェック
    File cacheFile = getCacheFile(fileOrDir.getAbsolutePath(), size);
    if (!cacheFile.exists()) return null;
    String exComment = fileOrDir + "," + cacheFile.getName();
    
    // 読み込み
    byte[] buff;
    try {
      buff = FileUtils.readFileToByteArray(cacheFile);
    }
    catch (IOException ex) {
      // 最初のアプリでないなら、存在チェックしてから読むまでのあいだに消されるかもしれないので、エラーにしない
      if (!isFirstApp) return null;
      throw ex;
    }
    if (buff.length < MIN_FOOTER_SIZE) throw new IOException(exComment + ": キャッシュファイルが壊れています");
    Cache cache = new Cache();
    
    // フッターを読む
    int footerIndex = readInt(buff, buff.length - 4);
    ByteArrayInputStream bin = new ByteArrayInputStream(buff, footerIndex, buff.length - footerIndex);
    assert bin.available() > MIN_FOOTER_SIZE;
    DataInputStream in = new DataInputStream(bin);
    
    // ファイル名
    int nameLen = in.readInt();
    if (nameLen > 64 * 1024) throw new RuntimeException("too long filename: " + nameLen + ", " + exComment);
    byte[] nameBytes = new byte[nameLen];
    nameLen = in.read(nameBytes);
    if (nameLen != nameBytes.length) throw new RuntimeException("broken filename: " + nameLen + ", " + nameBytes.length + ", " + exComment);
    cache.absolutePath = new String(nameBytes);
    // ファイルサイズ
    cache.length = in.readLong();
    // 日時
    cache.lastModified = in.readLong();
    // 画像サイズ
    cache.targetSize = new Dimension(in.readInt(), in.readInt());
    // フッター位置
    int dummy = in.readInt();
    assert dummy == footerIndex;
    assert bin.available() == 0;
    
    // チェック
    if (!fileOrDir.getAbsolutePath().equals(cache.absolutePath)) return null;
    if (fileOrDir.length() != cache.length) return null;
    if (fileOrDir.lastModified() != cache.lastModified) return null;
    if (!size.equals(cache.targetSize)) return null;
    
    // 画像
    cache.img = new ImageData(new ByteArrayInputStream(buff));
    return cache;
  }

  /*
   * int を読む
   */
  private static int readInt(byte[] buff, int start) throws IOException {
    DataInputStream in = new DataInputStream(new ByteArrayInputStream(buff, start, 4));
    return in.readInt();
  }
  
  /**
   * ファイル書き込み
   */
  protected static void save(FileInfo fileOrDir, Cache cache) throws IOException {
    Log.ut_start(fileOrDir, cache.targetSize);
    
    // 画像書き込み
    ByteArrayOutputStream bout = new ByteArrayOutputStream();
    ImageLoader writer = new ImageLoader();
    writer.data = new ImageData[] { cache.img };
    writer.save(bout, IMG_TYPE);
    
    // フッター書き込み
    DataOutputStream out = new DataOutputStream(bout);
    int footerIndex = bout.size();
    byte[] filename = cache.absolutePath.getBytes();
    out.writeInt(filename.length);
    out.write(filename);
    out.writeLong(cache.length);
    out.writeLong(cache.lastModified);
    out.writeInt(cache.targetSize.width);
    out.writeInt(cache.targetSize.height);
    out.writeInt(footerIndex);
    
    // ファイルに書き込み
    File file = getCacheFile(fileOrDir.getAbsolutePath(), cache.targetSize);
    File dir = file.getParentFile();
    if (!dir.isDirectory() && !dir.mkdirs()) throw new IOException("mkdir failed: " + file);
    File tmpFile = File.createTempFile("_tmp_", IMG_EXT, file.getParentFile());
    FileUtils.writeByteArrayToFile(tmpFile, bout.toByteArray());
    file.delete();
    if (!tmpFile.renameTo(file)) throw new IOException("rename failed: " + tmpFile + "=>" + file);
  }
  
  /**
   * ファイル削除
   */
  protected static void remove(File fileOrDir, Dimension size) throws IOException {
    Log.ut_start(fileOrDir, size);
    
    File f = getCacheFile(fileOrDir.getAbsolutePath(), size);
    String exComment = fileOrDir + ", " + size + ", " + f;
    if (!f.exists()) return; // 別のファイル同士が同じハッシュになったときは存在しないこともある
    if (!f.delete()) throw new IOException("file can not remove: " + exComment);
  }
  
  /**
   * ファイル存在チェック
   */
  protected static boolean isExist(Info info) {
    File f = getCacheFile(info.absolutePath, info.targetSize);
    return f.exists();
  }
  
  /**
   * 一覧に存在しないファイルをゴミ掃除
   * @param enableInfos 一覧
   * @param removeLimitTime これより新しいファイルは初期化後に作られたので、削除しない
   */
  protected static void clean(Info[] enableInfos, long removeLimitTime) throws IOException {
    Log.start();
    
    // 一覧に存在する全ファイル
    HashSet<File> enableFiles = new HashSet<File>();
    for (Info info : enableInfos) {
      enableFiles.add(getCacheFile(info.absolutePath, info.targetSize));
    }
    // 全ディレクトリを掃除する
    int max = Integer.MAX_VALUE & DIR_MASK;
    for (int i = 0; i < max; i++) {
      File dir = new File(FileCacheManager.CACHE_DIR, String.format(DIR_FORMAT_STRING, i));
      clean(enableFiles, dir, removeLimitTime);
    }
  }
  private static void clean(HashSet<File> enableFiles, File dir, long removeLimitTime) throws IOException {
    if (!dir.exists()) return;
    Iterator<File> ite = FileUtils.iterateFiles(dir, null, true);
    while (ite.hasNext()) {
      File f = ite.next();
      if (enableFiles.contains(f)) continue;
      if (f.lastModified() > removeLimitTime) continue;
      Log.d("delete thumbnail cache file: ", f);
      if (!f.delete()) throw new IOException("delete failed: " + f);
    }
  }
  
  /*
   * キャッシュファイル名を取得
   */
  private static File getCacheFile(String absolutePath, Dimension size) {
    int hash = ObjectUtils.hashCodeMulti(absolutePath, size);
    File dir = new File(FileCacheManager.CACHE_DIR, String.format(DIR_FORMAT_STRING, hash & DIR_MASK));
    return new File(dir, String.format("%08X", hash) + IMG_EXT);
  }
}
