package cvj2.app.file.thumbnail.fs;

import java.awt.Dimension;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.util.ArrayList;

import org.apache.commons.lang3.StringUtils;
import org.eclipse.swt.graphics.ImageData;

import cvj2.app.Config;
import cvj2.app.ParseArg;
import cvj2.app.file.thumbnail.ThumbnailCache.Cache;
import cvj2.app.file.thumbnail.ThumbnailCache.Info;
import cvj2.file.fs.FileInfo;
import cvj2.util.CacheMap;
import cvj2.util.Log;

/**
 * ファイルシステム内でサムネイルキャッシュを管理するクラス。<br>
 * CacheMapへのアクセスで同期されるので、このクラスはスレッドセーフです
 */
public class FileCacheManager {
  static final File CACHE_DIR = new File(ParseArg.configDir, "cache_file");
  static final File INFOTBL_FILE = new File(CACHE_DIR, "list.txt");
  
  static private FileCacheManager instance = new FileCacheManager();

  // 遅延初期化スレッド
  // 遅延初期化実行中は、!null になる
  // null にするのは、FileCacheBgInitializer の中から行われる
  volatile FileCacheBgInitializer bgInitializer;
  
  // サムネイル情報のキャッシュ (Key のみ使用して、Value は無視する)
  // 先に起動しているアプリがある場合は、この変数が null になる
  // 削除されたときだけコールバックが呼ばれ、更新された時は呼ばれない
  private CacheMap<Info, Object> infoTbl = new CacheMap<Info, Object>(Config.get().getThumnailFileCacheCount(), new RemoveNotify(), false);
  private final Object DUMMY_CACHE_VAL = new String("_DUMMY_CACHE_VAL_");
  private class RemoveNotify implements CacheMap.RemoveCallback<Info, Object> {
    @Override
    public void removed(Info key, Object val) {
      Log.d(key);
      assert val == DUMMY_CACHE_VAL;
      try {
        FileCacheLoader.remove(new File(key.absolutePath), key.targetSize);
      }
      catch (IOException e) {
        // 遅延初期化実行中にファイルが存在しない情報の削除が行われるので、無視する
        if (bgInitializer != null) return;
        
        throw new RuntimeException(e);
      }
    }
  }
  
  private FileCacheManager() {}

  /**
   * インスタンス取得
   */
  public static FileCacheManager instance() {
    return instance;
  }
  
  /**
   * 初期化
   */
  public void initialize(boolean writable) throws IOException {
    Log.start();
    
    if (!writable) {
      infoTbl = null;
      return;
    }
    
    try {
      // 一覧読み込み
      readInfoTbl();

      // 遅延初期化
      bgInitializer = new FileCacheBgInitializer(getAll());
      bgInitializer.start();
    }
    catch (IOException e) {
      Log.e(e);
      String msg = e.getMessage() + "\n" + 
            "サムネイルキャッシュの初期化中に深刻なエラーが発生しました。\n" +
            "うまく動かない場合は、アプリケーションを終了させてから\n" + 
            CACHE_DIR + " ディレクトリを削除してみてください。";
      throw new IOException(msg);
    }
  }

  /**
   * 終了
   */
  public void shutdown() throws IOException {
    Log.start();
    try {
      if (infoTbl != null) writeInfoTbl();
    }
    catch (IOException e) {
      Log.e(e);
      String msg = e.getMessage() + "\n" + 
            "サムネイルキャッシュの書き込み中に深刻なエラーが発生しました。\n" +
            "再起動してもうまく動かない場合は、アプリケーションを終了させてから\n" + 
            CACHE_DIR + " ディレクトリを削除してみてください。";
      throw new IOException(msg);
    }
  }

  /*
   * 一覧読み込み
   */
  private void readInfoTbl() throws IOException {
    Log.start();

    infoTbl.clear();
    if (!INFOTBL_FILE.exists()) return;
    LineNumberReader in = new LineNumberReader(new FileReader(INFOTBL_FILE));
    try {
      while (true) {
        // 1行読み込み
        String line = in.readLine();
        if (line == null) break;
        
        // 解析
        String[] spl = StringUtils.split(line, ",", 3);
        if (spl.length != 3) throw new IOException("thumbnail infotbl format error: [" + in.getLineNumber() + "]: " + line);
        Info info = new Info();
        info.targetSize = new Dimension(Integer.parseInt(spl[0]), Integer.parseInt(spl[1]));
        info.absolutePath = spl[2];
        
        // 一覧に追加 (実ファイルが存在しない時は、BgInitializer で削除される)
        infoTbl.put(info, DUMMY_CACHE_VAL);
      }
    }
    finally {
      in.close();
    }
  }

  /*
   * 一覧書き込み
   */
  private void writeInfoTbl() throws IOException {
    Log.start();

    CACHE_DIR.mkdirs(); 
    if (!CACHE_DIR.isDirectory()) throw new IOException(CACHE_DIR + " ディレクトリを作成できません");
    
    File tmpFile = File.createTempFile("_infotbl_", ".tmp", INFOTBL_FILE.getParentFile());
    PrintWriter out = new PrintWriter(new BufferedOutputStream(new FileOutputStream(tmpFile)));
    try {
      for (Info info : getAll()) {
        out.println("" + info.targetSize.width + "," + info.targetSize.height + "," + info.absolutePath);
      }
    }
    finally {
      out.close();
    }
    INFOTBL_FILE.delete();
    if (!tmpFile.renameTo(INFOTBL_FILE)) throw new IOException("rename failed: " + tmpFile + "=>" + INFOTBL_FILE);
  }

  /**
   * キャッシュ制限設定
   */
  public void setCacheLimit(int count) {
    if (infoTbl == null) return;
    infoTbl.setLimit(count);
  }
  
  /**
   * 一覧取得
   */
  public Info[] getAll() {
    ArrayList<Info> list = new ArrayList<Info>();
    infoTbl.getAll(list, null);
    return list.toArray(new Info[0]);
  }
  
  /**
   * キャッシュに保存する
   */
  public void put(FileInfo fileOrDir, Dimension targetSize, ImageData image) throws IOException {
    Log.ut_start(fileOrDir, image.width, image.height);

    if (infoTbl == null) return;
    
    // データの準備
    Cache cache = new Cache();
    cache.absolutePath = fileOrDir.getAbsolutePath();
    cache.targetSize = targetSize;
    cache.img = image;
    cache.length = fileOrDir.length();
    cache.lastModified = fileOrDir.lastModified();
    
    // ファイルに書き込み
    FileCacheLoader.save(fileOrDir, cache);
    
    // リストを更新
    putInfoTbl(cache);
  }
  
  /**
   * 画像取得
   * @param fileOrDir サムネイルを表示する ファイル、ディレクトリ
   * @return 該当する画像があればそれをかえす
   */
  public ImageData find(FileInfo fileOrDir, Dimension size) throws IOException {
    Log.ut_start(fileOrDir, size);
    
    // ファイルを読む
    Cache cache = FileCacheLoader.load(fileOrDir, size, infoTbl != null);
    if (cache == null) return null;
    
    // 画像サイズチェック
    if (!cache.targetSize.equals(size)) return null;
    // 日付チェック
    if (cache.lastModified != fileOrDir.lastModified()) return null;
    // ファイルサイズチェック
    if (cache.length != fileOrDir.length()) return null;

    // リストを更新
    if (infoTbl != null) {
      putInfoTbl(cache);
    }
    
    return cache.img;
  }

  /**
   * 存在チェック
   */
  public boolean exists(FileInfo fileOrDir, Dimension size) {
    Info info = new Info();
    info.absolutePath = fileOrDir.getAbsolutePath();
    info.targetSize = size;
    return infoTbl.exists(info);
  }

  /*
   * 一覧を更新
   */
  private void putInfoTbl(Cache cache) {
    Info info = new Info();
    info.absolutePath = cache.absolutePath;
    info.targetSize = cache.targetSize;
    infoTbl.put(info, DUMMY_CACHE_VAL);
  }

  /**
   * キャッシュを削除する
   */
  public void remove(Info info) throws IOException {
    Log.ut_start(info);
    
    // ※) 一覧から削除すると、コールバックされてキャッシュファイルが削除される
    // (このメソッドは画面の削除ボタンが押されたときに呼ばれるので、タイミングの差で既に存在しない可能性がある)
    Object dummy = infoTbl.remove(info);
    assert dummy == null || dummy == DUMMY_CACHE_VAL;
  }
}
