package cvj2.app.gui.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import cvj2.file.fs.FileInfo;
import cvj2.file.fs.FileInfoList;

/**
 * FileListModel のソートされたファイルのリスト
 */
public class FileListSort extends FileInfoList {
  // カレントディレクトリ内のファイル/ディレクトリ
  private FileInfoList fileList;
  // データの列
  private FileListColumn[] columns;
  // ソート順
  private List<FileListSortElement> sortElements = new ArrayList<FileListSortElement>();
  // ソートされた行からオリジナルデータの行への変換リスト (ソートされていないときは、null)
  private SortIndex sortIndex = null;

  /**
   * コンストラクタ
   */
  FileListSort(FileInfoList fileList, FileListColumn[] columns) {
    // 全てのメソッドを再実装するので、空リストを渡す
    super(new FileInfo[0]);
    
    this.fileList = fileList;
    this.columns = columns;
  }
  
  /**
   * 別のソート条件を再設定する
   */
  public void resetSortColumns(FileListSort other) {
    this.sortElements = new ArrayList<FileListSortElement>(other.sortElements);
    if (this.sortElements.size() != 0) sortIndex = new SortIndex();
  }

  /**
   * ソート条件がファイル情報の取得を必要とするか？
   */
  public boolean isNeedFileAttribute() {
    for (FileListSortElement elm : sortElements) {
      // 名前でソートするものが先頭にあればファイル情報は不要
      if (elm.getColumn() == FileListColumn.NAME) return false;
      // 拡張子は判断保留
      if (elm.getColumn() == FileListColumn.EXT) continue;
      // ファイル名以外の情報が必要なら、true
      if (!elm.getColumn().isFilenameAttribute()) return true;
    }
    return false;
  }
  
  /**
   * サイズ取得
   */
  @Override
  public int size() {
    return fileList.size();
  }

  /**
   * ファイルを探す。<br>
   * ※) 何でソートされているかわからないため、リスト中に存在しないファイルを探すと
   *     {@link FileInfoList#find(FileInfo)} とは違い、挿入位置を返すことができない。
   */
  @Override
  public int find(FileInfo f) {
    int i = fileList.find(f);
    // ※) バックグラウンドから呼ばれているかもしれないので、状態をコピーしておく
    SortIndex idx = sortIndex;
    // ソートされていないのなら、そのままかえす
    if (idx == null) return i;
    // ソートされているのに、存在しない場合はどうしようもない
    if (i < 0) return -1;
    
    return idx.rev[i];
  }
  
  /**
   * ソートされたインデックスでファイルを取得
   */
  @Override
  public FileInfo get(int index) {
    // ※) バックグラウンドから呼ばれているかもしれないので、状態をコピーしておく
    SortIndex idx = sortIndex;
    if (idx == null) return fileList.get(index);
    return fileList.get(idx.idx[index]);
  }
  
  /**
   * 先頭のソート条件を取得
   */
  public FileListSortElement getFirst() {
    if (sortElements.size() > 0) return sortElements.get(0);
    return null;
  }

  /**
   * 列に対応するインデックスを取得
   */
  public int getColumnIndex(FileListColumn column) {
    for (int i = 0; i < columns.length; i++) {
      if (columns[i] == column) return i;
    }
    throw new RuntimeException("" + column);
  }
  
  /**
   * データ列がクリックされた
   */
  public void clickColumn(FileListColumn col) {
    FileListSortElement s = null;
    int i = 0;
    for (i = 0; i < sortElements.size(); i++) {
      if (sortElements.get(i).getColumn() != col) continue;
      s = sortElements.get(i);
      sortElements.remove(i);
      break;
    }
    if (s == null) {
      // 新しいソート列
      // 最初のソート条件がファイル名の時は、降順にする
      boolean order = !(sortElements.size() == 0 && col == FileListColumn.NAME);
      s = new FileListSortElement(col, order);
    }
    else if (i == 0) {
      // 現在のソート列がクリックされたら反転する
      s = s.createReverse();
    }
    else {
      // 2番目以降のソート列がクリックされたら、昇順にする
      s = s.createAscent();
    }

    // ソート条件を先頭に追加
    sortElements.listIterator().add(s);
    // ソートする
    sortIndex = new SortIndex();
  }

  private class SortIndex {
    Integer idx[] = null;
    int[] rev = null;
    SortIndex() {
      int len = fileList.size();
      // 初期化
      idx = new Integer[len];
      rev = new int[len];
      for (int i = 0; i < len; i++) idx[i] = i;
      // ソート
      Arrays.sort(idx, new Cmp());
      // 逆順リスト生成
      for (int i = 0; i < idx.length; i++) {
        rev[idx[i]] = i;
      }
    }
  }

  // 比較クラス
  private class Cmp implements Comparator<Integer> {
    @Override
    public int compare(Integer i1, Integer i2) {
      FileInfo f1 = fileList.get(i1);
      FileInfo f2 = fileList.get(i2);
      for (int i = 0; i < sortElements.size(); i++) {
        FileListColumn c = sortElements.get(i).getColumn();
        int cmp = c.cmpare(f1, f2) * (sortElements.get(i).isAscent() ? 1 : -1);
        if (cmp != 0) return cmp;
      }
      return f1.compareTo(f2);
    }
  }
}
