package cvj2.app.gui.data;

import java.util.Collections;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import cvj2.app.file.imgload.BackgroundImageLoader;
import cvj2.util.Log;

/**
 * アプリケーションの内部状態の監視
 */
public class SystemMonitor {
  private static final int DATA_INTERVAL_SEC = 1;
  private static final int MAX_DATA_NUM = 3600;
  private static SystemMonitor instance = new SystemMonitor();
  
  private BgThread thread;
  private List<Data> datas = Collections.synchronizedList(new LinkedList<Data>());
  
  private SystemMonitor() {}
  
  /**
   * インスタンス取得
   */
  public static SystemMonitor instance() {
    return instance;
  }
  
  /**
   * 初期化
   */
  public void initialize() {
    assert thread == null;
    thread = new BgThread();
    thread.setDaemon(true);
    thread.setName("SysMon");
    thread.start();
  }

  /**
   * 統計データ取得
   * @return 古いデータが先頭に来る
   */
  public Data[] getDatas() {
    return datas.toArray(new Data[0]);
  }
  
  /**
   * 統計データ
   */
  public static class Data {
    /** 時刻 */
    public long time;
    
    /** 仮想マシンの最大メモリー容量 (MB) */
    public double maxMemoryMB;
    /** 仮想マシンの空きメモリー容量 (MB) */
    public double freeMemoryMB;
    /** 仮想マシンの割り当て済みメモリー容量 (MB) */
    public double totalMemoryMB;
    
    /** バックグラウンドの画像読み込み待ち数 */
    public double bgWaitNum;
    /** バックグラウンドの画像読み処理中の数 */
    public double bgRunNum;
    
    /** メモリーのサムネイルキャッシュ数 */
    public double thumbnailCacheNum;
    /** メモリーの一枚画像キャッシュ数 */
    public double imageCacheNum;

    /** バックグラウンドの画像読み処理に使用している作業メモリー容量 (MB) */
    public double bgWorkMemoryMB;
    
    /** バックグラウンドで処理した画像のピクセルバイト数 (MB) */
    public double bgPixelThroughputMB;
    /** バックグラウンドで処理した画像ファイルのサイズ (MB) */
    public double bgFileThroughtputMB;
    /** バックグラウンドで処理した画像ファイルの数 */
    public double bgFileReadNum;
  }

  /**
   * データ収集スレッド
   */
  private class BgThread extends Thread {
    // *.000 秒ちょうどになっている時刻
    private long lastTime;
    // 前回の統計情報
    private long[] lastStat = new long[8];
    
    @Override
    public void run() {
      Log.start();
      try {
        waitStart();
        while (true) {
          getData();
          sleepToNext();
        }
      }
      catch (InterruptedException e) {
        Log.d("interuppted");
      }
      finally {
        Log.d("end");
      }
    }

    /* *.000 秒ちょうどのデータ収集開始を待つ  */
    private void waitStart() throws InterruptedException {
      lastTime = new Date().getTime();
      long sleep = 1000 - (lastTime % 1000);
      lastTime = lastTime / 1000 * 1000;
      Thread.sleep(sleep);
    }

    /* データ収集 */
    public void getData() {
      Data d = new Data();

      Runtime r = Runtime.getRuntime();
      long[] bgStat = BackgroundImageLoader.instance().getStat();
      
      d.time = new Date().getTime();
      
      d.maxMemoryMB = (double)r.maxMemory() / 1024 / 1024;
      d.freeMemoryMB = (double)r.freeMemory() / 1024 / 1024;
      d.totalMemoryMB = (double)r.totalMemory() / 1024 / 1024;
      
      d.bgWaitNum = (double)bgStat[0];
      d.bgRunNum = (double)bgStat[1];

      d.thumbnailCacheNum = (double)bgStat[2];
      d.imageCacheNum = (double)bgStat[3];

      d.bgWorkMemoryMB = (double)bgStat[4] / 1024 / 1024;
      
      d.bgPixelThroughputMB = (double)(bgStat[5] - lastStat[5]) / 1024 / 1024;
      d.bgFileThroughtputMB = (double)(bgStat[6] - lastStat[6]) / 1024 / 1024;
      d.bgFileReadNum = (double)(bgStat[7] - lastStat[7]);
      
      lastStat = bgStat;

      datas.add(d);
      while (datas.size() > MAX_DATA_NUM) {
        datas.remove(0);
      }
    }

    /* 次のデータ収集まで sleep */
    private void sleepToNext() throws InterruptedException {
      long now = new Date().getTime();
      long next = lastTime + DATA_INTERVAL_SEC * 2 * 1000;
      long sleep = Math.max(0, next - now);
      lastTime += DATA_INTERVAL_SEC * 1000;
      Thread.sleep(sleep);
    }
  }
}
