package cvj2.app.gui.dialogs;

import java.awt.Dimension;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;

import cvj2.app.Config;
import cvj2.app.file.imgload.BackgroundImageLoader;
import cvj2.app.file.thumbnail.ThumbnailCache;
import cvj2.file.fs.FileInfo;
import cvj2.util.Log;

/**
 * サムネイル一括登録スレッド
 */
public class AddThumbnailThread extends Thread {
  private AddThumbnailDialog2 dlg;
  private FileInfo startDir;
  private boolean checkDir;
  private boolean checkFile;
  private int maxDepth;
  private HashMap<File, FileInfo> finishedCannonicalFileMap = new HashMap<File, FileInfo>();

  volatile int dirCount = 0;
  volatile int imageFileCount = 0;
  volatile int fileCount = 0;
  volatile int oldDirThumbnailCount = 0;
  volatile int oldImageThumbnailFileCount = 0;
  volatile int addDirThumbnailCount = 0;
  volatile int addImageThumbnailFileCount = 0;
  volatile String currentFind = "";
  volatile String currentLoad = "";
  
  volatile private boolean cancel = false;

  /**
   * コンストラクタ
   * @param depth 探索深さ (0 ならstartDirの子のみ、-1 なら制限なし)
   */
  public AddThumbnailThread(AddThumbnailDialog2 dlg, FileInfo startDir, boolean checkDir, boolean checkFile, int depth) {
    Log.start(startDir, checkDir, checkFile);
    
    this.dlg = dlg;
    this.startDir = startDir;
    this.checkDir = checkDir;
    this.checkFile = checkFile;
    this.maxDepth = depth;
    
    setDaemon(true);
    setName("AddThumbnail");
  }

  /**
   * キャンセルする
   */
  public void cancel() {
    this.cancel = true;
    this.interrupt();
  }
  
  /*
   * 画面を閉じる
   */
  private void finish() {
    if (dlg.getShell() == null || dlg.getShell().getDisplay() == null) return;
    dlg.getShell().getDisplay().asyncExec(new Runnable() {
      @Override public void run() {
        if (dlg.getShell() == null || dlg.getShell().isDisposed()) return;
        dlg.close();
      }
    });
  }

  /**
   * 処理実行
   */
  @Override
  public void run() {
    Log.start();

    try {
      // 処理開始
      exec(startDir, maxDepth >= 0 ? maxDepth + 1 : Integer.MAX_VALUE);
    }
    catch (IOException e) {
      throw new RuntimeException(e);
    }
    finally {
      finish();
    }
  }

  /*           ┌── (再帰)───┐
   *           ↓                 │
   * run ─→ exec ─┬→ execDir ┴──┐
   *                 │                 ↓
   *                 └→ execFile ─→ load  
   */

  /*
   * 処理する
   */
  private boolean exec(FileInfo f, int depth) throws IOException {
    if (cancel) return false;
    if (depth < 0) return true;
    
    currentFind = f.getAbsolutePath();
    
    if (f.isDirectory()) {
      if (!execDir(f, depth)) return false;
    }
    else {
      if (!execFile(f)) return false;
    }
    
    return true;
  }

  /*
   * ディレクトリの処理
   */
  private boolean execDir(FileInfo dir, int depth) throws IOException {
    dirCount++;

    // サムネイルロード
    if (!load(dir)) return false;
    
    // シンボリックリンクの自己再帰を防止する
    if (!checkSelfSymLink(dir)) return true;

    for (FileInfo f : dir.getList()) {
      if (!exec(f, depth - 1)) return false;
    }
    return true;
  }

  /*
   * シンボリックリンクの自己再帰かどうか
   */
  private boolean checkSelfSymLink(FileInfo dir) throws IOException {
    // 処理対象ディレクトリの正規名称
    if (dir == FileInfo.ROOT) return false;
    File cf = dir.getFile().getCanonicalFile();
    
    // 正規名称が登録されているかどうか
    FileInfo checkParent = finishedCannonicalFileMap.get(cf);
    if (checkParent == null) {
      finishedCannonicalFileMap.put(cf, dir);
      return true;
    }
    
    while (dir != FileInfo.ROOT) {
      // 処理対象の親が同じ正規名称を登録していたら、処理しない
      if (dir.equals(checkParent)) return false;
      dir = dir.getParent();
    }
    return true;
  }

  /*
   * ファイルのの処理
   */
  private boolean execFile(FileInfo f) {
    if (f.isImageFile()) {
      imageFileCount++;

      // サムネイルロード
      if (!load(f)) return false;
    }
    else {
      fileCount++;      
    }
    
    return true;
  }

  /*
   * サムネイルロード
   */
  private boolean load(final FileInfo fileOrDir) {
    // 存在チェック
    final Dimension thumbnailSize = new Dimension(Config.get().getMainWindowThumbnailSizeX(), Config.get().getMainWindowThumbnailSizeY());
    if (ThumbnailCache.instance().exists(fileOrDir, thumbnailSize)) {
      if (fileOrDir.isDirectory()) {
        oldDirThumbnailCount++;
      }
      else {
        oldImageThumbnailFileCount++;
      }
      return true;
    }

    // ロードするかどうかの判定
    if (fileOrDir.isDirectory()) {
      if (!checkDir) return true;
    }
    else {
      if (!checkFile) return true;
    }
    
    currentLoad = fileOrDir.getAbsolutePath();

    // 読み込み待ちがたくさんあれば、待つ
    while (true) {
      if (cancel) return false;
      if (BackgroundImageLoader.instance().getWaitNum() < Config.get().getImageLoadBackgroundThreadNum()) break;
      try {
        Thread.sleep(100);
      }
      catch (InterruptedException e) {
        return false;
      }
    }

    // 読み込む
    BackgroundImageLoader.Notify notify = new BackgroundImageLoader.Notify() {
      @Override
      public void stateChanged(FileInfo f) {
        if (!ThumbnailCache.instance().exists(fileOrDir, thumbnailSize)) return;
        
        if (fileOrDir.isDirectory()) {
          addDirThumbnailCount++;
        }
        else {
          addImageThumbnailFileCount++;
        }
      }
    };
    BackgroundImageLoader.instance().getThumbnail(fileOrDir, notify, true);
    
    return true;
  }
}
