package cvj2.app.gui.dialogs;

import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.apache.commons.io.FileUtils;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowData;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.program.Program;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;

import cvj2.app.Config;
import cvj2.app.gui.util.MsgBox;
import cvj2.file.fs.FileInfo;
import cvj2.file.util.FileUtil;
import cvj2.gui.swt.util.SWTUtil;
import cvj2.util.Log;

/**
 * ファイルプロパティダイアログ
 */
public class FilePropertyDialog extends Dialog {
  private Label lblIcon;
  private Label lblName;
  private Label lblType;
  private Label lblDir;
  private Label lblSize;
  private Label lblDate;

  private boolean initialized = false;
  private FileInfo file;
  private Image icon;
  private Label lblNewLabel_4;
  private Label lblNewLabel_5;
  private Label lblNewLabel_6;
  private Button btnCheckRead;
  private Button btnCheckWrite;
  private Button btnCheckExecute;
  private Composite composite;
  private FilePropertyDialogThread bgThread;
  
  /**
   * コンストラクタ
   */
  public FilePropertyDialog(Shell parentShell, FileInfo file) {
    super(parentShell);
    setShellStyle(SWT.RESIZE | SWT.TITLE | SWT.PRIMARY_MODAL);
    
    assert file != FileInfo.ROOT;
    this.file = file;
  }

  /*
   * 初期化
   */
  private void initializeDialog(ControlEvent e) {
    if (initialized) return;
    initialized = true;
    Log.start(e);

    bgThread = new FilePropertyDialogThread(this, file.getFile());
    bgThread.start();
    
    getShell().setText(file.getName() + " のプロパティ");
    
    // ラベル設定
    getLblType().setText(file.isDirectory() && !file.isArchiveRoot() ? "ディレクトリ" : "ファイル");
    getLblName().setText(SWTUtil.escapeControlText(file.getName()));
    getLblDir().setText(SWTUtil.escapeControlText(file.getParent().getAbsolutePath()));
    updateSize(file.length(), 0, -1, -1, true);
    getLblDate().setText(SimpleDateFormat.getDateInstance().format(new Date(file.lastModified())) + " " + SimpleDateFormat.getTimeInstance().format(new Date(file.lastModified())));
    
    // チェックボタン設定
    getBtnCheckRead().setSelection(file.getFile().canRead());
    getBtnCheckWrite().setSelection(file.getFile().canWrite());
    getBtnCheckExecute().setSelection(file.getFile().canExecute());
    
    // アイコン表示
    String ext = FileUtil.splitBasenameExt(file.getName())[1];
    Program p = Program.findProgram(ext);
    if (p != null) {
      ImageData id = p.getImageData();
      if (id != null) {
        icon = new Image(getShell().getDisplay(), id);
        getLblIcon().setImage(icon);
      }
      getLblType().setText(p.getName());
    }
    
    if (Config.get().isPropertyDialogHideReadableCheckBox()) {
      getBtnCheckRead().setVisible(false);
      RowData rd_btnCheckRead = new RowData(SWT.DEFAULT, SWT.DEFAULT);
      rd_btnCheckRead.exclude = true;
      getBtnCheckRead().setLayoutData(rd_btnCheckRead);
    }
    if (Config.get().isPropertyDialogHideExecuteCheckBox()) {
      getBtnCheckExecute().setVisible(false);
    }
    getButton(IDialogConstants.CANCEL_ID).setFocus();
    getShell().layout();
  }

  /*
   * サイズ表示更新
   */
  void updateSize(long size, int failCount, int dirCount, int fileCount, boolean finish) {
    if (getShell() == null || getShell().isDisposed()) return;
    
    String s = FileUtils.byteCountToDisplaySize(size) + " (" + DecimalFormat.getNumberInstance().format(size) + " バイト)";
    if (size < 0) s = "サイズ不明";
    if (!finish) s += " …… ";
    if (failCount > 0) s += " (不明なディレクトリ " + failCount + ")";
    if (dirCount >= 0) s += " ディレクトリ: " + dirCount;
    if (fileCount >= 0) s += " ファイル: " + fileCount;
    getLblSize().setText(s);
  }
  
  /*
   * 破棄
   */
  private void disposeDialog(DisposeEvent e) {
    Log.start(e);
    
    if (bgThread != null) bgThread.cancel();
    if (icon != null) icon.dispose();
  }
  
  @Override
  protected void buttonPressed(int buttonId) {
    if (buttonId == IDialogConstants.OK_ID) {
      // 属性変更
      try {
        File f = file.getFile();
        if (f.canRead() != getBtnCheckRead().getSelection()) {
          if (!f.setReadable(getBtnCheckRead().getSelection())) throw new IOException("読み取り属性を変更できません");
        }
        if (f.canWrite() != getBtnCheckWrite().getSelection()) {
          if (!f.setWritable(getBtnCheckWrite().getSelection())) throw new IOException("書き込み属性を変更できません");
        }
        if (f.canExecute() != getBtnCheckExecute().getSelection()) {
          if (!f.setExecutable(getBtnCheckExecute().getSelection())) throw new IOException("実行属性を変更できません");
        }
      }
      catch (IOException e) {
        MsgBox.error("属性変更失敗", e.getMessage());
        return;
      }
    }
    
    super.buttonPressed(buttonId);
  }

  /**
   * Create contents of the dialog.
   * @param parent
   */
  @Override
  protected Control createDialogArea(Composite parent) {
    parent.addDisposeListener(new DisposeListener() {
      public void widgetDisposed(DisposeEvent e) {
        disposeDialog(e);
      }
    });
    parent.addControlListener(new ControlAdapter() {
      @Override
      public void controlResized(ControlEvent e) {
        initializeDialog(e);
      }
    });
    Composite container = (Composite) super.createDialogArea(parent);
    GridLayout gridLayout = (GridLayout) container.getLayout();
    gridLayout.numColumns = 3;
    
    lblIcon = new Label(container, SWT.NONE);
    lblIcon.setText(" ");
    
    lblNewLabel_4 = new Label(container, SWT.NONE);
    lblNewLabel_4.setText("  ");
    
    lblName = new Label(container, SWT.NONE);
    lblName.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
    lblName.setText("name");
    
    Label lblNewLabel = new Label(container, SWT.NONE);
    lblNewLabel.setText("種類");
    new Label(container, SWT.NONE);
    
    lblType = new Label(container, SWT.NONE);
    lblType.setLayoutData(new GridData(SWT.LEFT, SWT.CENTER, true, false, 1, 1));
    lblType.setText(" ");
    
    Label lblNewLabel_1 = new Label(container, SWT.NONE);
    lblNewLabel_1.setText("場所");
    new Label(container, SWT.NONE);
    
    lblDir = new Label(container, SWT.NONE);
    lblDir.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
    lblDir.setText("dir");
    
    Label lblNewLabel_2 = new Label(container, SWT.NONE);
    lblNewLabel_2.setText("サイズ");
    new Label(container, SWT.NONE);
    
    lblSize = new Label(container, SWT.NONE);
    lblSize.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
    lblSize.setText("size");
    
    Label lblNewLabel_3 = new Label(container, SWT.NONE);
    lblNewLabel_3.setText("日時");
    new Label(container, SWT.NONE);
    
    lblDate = new Label(container, SWT.NONE);
    lblDate.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
    lblDate.setText("date");
    
    lblNewLabel_5 = new Label(container, SWT.SEPARATOR | SWT.HORIZONTAL);
    lblNewLabel_5.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 3, 1));
    lblNewLabel_5.setText("New Label");
    
    lblNewLabel_6 = new Label(container, SWT.NONE);
    lblNewLabel_6.setText("属性");
    new Label(container, SWT.NONE);
    
    composite = new Composite(container, SWT.NONE);
    composite.setLayout(new RowLayout(SWT.HORIZONTAL));
    
    btnCheckRead = new Button(composite, SWT.CHECK);
    btnCheckRead.setText("読み取り可(&R)");
    
    btnCheckWrite = new Button(composite, SWT.CHECK);
    btnCheckWrite.setText("書き込み可(&W)");
    
    btnCheckExecute = new Button(composite, SWT.CHECK);
    btnCheckExecute.setText("実行可(&X)");

    return container;
  }

  /**
   * Create contents of the button bar.
   * @param parent
   */
  @Override
  protected void createButtonsForButtonBar(Composite parent) {
    createButton(parent, IDialogConstants.OK_ID, IDialogConstants.OK_LABEL, true);
    createButton(parent, IDialogConstants.CANCEL_ID, IDialogConstants.CANCEL_LABEL, false);
  }

  /**
   * Return the initial size of the dialog.
   */
  @Override
  protected Point getInitialSize() {
    return new Point(465, 384);
  }

  private Label getLblIcon() {
    return lblIcon;
  }
  private Label getLblName() {
    return lblName;
  }
  private Label getLblType() {
    return lblType;
  }
  private Label getLblDir() {
    return lblDir;
  }
  private Label getLblSize() {
    return lblSize;
  }
  private Label getLblDate() {
    return lblDate;
  }
  private Button getBtnCheckRead() {
    return btnCheckRead;
  }
  private Button getBtnCheckWrite() {
    return btnCheckWrite;
  }
  private Button getBtnCheckExecute() {
    return btnCheckExecute;
  }
}
