package cvj2.app.gui.dialogs;

import org.apache.commons.lang3.SystemUtils;
import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CLabel;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.wb.swt.SWTResourceManager;

import cvj2.app.Main;
import cvj2.gui.swt.util.SWTUtil;
import cvj2.util.Log;

/**
 * メッセージボックスを表示するダイアログ
 */
public class MsgBoxDialog extends Dialog {
  private final int BUTTON_ID_SHIFT = 100;
  
  private Composite container;
  private CLabel label_icon;
  private CLabel text_message;

  private boolean initialized = false;
  private String[] buttonNames;
  private char[] shortcutChars;
  private int defaultButtonIndex;
  private String title;
  private int icon;
  private String dispMessage;
  private String innerMessage;
  private int returnCode;
  
  /**
   * コンストラクタ
   * 
   * @param icon {@link Display#getSystemImage(int)} の引数
   * 
   * @see SWT#ICON_ERROR
   * @see SWT#ICON_INFORMATION
   * @see SWT#ICON_QUESTION
   * @see SWT#ICON_WARNING
   * @see SWT#ICON_WORKING
   */
  public MsgBoxDialog(Shell parentShell, String title, int icon, String[] buttonNames, char[] shotcutChars, int defaultButtonIndex) {
    super(parentShell);
    setShellStyle(SWT.BORDER | SWT.RESIZE | SWT.TITLE | SWT.PRIMARY_MODAL);

    assert shortcutChars == null || buttonNames.length == shotcutChars.length;
    
    this.title = title;
    this.icon = icon;
    this.buttonNames = buttonNames;
    this.shortcutChars = shotcutChars;
    this.defaultButtonIndex = defaultButtonIndex;
  }

  /**
   * ダイアログの初期化
   */
  private void initDialog(ControlEvent e) {
    if (initialized) return;
    initialized = true;
    
    // ボタン以外の方法で閉じられると、この値をかえす
    this.returnCode = -1;
    
    if (icon >= 0) {
      getLabel_icon().setImage(getShell().getDisplay().getSystemImage(icon));
    }
    getShell().setText(title);
    getText_message().setText(SWTUtil.escapeControlText(dispMessage));
    getButton(BUTTON_ID_SHIFT + defaultButtonIndex).setFocus();
    
    // ショートカットを認識するためにボタンにキーリスナーを登録
    for (int i = 0; i < buttonNames.length; i++) {
      getButton(BUTTON_ID_SHIFT + i).addKeyListener(new KeyAdapter() {
        @Override
        public void keyReleased(KeyEvent e) {
          button_keyPressed(e);
        }
      });
    }
    getContainer().layout();
    getShell().pack();
  }

  /**
   * ボタンのキー入力イベントハンドラー
   * (ショートカットキーが押されたかどうかチェックする)
   */
  private void button_keyPressed(KeyEvent e) {
    // CTRL+C が押されたら、メッセージをコピーする
    if (e.stateMask == SWT.CTRL && e.keyCode == 'c') {
      if (Main.getClipboard() != null) {
        String s = getShell().getText() + "\n\n";
        s += innerMessage != null ? innerMessage : getText_message().getText();
        Main.getClipboard().setContents(new Object[] { s }, new Transfer[] { TextTransfer.getInstance() });
        return;
      }
    }
    
    if (!SystemUtils.IS_OS_WINDOWS) {
      if (e.stateMask == 0 && e.keyCode == SWT.ARROW_RIGHT) {
        ((Control)e.widget).traverse(SWT.TRAVERSE_TAB_NEXT);
      }
      if (e.stateMask == 0 && e.keyCode == SWT.ARROW_LEFT) {
        ((Control)e.widget).traverse(SWT.TRAVERSE_TAB_PREVIOUS);
      }
    }
    
    // ショートカットキーは単独押しのみ対応
    if (e.stateMask != 0) return;
    if (shortcutChars == null) return;
    for (int i = 0; i < shortcutChars.length; i++) {
      if (shortcutChars[i] != e.keyCode) continue;
      buttonPressed(i + BUTTON_ID_SHIFT);
    }
  }

  /**
   * テキストのマウス移動イベントハンドラー
   * (マウスが動かされたら、ツールチップを有効化する)
   */
  protected void text_message_mouseMove(MouseEvent e) {
    if (getText_message().getToolTipText() != null && getText_message().getToolTipText().length() != 0) return;
    getText_message().setToolTipText(dispMessage);
  }
  
  /**
   * @deprecated 使用禁止
   */
  @Override
  public int open() {
    throw new RuntimeException();
  }

  /**
   * メッセージを表示する
   * @return 押されたボタンのインデックス
   *          -1: ボタン以外の方法でダイアログが閉じられた
   */
  public int showMessage(String dispMsg, String innerMsg) {
    this.dispMessage = dispMsg;
    this.innerMessage = innerMsg;
    super.open();
    return this.returnCode;
  }
  
  /**
   * ボタンが押された
   */
  @Override
  protected void buttonPressed(int buttonId) {
    Log.ut_start(buttonId);
    this.returnCode = buttonId - BUTTON_ID_SHIFT;
    close();
  }
  
  /**
   * Create contents of the dialog.
   * @param parent
   */
  @Override
  protected Control createDialogArea(Composite parent) {
    parent.addControlListener(new ControlAdapter() {
      @Override
      public void controlResized(ControlEvent e) {
        initDialog(e);
      }
    });
    container = (Composite) super.createDialogArea(parent);
    GridLayout gridLayout = (GridLayout) container.getLayout();
    gridLayout.numColumns = 2;
    
    label_icon = new CLabel(container, SWT.NONE);
    label_icon.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false, 1, 1));
    label_icon.setText(" ");
    
    text_message = new CLabel(container, SWT.BORDER);
    text_message.addMouseMoveListener(new MouseMoveListener() {
      public void mouseMove(MouseEvent e) {
        text_message_mouseMove(e);
      }
    });
    text_message.setBackground(SWTResourceManager.getColor(SWT.COLOR_WIDGET_BACKGROUND));
    GridData gd_text_message = new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1);
    gd_text_message.minimumHeight = 70;
    gd_text_message.minimumWidth = 300;
    text_message.setLayoutData(gd_text_message);

    return container;
  }

  /**
   * Create contents of the button bar.
   * @param parent
   */
  @Override
  protected void createButtonsForButtonBar(Composite parent) {
    for (int i = 0; i < buttonNames.length; i++) {
      createButton(parent, BUTTON_ID_SHIFT + i, buttonNames[i], false);
    }
  }

  /**
   * Return the initial size of the dialog.
   */
  @Override
  protected Point getInitialSize() {
    return new Point(500, 200);
  }
  
  private CLabel getText_message() {
    return text_message;
  }
  private Composite getContainer() {
    return container;
  }
  private CLabel getLabel_icon() {
    return label_icon;
  }
}
