package cvj2.app.gui.dialogs.preference;

import java.util.ArrayList;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.lang3.SystemUtils;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;

import cvj2.app.Config;
import cvj2.app.command.Command;
import cvj2.app.command.CommandMapInput;
import cvj2.app.command.DefaultCommandMapInputCreator;
import cvj2.gui.swt.ui.input.InputCombination;
import cvj2.util.Log;

/**
 * キーカスタマイズ設定ページ
 */
public class KeyMapPreferencePage extends CommonPreferencePage {
  private Table table;
  private TableViewer tableViewer;
  
  private TreeMap<Command, Set<InputCombination>> tableData = new TreeMap<Command, Set<InputCombination>>();
  private boolean isMainWindow;

  /* テーブルヘッダー */
  private class TableLabelProvider extends LabelProvider implements ITableLabelProvider {
    public Image getColumnImage(Object element, int columnIndex) { return null; }
    public String getColumnText(Object element, int columnIndex) {
      String[] row = (String[]) element;
      return row[columnIndex];
    }
  }
  /* テーブルデータ */
  private static class ContentProvider implements IStructuredContentProvider {
    public Object[] getElements(Object inputElement) {
      return (Object[]) inputElement;
    }
    public void dispose() { }
    public void inputChanged(Viewer viewer, Object oldInput, Object newInput) { }
  }
  
  /**
   * コンストラクタ
   */
  public KeyMapPreferencePage(boolean isMainWindow) {
    setTitle("入力(" + (isMainWindow ? "メイン画面" : "画像表示画面") + ")");
    this.isMainWindow = isMainWindow;
  }

  /**
   * 入力データ検証
   */
  @Override
  public boolean validate() {
    Log.start();
    setErrorMessage(null);
    return true;
  }

  /**
   * 設定データで画面を更新する
   */
  @Override
  public void updateDialogFromData(boolean forceUpdate) {
    Log.start(forceUpdate);
    if (!forceUpdate && super.initialized) return;
    super.initialized = true;

    Config c = ConfigDialog.getConfig();
    
    // データ作成
    CommandMapInput def = isMainWindow ? new DefaultCommandMapInputCreator().createMainWindowCommandMap() :
                                         new DefaultCommandMapInputCreator().createViewWindowCommandMap();
    CommandMapInput cmdMap = isMainWindow ? c.getCmdMapMain() : c.getCmdMapView();
    tableData.clear();
    for (Command cmd : def.getAllCommands()) {
      tableData.put(cmd, cmdMap.reverseFindInputs(cmd));
    }
    
    // テーブルにデータをセット
    updateTable();
  }

  /**
   * 画面入力で設定データを更新する
   */
  @Override
  public void updateDataFromDialog() {
    Log.start();
    
    Config c = ConfigDialog.getConfig();
    
    if (isMainWindow) c.setCmdMapMain(getCommandMapInput());
    else              c.setCmdMapView(getCommandMapInput());
  }

  /*
   * 現状のデータから CommandMapInput を作成する
   */
  private CommandMapInput getCommandMapInput() {
    CommandMapInput cmds = new CommandMapInput();
    for (Command cmd : tableData.keySet()) {
      for (InputCombination i : tableData.get(cmd)) {
        cmds.addCommandInput(i, cmd);
      }
    }
    return cmds;
  }

  /*
   * テーブルにデータをセット
   */
  private void updateTable() {
    ArrayList<String[]> rows = new ArrayList<String[]>();
    for (Command cmd : tableData.keySet()) {
      StringBuilder s = new StringBuilder();
      for (InputCombination input : tableData.get(cmd)) {
        if (s.length() != 0) {
          s.append("\n");
          if (SystemUtils.IS_OS_WINDOWS) s.append(" / ");
        }
        s.append(input);
      }
      rows.add(new String[] { cmd.getGroup() + " - " + cmd.getName(), s.toString() });
    }
    getTableViewer().setInput(rows.toArray(new String[0][]));
  }
  
  /*
   * テーブルダブルクリック イベントハンドラー
   */
  private void table_mouseDoubleClick(MouseEvent e) {
    int sel = getTableViewer().getTable().getSelectionIndex();
    if (sel < 0) return;
    
    Command cmd = tableData.keySet().toArray(new Command[0])[sel];
    
    // 編集ダイアログ表示
    KeyEditDialog dlg = new KeyEditDialog(getShell(), cmd, tableData.get(cmd), getCommandMapInput());
    if (Window.OK != dlg.open()) return;

    // 表示更新
    tableData.put(cmd, new TreeSet<InputCombination>(dlg.getInputs()));
    updateTable();
    getTableViewer().getTable().setSelection(sel);
  }

  /*
   * テーブルサイズ変更 イベントハンドラー
   */
  private void table_controlResized(ControlEvent e) {
    int w = getTableViewer().getTable().getClientArea().width / 2;
    getTableViewer().getTable().getColumn(0).setWidth(w);
    getTableViewer().getTable().getColumn(1).setWidth(w);
  }

  @Override
  protected Control createContents(Composite parent) {
    Composite composite = new Composite(parent, SWT.NONE);
    composite.setLayout(new GridLayout(1, false));
    
    tableViewer = new TableViewer(composite, SWT.BORDER | SWT.FULL_SELECTION);
    table = tableViewer.getTable();
    table.addMouseListener(new MouseAdapter() {
      @Override
      public void mouseDoubleClick(MouseEvent e) {
        table_mouseDoubleClick(e);
      }
    });
    table.addControlListener(new ControlAdapter() {
      @Override
      public void controlResized(ControlEvent e) {
        table_controlResized(e);
      }
    });
    table.setLinesVisible(true);
    table.setHeaderVisible(true);
    table.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));
    
    TableViewerColumn tableViewerColumn = new TableViewerColumn(tableViewer, SWT.NONE);
    TableColumn tblclmnCmd = tableViewerColumn.getColumn();
    tblclmnCmd.setWidth(100);
    tblclmnCmd.setText("コマンド");
    
    TableViewerColumn tableViewerColumn_1 = new TableViewerColumn(tableViewer, SWT.NONE);
    TableColumn tblclmnKey = tableViewerColumn_1.getColumn();
    tblclmnKey.setWidth(100);
    tblclmnKey.setText("入力");
    tableViewer.setLabelProvider(new TableLabelProvider());
    tableViewer.setContentProvider(new ContentProvider());
    
    Button btnNewButton = new Button(composite, SWT.NONE);
    btnNewButton.addSelectionListener(new SelectionAdapter() {
      @Override
      public void widgetSelected(SelectionEvent e) {
        table_mouseDoubleClick(null);
      }
    });
    btnNewButton.setText("編集(&E)");
    return composite;
  }

  private TableViewer getTableViewer() {
    return tableViewer;
  }
}
