package cvj2.app.gui.util;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.eclipse.swt.SWT;

import cvj2.util.Log;

/**
 * バックグラウンドスレッドなどで、エラーが起きたときにフォアグラウンドスレッドでエラー表示を行う
 */
public class GlobalErrorHandler {
  static private GlobalErrorHandler instance = new GlobalErrorHandler();
  
  /** デバッグモード */
  static public boolean isDebug;
  
  private Object LIST_LOCK = new String("LIST_LOCK");
  private List<ErrorInfo> list = new ArrayList<GlobalErrorHandler.ErrorInfo>();
  private static class ErrorInfo {
    boolean isFatal;
    Throwable exception;
    String msg;
  }
  private CheckThread checkThread;
  private String lastErrorMessage = "";
  
  /**
   * インスタンス取得
   */
  static public GlobalErrorHandler instance() {
    return instance;
  }
  
  /**
   * コンストラクタ
   */
  private GlobalErrorHandler() {
    checkThread = new CheckThread();
    checkThread.setDaemon(true);
    checkThread.setName("GEH");
    checkThread.start();
  }
  
  /**
   * 画面に出力してほしい深刻なエラーをセットする
   */
  public void setFatalMessage(String msg, Throwable e) {
    Log.start(msg, e);
    addError(msg, e, true);
  }
  
  /**
   * 画面に出力してほしいエラーをセットする
   */
  public void setErrorMessage(String msg, Throwable e) {
    Log.start(msg, e);
    addError(msg, e, false);
  }
  
  /**
   * 画面に出力してほしいエラーをセットする (デバッグモードの場合のみ出力)
   */
  public void setDebugMessage(String msg, Throwable e) {
    Log.start(msg, e);
    if (!isDebug) return;
    addError("デバッグ用メッセージ\n\n" + msg, e, false);
  }

  /*
   * リストに追加
   */
  private void addError(String msg, Throwable e, boolean fatal) {
    ErrorInfo ei = new ErrorInfo();
    ei.exception = e;
    ei.msg = msg;
    ei.isFatal = fatal;
    synchronized (LIST_LOCK) {
      list.add(ei);
    }
  }
  
  /*
   * エラーメッセージがあれば、画面に出力する
   * @return メッセージがなければ、false
   */
  private boolean reportErrorMsgBox() {
    ErrorInfo ei;
    synchronized (LIST_LOCK) {
      if (list.size() == 0) return false;
      ei = list.remove(0);
    }

    // 同じメッセージを表示しない
    if (ei.msg.equals(this.lastErrorMessage)) return true;
    String dispMsg = getErrMsg(ei, true);
    String innerMsg = getErrMsg(ei, false);
    
    if (!ei.isFatal) {
      // 深刻でないメッセージなら、表示するだけ
      int ret = MsgBox.showMsgBoxDialog("エラー", SWT.ICON_ERROR, 
          new String[] { "閉じる", "同じエラーを表示しない", }, new char[] { '\0', '\0', }, 0, 
          dispMsg, innerMsg);
      if (ret == 1) this.lastErrorMessage = ei.msg;
      return true;
    }
    
    boolean oomErr = false;
    Throwable parentEx = ei.exception;
    while (parentEx != null) {
      if (parentEx instanceof OutOfMemoryError) oomErr = true;
      parentEx = parentEx.getCause();
    }
    
    // 深刻な時は、強制終了するか尋ねる
    String s = "大変恐縮ですが重大なエラーが発生してしまいました。\n" +
        "プログラム内部が正常でなくなっているかもしれません。\n" +
        "もし可能なら、アプリケーションを再起動させてみてください。\n\n";
    if (oomErr) {
      s = "メモリー不足のエラーが発生してしまいました。\n" + 
          "システム設定を見直してみてください。\n" +
          "プログラム内部が正常でなくなっているかもしれません。\n" +
          "もし可能なら、アプリケーションを再起動させてみてください。\n\n";
    }
    dispMsg = s + dispMsg;
    innerMsg = s + innerMsg;
    int ret = MsgBox.showMsgBoxDialog("重大なエラー", SWT.ICON_ERROR, 
        new String[] { "閉じる", "同じエラーを表示しない", "今すぐ強制終了" }, new char[] { '\0', '\0', '\0' }, 0, 
        dispMsg, innerMsg);
    if (ret == 2) {
      Log.e("強制終了します");
      System.exit(1);
    }
    if (ret == 1) this.lastErrorMessage = ei.msg;
    return true;
  }

  /*
   * 例外を文字列に変換
   */
  private String getErrMsg(ErrorInfo ei, boolean dispMsg) {
    if (ei.exception == null) return ei.msg;

    Throwable e = ei.exception;
    if (dispMsg) {
      while (e.getCause() != null) {
        e = e.getCause();
      }
    }
    
    String[] lines = StringUtils.split(Log.getErrString(e), '\n');
    int maxline = dispMsg ? 15 : Integer.MAX_VALUE;
    return ei.msg + "\n\n" + ei.exception.getMessage() + "\n" + StringUtils.join(lines, '\n', 0, Math.min(maxline, lines.length));
  }
  
  /*
   * エラーチェックスレッド
   */
  private class CheckThread extends Thread {
    @Override
    public void run() {
      while (true) {
        try {
          while (reportErrorMsgBox()) { /* */ }
          Thread.sleep(500);
        }
        catch (InterruptedException e) {
          Log.e(e);
        }
      }
    }
  }
}
