package cvj2.app.gui.widgets;


import java.awt.Dimension;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Set;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.wb.swt.SWTResourceManager;

import cvj2.app.Config;
import cvj2.app.file.imgload.BackgroundImageLoader;
import cvj2.app.file.imgload.CachedImage;
import cvj2.app.file.imgload.ImageLoadProperty;
import cvj2.app.gui.window.WindowController;
import cvj2.app.gui.window.WindowSelector;
import cvj2.app.gui.window.WindowSelector.Type;
import cvj2.app.gui.window.view.ViewWindow;
import cvj2.app.gui.window.view.ViewWindowCommandHandler;
import cvj2.app.gui.window.view.ViewWindowFullscreen;
import cvj2.file.fs.FileInfo;
import cvj2.file.fs.FileInfoList;
import cvj2.gui.swt.ui.input.KeyCombination;
import cvj2.gui.swt.ui.input.MouseCombination;
import cvj2.gui.swt.util.NullCursor;
import cvj2.gui.swt.widgets.CustomScrollbar;
import cvj2.gui.swt.widgets.DoubleBufferedCanvas;
import cvj2.gui.swt.widgets.gdidpanel.GridSelection;
import cvj2.img.SWT2AWT;
import cvj2.util.Log;

/**
 * 一枚画像を表示するパネル。<br>
 * <pre>
 * {@link ViewWindow} と {@link ViewWindowFullscreen} の部品として使用される。
 * {@link ViewWindow} では、タブの中にそれぞれ1枚づつ、
 * {@link ViewWindowFullscreen} では1枚のみ存在する。
 * 
 * 画面の描画は {@link ViewPanelPainter} が行い、
 * 描画する内容の管理は、{@link ViewPanelState} が行う。
 * </pre>
 */
public class ViewPanel extends Composite {
  private DoubleBufferedCanvas canvas;
  private CustomScrollbar scrollV;
  private CustomScrollbar scrollH;
  
  // キー、マウスのイベントハンドラー
  private ViewWindowCommandHandler cmdHandler;
  // ウインドウを操作するためのクラス
  private WindowController wndController;

  // 画像描画クラス
  private ViewPanelPainter painter;

  // ズームされた状態でキー入力された時刻
  private long lastZoomEventTime = 0;
  
  // マウスが動いた時刻
  private long lastMouseMoveTime = new Date().getTime();
  
  // ステータスバーのメッセージ
  private String statusMessage = "";
  
  /**
   * コンストラクタ
   */
  public ViewPanel(Composite parent, ViewWindowCommandHandler cmdHandler, WindowController wndController) {
    super(parent, 0);
    this.cmdHandler = cmdHandler;
    this.wndController = wndController;
    
    scrollV = new CustomScrollbar(this, SWT.BORDER | SWT.V_SCROLL);
    scrollV.setBounds(305, 28, 25, 144);
    
    scrollH = new CustomScrollbar(this, SWT.BORDER | SWT.H_SCROLL);
    scrollH.setBounds(53, 174, 221, 30);

    canvas = new DoubleBufferedCanvas(this, SWT.NO_BACKGROUND);
    canvas.addMouseMoveListener(new MouseMoveListener() {
      public void mouseMove(MouseEvent e) {
        canvas_mouseMove(e);
      }
    });
    canvas.setBounds(35, 28, 279, 163);
    
    super.addControlListener(new ControlAdapter() {
      @Override
      public void controlResized(ControlEvent e) {
        layoutControl();
      }
    });
    canvas.addKeyListener(new KeyAdapter() {
      @Override
      public void keyPressed(KeyEvent e) {
        canvas_keyPressed(e);
      }
    });
    canvas.addMouseListener(new MouseAdapter() {
      @Override
      public void mouseDown(MouseEvent e) {
        canvas_mouseDown(e);
      }
    });
    canvas.addPaintListener(new PaintListener() {
      @Override
      public void paintControl(PaintEvent e) {
        canvas_paintControl(e);
      }
    });
    
    initializeData();
  }

  /**
   * 画面部品をレイアウト
   */
  private void layoutControl() {
    int scrollWidth = 16;
    
    Rectangle rect = super.getClientArea();
    canvas.setBounds(super.getClientArea());
    
    rect.x = rect.x + rect.width - scrollWidth;
    rect.width = scrollWidth;
    rect.height -= scrollWidth;
    scrollV.setBounds(rect);
    
    rect = super.getClientArea();
    rect.y = rect.y + rect.height - scrollWidth;
    rect.height = scrollWidth;
    rect.width -= scrollWidth;
    scrollH.setBounds(rect);
  }

  /*
   * マウス移動イベントハンドラ
   */
  private void canvas_mouseMove(MouseEvent e) {
    lastMouseMoveTime = new Date().getTime();
  }
  
  /*
   * マウスボタン イベントハンドラー
   */
  protected void canvas_mouseDown(MouseEvent e) {
    Log.start(e);
    cmdHandler.handle(new MouseCombination(e));
  }
  
  /*
   * キーイベントハンドラー
   */
  private void canvas_keyPressed(KeyEvent e) {
    Log.start(e);
    cmdHandler.handle(new KeyCombination(e));
  }

  /*
   * 画面描画ハンドラー
   */
  private void canvas_paintControl(PaintEvent e) {
    painter.paintControl(e);
  }
  
  /**
   * クラス内データの初期化
   */
  private void initializeData() {
    painter = new ViewPanelPainter(this);

    // フォントセット
    getCanvas().setFont(Config.get().getFont());
    
    // スクロールバーを最前面に
    scrollH.moveAbove(canvas);
    scrollV.moveAbove(canvas);
    
    // スクロールバーの背景色設定
    class ScrollBgPainter implements PaintListener {
      @Override public void paintControl(PaintEvent e) {
        e.gc.setBackground(SWTResourceManager.getColor(SWT.COLOR_WIDGET_DARK_SHADOW));
        e.gc.fillRectangle(e.x, e.y, e.width, e.height);
      }
    };
    scrollH.setBgPaintListener(new ScrollBgPainter());
    scrollV.setBgPaintListener(new ScrollBgPainter());
    scrollH.setIgnoreEvent(true);
    scrollV.setIgnoreEvent(true);
    
    // 画面内のタイマー
    getDisplay().timerExec(100, new TimerUpdate());
  }
  
  /*
   * 画面内のタイマー
   */
  private class TimerUpdate implements Runnable {
    @Override
    public void run() {
      if (isDisposed()) return;
      
      // スクロールバー表示更新
      scrollUnvisibleUpdate();
      // マウスカーソル表示更新
      mouseCursorUpdate();
      
      getDisplay().timerExec(100, new TimerUpdate());
    }
  }

  /*
   * 時間が経ったら、マウスカーソルを消す
   */
  private void mouseCursorUpdate() {
    boolean hide = lastMouseMoveTime + Config.get().getViewWindowFullscreenHideCursorTime() < new Date().getTime();
    hide = hide && getParent().getClass().equals(ViewWindowFullscreen.class);
    getShell().setCursor(hide ? NullCursor.get() : null);
  }
  
  /*
   * 時間が経ったら、スクロールバーを非表示にする
   */
  private void scrollUnvisibleUpdate() {
    boolean visible = isScrollVisible();
    if (!visible && scrollH.isVisible()) scrollH.setVisible(false);
    if (!visible && scrollV.isVisible()) scrollV.setVisible(false);
  }
  
  /*
   * スクロールバーを表示してもよいか
   */
  private boolean isScrollVisible() {
    long time = new Date().getTime();
    return (time < lastZoomEventTime + Config.get().getViewWindowScrollVisibleTime());
  }
  
  /**
   * スクロールバーを表示する
   * ※) このメソッドは、{@link ViewPanelPainter} の描画メソッド中から呼ばれる
   */
  void scrollUpdate(Rectangle imgRect, Rectangle viewRect) {
    if (!isScrollVisible()) return;
    Log.ut("imgRect = ", imgRect, ", viewRect = ", viewRect);

    boolean visibleH = imgRect.width > viewRect.width;
    boolean visibleV = imgRect.height > viewRect.height;
    scrollH.setVisible(visibleH);
    scrollV.setVisible(visibleV);

    if (visibleH) {
      scrollH.moveAbove(canvas);
      scrollH.setup(0, imgRect.width, viewRect.width, imgRect.width);
      scrollH.setSelection(viewRect.x);
      scrollH.redraw();
    }
    if (visibleV) {
      scrollV.moveAbove(canvas);
      scrollV.setup(0, imgRect.height, viewRect.height, imgRect.height);
      scrollV.setSelection(viewRect.y);
      scrollV.redraw();
    }
  }

  /*
   * スクロールバーを表示する
   * ※) ここではタイマー用の時刻をセットするだけ
   */
  private void showScrollbar() {
    lastZoomEventTime = new Date().getTime();
  }

  /*
   * 選択されているタブ位置
   */
  int getCurrentTabIndex() {
    return wndController.getCurrentTabIndex();
  }
  
  /**
   * 選択されているタブの画像表示情報
   */
  public ViewPanelState getCurrentTabState() {
    return wndController.getCurrentListPanel().getViewPanelState().clone();
  }

  /**
   * 表示する画像ファイルの状態を設定して再読み込みする
   */
  public void setCurrentTabState(ViewPanelState state) {
    wndController.getCurrentListPanel().setViewPanelState(state.clone());
    canvas.redraw();
  }

  /**
   * ズームの設定
   * @param direction 1: ズームイン、-1: ズームアウト
   */
  public void setZoom(int direction) {
    Config conf = Config.get();
    showScrollbar();
    ViewPanelState state = getCurrentTabState();
    if (direction > 0) {
      state.zoom.zoomIn();
    }
    else {
      state.zoom.zoomOut();
    }
    if (conf.getViewWindowZoomInRatioMax() < state.zoom.getRatio()) return;
    if (conf.getViewWindowZoomOutRatioMin() > state.zoom.getRatio()) return;
    setCurrentTabState(state);
  }
  
  /**
   * 画像表示位置の移動
   */
  public void moveImageLocation(int directionH, int directionV) {
    Rectangle canvasRect = canvas.getBounds();
    ViewPanelState state = getCurrentTabState();
    CachedImage img = state.getImage(this);
    if (img == null || !img.ref()) return;
    showScrollbar();
    try {
      Dimension canvasSize = SWT2AWT.toDimension(canvasRect);
      Dimension drawImageSize = state.getDisplayImageSize(img, canvasRect);
      boolean move = state.location.move(directionH, directionV, drawImageSize, canvasSize);
      if (!move) return;
      setCurrentTabState(state);
    }
    finally {
      img.unref();
    }
  }

  /**
   * ListPanel で選択されているファイルで、表示する画像を更新する
   */
  public void updateListPanelSelectionImage() {
    // ロードする画像情報の取得
    ViewPanelState state = getForwardLoadFile(0);
    
    // 画像ファイルがみつからないなら、メイン画面に戻る
    if (state.file == null) {
      WindowSelector.instance().select(Type.MAIN);
      return;
    }

    // 画像表示
    setCurrentTabState(state);
  }
  
  private ImageLoadProperty lastLoadProperty = null;
  /**
   * 画像の先読み
   */
  public void loadForwardImage(CachedImage currentImage) {
    // 現在表示している画像をキャッシュに収めるのに必要なメモリを計算する
    ImageLoadProperty loadProp = getCurrentTabState().getImageLoadProperty(this);
    long needMem = 0;
    if (currentImage != null) {
      needMem = currentImage.getObjectBytes();
    }
    else {
      Dimension size = SWT2AWT.toDimension(this.getClientArea());
      if (null != loadProp.getTargetSize()) size = loadProp.getTargetSize();
      needMem = size.width * size.height * 4;

      // 現在表示しようとしている画像がロードされていなかったら、自分以外をキャンセルする
      if (!loadProp.equals(lastLoadProperty)) {
        for (ImageLoadProperty prop : BackgroundImageLoader.instance().getAllTaskInfo()) {
          if (loadProp.equals(prop)) continue;
          BackgroundImageLoader.instance().cancelTask(prop);
          Log.d("cancel: " + prop);
        }
      }
      
      lastLoadProperty = loadProp;
    }

    // キャッシュに収まるぶんだけ先読みする画像を探す
    Config conf = Config.get();
    long cacheLimit = (long)(conf.getImageLoadMemCacheBytesView() * conf.getViewWindowForwardLoadInCacheRatio());
    List<ViewPanelState> stateList = new ArrayList<ViewPanelState>();
    List<ImageLoadProperty> loadPropList = new ArrayList<ImageLoadProperty>();
    for (int i = 1; i <= conf.getViewWindowForwardLoadNum(); i++) {
      cacheLimit -= needMem;
      if (cacheLimit < 0) break;
      ViewPanelState state = getForwardLoadFile(i);
      if (state.file == null) break;
      stateList.add(state);
      loadPropList.add(state.getImageLoadProperty(this));
    }
    
    // 先読みリストに入っていないタスクはキャンセルする
    Set<ImageLoadProperty> allTaskInfo = BackgroundImageLoader.instance().getAllTaskInfo();
    for (ImageLoadProperty taskLoadProp : allTaskInfo) {
      if (loadProp.equals(taskLoadProp)) continue;
      if (loadPropList.contains(taskLoadProp)) continue;
      BackgroundImageLoader.instance().cancelTask(taskLoadProp);
    }
    
    // 読み込む
    for (ViewPanelState state : stateList) {
      ImageLoadProperty forwardProp = state.getImageLoadProperty(this);
      if (allTaskInfo.contains(forwardProp)) continue;
      Log.d("forward load: ", forwardProp);
      state.getImage(this);
    }
  }
  
  /**
   * 読み込む画像情報の取得
   * @param forwardIndex 現在位置からの移動量
   *         0: 現在位置
   *         1: 1つ先
   *         2: 2つ先
   */
  private ViewPanelState getForwardLoadFile(int forwardIndex) {
    assert 0 <= forwardIndex; 
    
    ViewPanelState state = getCurrentTabState().clone();
    ListPanel lp = wndController.getCurrentListPanel();
    FileInfoList list = lp.getModel().getSortedList();
    GridSelection sel = lp.getSelection();
    int lastMoveDirection = sel.getLastMoveDirection();
    if (lastMoveDirection == 0) lastMoveDirection = 1;
    int moveCount = forwardIndex * lastMoveDirection;
    if (state.page2) moveCount *= 2;
    int imageIndex = sel.getFocusIndex() + moveCount;
    
    state.file = null;
    state.file2 = null;
    
    // ファイルがないとき
    if (list.size() == 0) return state;
    if (!state.page2 && imageIndex < 0) return state;
    if (state.page2 && imageIndex < -1) return state;
    if (list.size() <= imageIndex) return state;
    
    // 2ページ表示で、[-1,0] を 読もうとしたら、[0,1]を読む
    if (state.page2 && imageIndex == -1) {
      state.file = list.get(0);
      state.file2 = list.size() == 1 ? null : list.get(1);
      return checkReadableImage(state);
    }

    // 1ページ表示のとき
    state.file = list.get(imageIndex);
    state.file2 = null;
    if (!state.page2) return checkReadableImage(state);
    
    // 2ページ表示のとき
    if (imageIndex < list.size() - 1) {
      if (state.page2Reverse) {
        state.file = list.get(imageIndex + 1);
        state.file2 = list.get(imageIndex);
      }
      else {
        state.file = list.get(imageIndex);
        state.file2= list.get(imageIndex + 1);
      }
    }
    return checkReadableImage(state);
  }

  /* 
   * 表示できる画像かどうかチェックする
   */
  private ViewPanelState checkReadableImage(ViewPanelState state) {
    // 1枚めが画像でないなら、2枚めもダメ
    if (!checkReadableImage(state.file)) {
      state.file = null;
      state.file2 = null;
      return state;
    }
    if (!checkReadableImage(state.file2)) state.file2 = null;
    return state;
  }
  /* 
   * 表示できる画像かどうかチェックする
   */
  private boolean checkReadableImage(FileInfo f) {
    if (f == null) return false;
    if (f.isDirectory()) return false;
    if (!f.isImageFile()) return false;
    return true;
  }

  /**
   * ステータスメッセージの取得
   */
  public String getStatusMessage() {
    return statusMessage;
  }

  /**
   * ステータスメッセージの設定
   */
  public void setStatusMessage(String statusMessage) {
    assert statusMessage != null;
    this.statusMessage = statusMessage;
    if (super.isDisposed()) return;
    getCanvas().redraw();
  }

  /** Canvas にフォーカスを設定するためにオーバーライド */
  @Override
  public boolean setFocus() {
    return canvas.setFocus();
  }

  @Override
  protected void checkSubclass() {
    // Disable the check that prevents subclassing of SWT components
  }
  
  /** 画像表示Canvasを取得 */
  protected Canvas getCanvas() {
    return canvas;
  }
}
