package cvj2.app.gui.widgets;

import java.awt.Dimension;
import java.text.DecimalFormat;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.wb.swt.SWTResourceManager;

import cvj2.app.Config;
import cvj2.app.file.imgload.BackgroundImageLoader;
import cvj2.app.file.imgload.CachedImage;
import cvj2.app.file.imgload.ImageLoadProperty;
import cvj2.file.fs.FileInfo;
import cvj2.gui.swt.util.GCX;
import cvj2.gui.swt.util.SWTUtil;
import cvj2.img.AWT2SWT;
import cvj2.img.ImageInfo;
import cvj2.img.ImageRotate;
import cvj2.img.ImageUtil;
import cvj2.img.SWT2AWT;
import cvj2.util.DelimString;
import cvj2.util.Log;

/**
 * ViewPanel の描画を行う
 */
public class ViewPanelPainter {
  private ViewPanel vp;
  
  /**
   * コンストラクタ
   */
  public ViewPanelPainter(ViewPanel vp) {
    this.vp = vp;
  }

  /**
   * 描画
   */
  public void paintControl(PaintEvent e) {
    Log.start(e);
    GCX gcx = new GCX(e.gc);

    ViewPanelState state = vp.getCurrentTabState();
    if (state == null) return;
    Rectangle canvasRect = vp.getCanvas().getClientArea();
    Config conf = Config.get();
    FileInfo file = state.file;
    if (file == null) return;
    String comment = null;

    // 指定サイズの画像をキャッシュから取得する
    CachedImage img = state.getImage(vp);
    boolean gotCorrectImage = img != null;
    // 指定サイズの画像がないときは、何でもいいから取得する
    if (img == null) {
      img = BackgroundImageLoader.instance().findViewCache(file);
      // ただし、サイズ以外が違っていると使用できない
      if (img != null && !img.getLoadProperty().equalsFileAndRotate(file, state.file2, state.rotate)) img = null;
    }

    if (img != null && img.ref()) {
      // 画像あり
      Log.ut("img = ", img, ", gotCorrectImage = ", gotCorrectImage);
      try {
        gcx.setBackground(SWTResourceManager.getColor(AWT2SWT.toRGB(conf.getImageBackgroundColor())));
        gcx.fillRectangle(canvasRect.x, canvasRect.y, canvasRect.width + 1, canvasRect.height + 1);

        if (!img.isSystemResource()) {
          // ロードされた画像は設定に従って拡大縮小回転して描画する
          drawImage(gcx, img, canvasRect, state, gotCorrectImage);

          if (!gotCorrectImage) {
            comment = getImageComment(null, state);
          }
          else if (conf.viewWindowShowImageInfo) {
            comment = getImageComment(img, state);
          }
        }
        else {
          // ロード失敗の場合は、画面の中心にそのまま描画する
          Rectangle imgDrawRect = AWT2SWT.toRectangle(ImageUtil.calcImageCenterRect(img.getSize(),
              SWT2AWT.toRectangle(canvasRect)));
          gcx.drawImage(img.getImage(), imgDrawRect.x, imgDrawRect.y);
          Point pt = gcx.stringExtent(img.getLoadResult().loadFailMessage);
          gcx.drawString(img.getLoadResult().loadFailMessage, imgDrawRect.x - pt.x / 2, imgDrawRect.y + imgDrawRect.height);
        }
      }
      finally {
        img.unref();
      }
    }
    else {
      // 画像まだ
      comment = getImageComment(null, state);
    }
    
    // 画像の先読み
    vp.loadForwardImage(img);

    // コメント描画
    if (comment != null) gcx.drawText(comment, canvasRect.x + 5, canvasRect.y + 5);
    
    // ステータスバーメッセージ表示
    if (vp.getStatusMessage().length() > 0) {
      String msg = "  " + vp.getStatusMessage();
      Point stPt = gcx.stringExtent(msg);
      gcx.setBackground(SWTResourceManager.getColor(SWT.COLOR_LIST_SELECTION));
      gcx.setForeground(SWTResourceManager.getColor(SWT.COLOR_LIST_SELECTION_TEXT));
      gcx.fillRectangle(canvasRect.x, canvasRect.y + canvasRect.height - stPt.y, canvasRect.width, stPt.y + 1);
      gcx.drawText(msg, 0, canvasRect.y + canvasRect.height - stPt.y);
    }
  }

  /**
   * ロードされた画像の表示
   */
  private void drawImage(GCX gcx, CachedImage img, Rectangle canvasRect, ViewPanelState state, boolean gotCorrectImage) {
    assert img.getRefCount() > 0;
    
    // 読み込んだ画像の大きさ
    Rectangle imgRect = img.getImage().getBounds();
    // 画像を拡大縮小回転して表示する領域の大きさ
    Dimension drawImageSize = state.getDisplayImageSize(img, canvasRect);
    // 画像を拡大縮小回転して表示する領域の位置
    Rectangle drawImageRect = AWT2SWT.toRectangle(ImageUtil.calcImageCenterRect(drawImageSize, SWT2AWT.toRectangle(canvasRect)));
    // スクロール後の表示位置
    java.awt.Point dispPoint = ImageUtil.getImageDispLocation(drawImageSize, SWT2AWT.toDimension(canvasRect), state.location.getLocation());
    drawImageRect.x = dispPoint.x;
    drawImageRect.y = dispPoint.y;

    // BITBLT の転送元領域
    Rectangle srcRect = SWTUtil.clone(imgRect);
    // BITBLT の転送先領域
    Rectangle destRect = SWTUtil.clone(drawImageRect);

    if (destRect.x < 0) {
      // 左にはみ出した領域
      double leftOver = (double) -destRect.x / destRect.width;
      srcRect.x += (int) (srcRect.width * leftOver);
      destRect.x = 0;
    }
    if (destRect.y < 0) {
      // 上にはみ出した領域
      double topOver = (double) -destRect.y / destRect.height;
      srcRect.y += (int) (srcRect.height * topOver);
      destRect.y = 0;
    }
    if (destRect.x + destRect.width > canvasRect.width) {
      // 右にはみ出した領域
      double rightOver = (double) ((destRect.x + destRect.width) - canvasRect.width) / destRect.width;
      srcRect.width -= (int) (srcRect.width * rightOver);
      destRect.width -= (int) (destRect.width * rightOver);
    }
    if (destRect.y + destRect.height > canvasRect.height) {
      // 下にはみ出した領域
      double bottomOver = (double) ((destRect.y + destRect.height) - canvasRect.height) / destRect.height;
      srcRect.height -= (int) (srcRect.height * bottomOver);
      destRect.height -= (int) (destRect.height * bottomOver);
    }

    // 描画
    gcx.drawImage(img.getImage(), srcRect.x, srcRect.y, srcRect.width, srcRect.height, destRect.x, destRect.y, destRect.width, destRect.height);
    
    // スクロールバーの表示更新
    if (gotCorrectImage) {
      vp.scrollUpdate(imgRect, srcRect);
    }
    
    Log.ut("drawImageSize = ", drawImageSize, ", drawImageRect = ", drawImageRect, ", imgRect = ", imgRect);
    Log.ut("canvasRect = ", canvasRect, ", dispPoint = ", dispPoint, ", location = ", state.location);
    Log.ut("srcRect = ", srcRect, ", destRect = ", destRect);
  }

  /**
   * 画像コメントとの取得
   */
  private String getImageComment(CachedImage img, ViewPanelState state) {
    DelimString comment = new DelimString(" \n");

    // タブ番号
    comment.append("Tab [ " + (vp.getCurrentTabIndex() + 1) + " ] ");
    
    // ロード中か？
    if (img == null) {
      comment.append(state.file);
      comment.append("＝＝＝＝＝＝＝＝＝＝＝＝");
      comment.append("＊＊＊＊＊＊＊＊＊＊＊＊");
      comment.append("　　　ロード中です");
      comment.append("＊＊＊＊＊＊＊＊＊＊＊＊");
      comment.append("＝＝＝＝＝＝＝＝＝＝＝＝");
      return comment.toString();
    }

    ImageLoadProperty prop = img.getLoadProperty();

    // ファイル名
    comment.append(prop.getFileOrDir());
    comment.append("  ファイルサイズ: " + DecimalFormat.getNumberInstance().format(prop.getFileOrDir().length()));
    if (prop.getFile2() != null) {
      comment.append(prop.getFile2());
      comment.append("  ファイルサイズ: " + DecimalFormat.getNumberInstance().format(prop.getFile2().length()));
    }
    comment.append("");

    ImageInfo[] infos = img.getLoadResult().originalImageInfos;
    ImageInfo info = img.getLoadResult().originalImageInfo;

    // 表示している画像のサイズ
    if (infos[1] == null) {
      // 1ページ表示
      comment.append(toString(infos[0]));
    }
    else {
      // 2ページ表示
      comment.append("1ページ目: " + toString(infos[0]));
      comment.append("2ページ目: " + toString(infos[1]));

      comment.append("  合成した画像は、幅 " + info.width + " x 高 " + info.height);
      
      if (state.page2Reverse) comment.append("2ページ表示: [1ページ目] ← [2ページ目]");
      else                    comment.append("2ページ表示: [2ページ目] ← [1ページ目]");
    }
    
    // 回転
    comment.append("画像の回転: " + toString(prop.getRotate()));
    
    // ズーム設定
    if (state.originalSize) {
      comment.append("ズーム設定: オリジナル画像サイズの" + String.format("%4.2f", state.zoom.getRatio()) + "倍 で表示します");
    }
    else {
      comment.append("ズーム設定: 画面にフィットするサイズの" + String.format("%4.2f", state.zoom.getRatio()) + "倍 で表示します");
    }
    
    comment.append("");
    comment.append("メモリキャッシュに格納されている画像サイズ: " + img.getSize().width + " x " + img.getSize().height + 
                   " (元画像の " + String.format("%4.2f", (double)img.getSize().width / info.width * 100) + " %)");
    comment.append("メモリキャッシュに格納されているバイト数: " + DecimalFormat.getNumberInstance().format(img.getObjectBytes())); 
    
    if (Config.get().isViewWindowShowImageInfoDebugMessage()) {
      comment.append("");
      comment.append("(画像ロード時のデバッグメッセージ)");
      comment.append(img.getLoadResult().convertDebugMessage);
    }
    
    return comment.toString();
  }
  
  private String toString(ImageInfo info) {
    return "幅 " + info.width + " x 高 " + info.height + " " + ImageUtil.getBufferedImageTypeString(info.type);
  }
  
  private String toString(ImageRotate rotate) {
    switch (rotate.getRotate()) {
    case 0: return "回転なし";
    case 1: return "時計回りに270度";
    case 2: return "時計回りに180度";
    case 3: return "時計回りに90度";
    default: return "回転???: " + rotate.getRotate();
    }
  }
}
