package cvj2.app.gui.window;

import java.io.File;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CTabFolder;
import org.eclipse.swt.custom.CTabItem;
import org.eclipse.swt.widgets.Control;

import cvj2.app.Config;
import cvj2.app.file.ApplicationLock;
import cvj2.app.gui.widgets.ListPanel;
import cvj2.app.gui.widgets.ViewPanel;
import cvj2.app.gui.window.view.ViewWindow;
import cvj2.app.gui.window.view.ViewWindowCommandHandler;
import cvj2.app.gui.window.view.ViewWindowFullscreen;
import cvj2.file.fs.FileInfo;
import cvj2.file.fs.FileInfoFactory;
import cvj2.file.fs.FileInfoList;
import cvj2.gui.swt.util.SWTUtil;
import cvj2.util.Log;

/**
 * ListPanel, ViewPanel から、MainWindow, ViewWindow, ViewWindowFullscreen に対して更新処理を行うためのクラス。<br>
 * MainWindow と ViewWindow は、メニュー,タブ,ステータスバー の構造が似ているので、共通の処理をここで実装する。<br>
 * ViewWindowFullscrenn はメニューとタブはないが、ステータスバーに表示されるメッセージを画面に表示する。<br>
 */
public class WindowController {
  // MainWindow
  private CommonApplicationWindow mainWnd;
  // MainWindow or ViewWindow or ViewWindowFullscreen
  private CommonApplicationWindow currentWnd;
  
  /**
   * コンストラクタ
   */
  public WindowController(CommonApplicationWindow mainWindow, CommonApplicationWindow currentWindow) {
    assert mainWindow != null && currentWindow != null;
    this.mainWnd = mainWindow;
    this.currentWnd = currentWindow;
  }

  private String windowTitle = "";
  /**
   * メインウインドウのタイトル設定
   */
  public void setTitle(String title) {
    windowTitle = (title.length() == 0 ? "" : title + " - ") + Config.APP_NAME + (ApplicationLock.instance().isFirstApp() ? "" : "(*)");
    currentWnd.getShell().setText(windowTitle);
  }
  
  /**
   * タブ名の更新
   */
  public void updateTabNames() {
    if (currentWnd instanceof ViewWindowFullscreen) return;
    assert mainWnd.getTabFolder().getItemCount() == currentWnd.getTabFolder().getItemCount();
    
    CTabItem[] mainItems = mainWnd.getTabFolder().getItems();
    for (int i = 0; i < mainWnd.getTabFolder().getItemCount(); i++) {
      boolean focus = (mainItems[i] == mainWnd.getTabFolder().getSelection());
      ListPanel lp = (ListPanel) mainItems[i].getControl();
      CTabItem item = currentWnd.getTabFolder().getItem(i);
      if (focus) {
        item.setText(SWTUtil.escapeControlText(lp.getCurrentDir().getAbsolutePath()));
      }
      else {
        item.setText(SWTUtil.escapeControlText(lp.getCurrentDir().getName()));
      }
    }
  }
  
  /**
   * メニュー状態の更新
   */
  public void updateMenuStatus() {
    ListPanel lp = getCurrentListPanel();
    currentWnd.updateMenuState(lp != null ? lp.isThumbnailMode() : false, lp == null ? null : lp.getSort());
  }
  
  private String lastStatusMessageTemporary = null;
  /**
   * メインウインドウのステータスメッセージ設定 (一時的に)
   */
  public void setStatusMessageTemporary(String message) {
    final String innerString = new String(message); // new しないと、オブジェクトの上書きを検知できない
    Log.ut_start(innerString);
    lastStatusMessageTemporary = innerString;
    currentWnd.setStatus(innerString, true);
    currentWnd.getShell().getDisplay().timerExec(Config.get().getWindowStatusClearTimeMillisec(), new Runnable() {
      @Override
      public void run() {
        if (lastStatusMessageTemporary != innerString) return;
        lastStatusMessageTemporary = null;
        currentWnd.setStatus(statusMessage, false);
      }
    });
  }
  
  private String statusMessage = "";
  /**
   * メインウインドウのステータスメッセージ設定
   */
  public void setStatusMessage(String message) {
    Log.ut_start(message);
    statusMessage = message;
    if (lastStatusMessageTemporary == null) currentWnd.setStatus(message, false);
  }
  
  private String lastRealtimeStatusMessage = null;
  /**
   * 一時メッセージを表示する。<br>
   * イベントハンドラースレッドの中で時間のかかる処理をするときに、ステータスバーでは更新されないのでタイトルを使用する<br>
   * ※) Windows の場合はうまくいった
   */
  public void setRealtimeStatusMessageTemporary(String message) {
    final String innerString = new String(message); // new しないと、オブジェクトの上書きを検知できない
    Log.ut_start(innerString);
    lastRealtimeStatusMessage = innerString;
    currentWnd.getShell().setText(innerString);
    currentWnd.getShell().getDisplay().timerExec(1, new Runnable() {
      @Override
      public void run() {
        if (lastRealtimeStatusMessage != innerString) return;
        if (currentWnd.getShell() == null) return;
        if (currentWnd.getShell().isDisposed()) return;
        currentWnd.getShell().setText(windowTitle);
      }
    });
  }
  
  /**
   * メインウインドウのファイルステータスメッセージ設定
   */
  public void setStatusFileMessage(String msg) {
    currentWnd.setStatusFileMessage(msg);
  }
  
  /**
   * 選択中のタブの ListPanel を取得
   */
  public ListPanel getCurrentListPanel() {
    CTabItem item = mainWnd.getTabFolder().getSelection();
    if (item == null) return null;
    return (ListPanel) item.getControl();
  }

  /**
   * 選択中のタブのインデックスを取得
   */
  public int getCurrentTabIndex() {
    return mainWnd.getTabFolder().getSelectionIndex();
  }
  
  /**
   * 画面更新
   */
  public void refresh() {
    ListPanel lp = getCurrentListPanel();
    if (lp == null) return;

    lp.setMouseScrollType(Config.get().getMainWindowMouseScrollType());
    
    setRealtimeStatusMessageTemporary("ファイル一覧更新");
    setCurrentDir(lp.getCurrentDir(), null);
    setStatusMessageTemporary("最新のファイル一覧を表示しました");
  }
  
  /**
   * 親に移動
   */
  public void moveToParent() {
    ListPanel lp = getCurrentListPanel();
    if (lp == null) return;
    
    FileInfo parent = lp.getCurrentDir().getParent();
    if (parent == FileInfo.ROOT && parent.getList().size() == 1) return;
    setRealtimeStatusMessageTemporary("親ディレクトリに移動");
    setCurrentDir(parent, lp.getCurrentDir());
    setStatusMessageTemporary("親ディレクトリに移動しました");
  }
  
  /**
   * ホームに移動
   */
  public void moveToHome() {
    ListPanel lp = getCurrentListPanel();
    if (lp == null) return;
    
    FileInfo home = FileInfoFactory.create(new File(Config.get().getHomeDir()));
    setRealtimeStatusMessageTemporary("ホームディレクトリに移動");
    setCurrentDir(home, null);
    setStatusMessageTemporary("ホームディレクトリに移動しました");
  }
  
  /*
   * カレントディレクトリの移動
   */
  private void setCurrentDir(FileInfo dir, FileInfo selFile) {
    ListPanel lp = getCurrentListPanel();
    assert lp != null;
    lp.setCurrentDir(dir, selFile);
    lp.updateApplicationWindowState(this);
  }
  
  /**
   * フォーカスの移動
   */
  public void moveCursor(int moveCol, int moveRow, int movePage, int headTailDirection) {
    ListPanel lp = getCurrentListPanel();
    if (lp == null) return;
    
    boolean move = lp.getSelection().moveCursor(moveCol, moveRow, movePage, headTailDirection, false, false);
    
    // MainWindow の場合は、GridPanel の GridSelectionListener が更新しているので update は不要
    if (mainWnd != currentWnd) {
      lp.updateApplicationWindowState(this);
      if (!move) setStatusMessageTemporary("これ以上進めません");
    }
  }
  
  /**
   * タブ移動
   */
  public void moveTabSelect(int direction) {
    if (getCurrentListPanel() == null) return;
    
    CTabFolder tab = mainWnd.getTabFolder();
    int i = tab.getSelectionIndex() + direction;
    if (i < 0) i = tab.getItems().length - 1;
    if (i >= tab.getItems().length) i = 0;
    tab.setSelection(i);
    if (!(currentWnd instanceof ViewWindowFullscreen)) {
      currentWnd.getTabFolder().setSelection(i);
    }
    
    // ステータス表示を更新
    refresh();
    getCurrentListPanel().updateApplicationWindowState(this);
  }

  /**
   * 兄弟ディレクトリに移動
   */
  public void moveSiblingDir(int direction) {
    ListPanel lp = getCurrentListPanel();
    if (lp == null) return;
    
    // 親ディレクトリ内での、カレントディレクトリの位置を探す
    FileInfo dir = lp.getCurrentDir();
    FileInfoList parentChilds = dir.getParent().getList();
    int index = parentChilds.find(dir);
    if (index >= 0) {
      index += direction;
    }
    else {
      index = -1 * (index + 1);
      if (direction < 0) index--;
    }

    if (direction < 0) setRealtimeStatusMessageTemporary("前のディレクトリに移動");
    else               setRealtimeStatusMessageTemporary("次のディレクトリに移動");
    
    // 次のディレクトリを探す
    while (0 <= index && index < parentChilds.size()) {
      FileInfo newDir = parentChilds.get(index);
      if (newDir.isDirectory()) {
        setCurrentDir(newDir, null);
        if (direction < 0) setStatusMessageTemporary("前のディレクトリに移動しました");
        else               setStatusMessageTemporary("次のディレクトリに移動しました");
        return;
      }
      index += direction;
    }
    
    setStatusMessageTemporary("これ以上進めません");
  }

  /**
   * ViewWindow のタブの初期化
   */
  public void initializeViewWindowTab(ViewWindowCommandHandler cmdHandler) {
    assert currentWnd instanceof ViewWindow;
    
    CTabFolder mainTab = mainWnd.getTabFolder();
    CTabFolder currentTab = currentWnd.getTabFolder();
    int tabCount = mainTab.getChildren().length;
    
    for (int i = 0; i < tabCount; i++) {
      CTabItem item = new CTabItem(currentTab, SWT.NONE);
      ViewPanel newVp = new ViewPanel(currentTab, cmdHandler, this);
      item.setControl(newVp);
    }

    moveTabSelect(0);
    updateTabNames();
  }
  
  /**
   * ListPanel の持っている ViewPanelState の中の FileInfo 情報を初期化する
   * (初期化しておかないと、初期表示直後に以前の画像を表示してしまう)
   */
  public void initViewPanelStateFileInfo() {
    for (Control c : mainWnd.getTabFolder().getChildren()) {
      ListPanel lp = (ListPanel) c;
      lp.getViewPanelState().file = null;
      lp.getViewPanelState().file2 = null;
    }
  }
}
