package cvj2.app.gui.window;

import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Shell;

import cvj2.app.file.imgload.BackgroundImageLoader;
import cvj2.app.gui.window.main.MainWindow;
import cvj2.app.gui.window.view.ViewWindow;
import cvj2.app.gui.window.view.ViewWindowFullscreen;
import cvj2.gui.swt.util.SWTUtil;
import cvj2.util.Log;

/**
 * 表示するアプリケーション画面の選択
 */
public class WindowSelector {
  private static WindowSelector instance = new WindowSelector();
  
  private MainWindow main;
  private ViewWindow view;
  private ViewWindowFullscreen viewFull;
  private Type current;
  
  private Rectangle mainBounds;
  private boolean mainMaximized;
  
  /**
   * 画面種別
   */
  public enum Type {
    MAIN,
    VIEW,
    VIEW_FULL,
  }

  // 使用禁止
  private WindowSelector() {}
  
  /**
   * インスタンス取得
   */
  public static WindowSelector instance() {
    return instance;
  }
  
  /**
   * MainWindow の初期表示 (閉じるまでリターンしない)
   */
  public void initialOpen() {
    Log.start();
    current = Type.MAIN;
    main = new MainWindow();
    main.setBlockOnOpen(true);
    main.open();
  }
  
  /**
   * MainWindow を取得
   */
  public MainWindow getMainWindow() {
    return main;
  }
  
  /**
   * MainWindow の位置とサイズ
   */
  public Rectangle getMainWindowBounds() {
    return SWTUtil.clone(mainBounds);
  }
  
  /**
   * MainWindow の最大化状態
   */
  public boolean isMainWindowMaximized() {
    return mainMaximized;
  }
  
  /**
   * 現在表示されている画面のShellを取得する
   * @return 表示されていない時は、null
   */
  public Shell getCurrentShell() {
    if (main != null && main.getShell() != null && !main.getShell().isDisposed()) return main.getShell();
    if (view != null && view.getShell() != null && !view.getShell().isDisposed()) return view.getShell();
    if (viewFull != null && !viewFull.isDisposed()) return viewFull;
    return null;
  }
  
  /**
   * 指定の画面を表示する
   */
  public void select(Type type) {
    Log.start(type);
    assert current != null;
    
    // 全ての画像読み込みをキャンセルする
    BackgroundImageLoader.instance().cancelAllTasks();
    
    // MainWindow の状態を憶えておく
    if (current == Type.MAIN) {
      mainBounds = main.getShell().getBounds();
      mainMaximized = main.getShell().getMaximized();
    }
    
    // 新しい画面を開く
    open(type);
    // 古い画面を閉じる
    close(current);
    
    // MainWindow の状態を復元する (Windowsの場合は必要ない)
    if (type == Type.MAIN) {
      main.getShell().setBounds(mainBounds);
      main.getShell().setMaximized(mainMaximized);
    }
    
    current = type;
  }

  private void open(Type type) {
    Log.start(type);

    switch (type) {
    case MAIN: 
      assert main != null;
      main.getShell().setVisible(true);
      main.getShell().setActive();
      break;
    case VIEW: 
      assert view == null;
      view = new ViewWindow(main);
      view.open();
      break;
    case VIEW_FULL: 
      assert viewFull == null;
      viewFull = new ViewWindowFullscreen(main);
      viewFull.showViewWindowFullscreen();
      break;
    }
  }
  
  private void close(Type type) {
    Log.start(type);
    
    switch (current) {
    case MAIN: 
      assert main != null;
      main.getShell().setVisible(false);
      break;
    case VIEW: 
      assert view != null;
      view.close();
      view = null;
      break;
    case VIEW_FULL:
      assert viewFull != null;
      if (!viewFull.isViewWindowFullscreenClosed()) {
        viewFull.closeViewWindowFullscreen();
        viewFull.dispose();
      }
      viewFull = null;
      break;
    }
  }
}
