package cvj2.app.gui.window.main;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.eclipse.swt.program.Program;

import cvj2.app.Config;
import cvj2.app.gui.util.MsgBox;
import cvj2.app.gui.widgets.ListPanel;
import cvj2.app.gui.window.WindowController;
import cvj2.file.fs.FileInfo;
import cvj2.file.util.FileUtil;
import cvj2.util.Log;

/**
 * MainWindow でのプログラム実行のためのヘルパークラス
 */
public class MainWindowProgramExecutor {
  private static HashSet<String> noConfirmExts = new HashSet<String>();
  private static String lastNoConfirmExts = new String("");

  private MainWindowProgramExecutor() {}
  
  /*
   * 確認不要拡張子かどうか
   */
  private static boolean isNoConfirmExt(String ext) {
    String newConf = Config.get().getExecProgramNoConfirmExts();
    if (newConf != lastNoConfirmExts) {
      lastNoConfirmExts = newConf;
      noConfirmExts.clear();
      for (String s : StringUtils.split(newConf, "\r\n")) {
        s = s.trim().toLowerCase();
        if (s.length() == 0) continue;
        if (!s.startsWith(".")) s = "." + s;
        noConfirmExts.add(s);
      }
    }
    return noConfirmExts.contains(ext.toLowerCase());
  }


  /**
   * 選択されているファイルを実行する
   */
  static public void exec(ListPanel lp, WindowController wndController) {
    if (lp == null) return;

    // 選択されているファイル
    List<FileInfo> list = new ArrayList<FileInfo>();
    for (int i : lp.getSelection().getSelectedIndices()) {
      list.add(lp.getModel().getSortedList().get(i));
    }
    if (list.size() == 0) return;
    
    // 警告
    if (0 < Config.get().getExecProgramMaxNum() && Config.get().getExecProgramMaxNum() < list.size()) {
      if (!MsgBox.okCancel("ファイルの実行", "選択されている数が設定された制限を越えています。\nこのまま続けてもよろしいですか？")) return;
    }
    
    // 実行
    for (FileInfo f : list) {
      exec(f, wndController);
    }
  }
 
  /**
   * 選択されているファイルを実行する
   */
  static public void exec(FileInfo sel, WindowController wndController) {
    // ディレクトリが選択されていたら、それを開く
    if (sel.isDirectory() && !sel.isArchiveRoot() && !sel.isArchiveChild()) {
      if (!Program.launch(sel.getAbsolutePath())) {
        MsgBox.error("実行不可", "ディレクトリを開けませんでした");
        return;
      }
      wndController.setStatusMessageTemporary(Config.OS_NAME + " の機能でディレクトリを開きます");
      return;
    }
    
    // アーカイブファイル内の子要素の場合は、親ファイルを対象にする
    if (sel.isArchiveChild()) {
      String msg = sel.getName() + "\n" + 
            "このファイルは、\n" +
            "  " + sel.getArchiveRoot() + "\n" +
            "の中の子要素です。\n" + 
            "親要素のファイルを開きますか？";
      if (!MsgBox.okCancel("アーカイブ内のファイルです", msg)) return;
      sel = sel.getArchiveRoot();
    }
    
    String[] spl = FileUtil.splitBasenameExt(sel.getName());
    
    // 拡張子がないとき
    if (spl[1].length() == 0) {
      launchWithConfirm(sel, wndController);
      return;
    }
    
    Program p = Program.findProgram(spl[1]);
    
    // 拡張子に対応するプログラムが見つからないとき
    if (p == null) {
      launchWithConfirm(sel, wndController);
      return;
    }
    
    // 拡張子が確認不要リストに入っていないなら、確認する
    if (!isNoConfirmExt(spl[1])) {
      if (!MsgBox.okCancel("ファイルの実行", p.getName() + " でファイルを開いてもよろしいですか？\n" + sel.getName())) return;
    }
    
    // 実行
    if (p.execute(sel.getAbsolutePath())) {
      wndController.setStatusMessageTemporary(p.getName() + " でファイルを開きます");
      return;
    }
    
    MsgBox.error("実行不可", sel.getName() + "\nは" + Config.OS_NAME + "の機能では実行できません。");
  }
  
  /*
   * 拡張子に対応していないファイルを実行する
   */
  static private void launchWithConfirm(FileInfo f, WindowController wndController) {
    // 確認
    String msg = f.getName() + "\n" + 
          "このファイルは" + Config.OS_NAME +"のアプリケーションに結びつけられていません。\n" +
          Config.OS_NAME + "の機能を使用して開いてもよろしいですか？";
    if (!MsgBox.okCancel("プログラムの実行", msg)) return;
    
    // 実行
    if (Program.launch(f.getAbsolutePath(), f.getParent().getAbsolutePath())) {
      wndController.setStatusMessageTemporary(Config.OS_NAME + "の機能でファイルを開きます");
      return;
    }
    
    MsgBox.error("実行不可", f.getName() + "\nは" + Config.OS_NAME + "の機能では実行できません。");
  }

  /**
   * プログラムを起動する
   */
  public static void launchProgram(ListPanel lp, int progIndex, WindowController wndController) {
    if (lp == null) return;

    // 起動対象プログラム
    Config conf = Config.get();
    if (conf.getLaunchPrograms().length <= progIndex) return;
    Config.LaunchProgram p = conf.getLaunchPrograms()[progIndex];

    // 引数準備
    ArrayList<String> args = new ArrayList<String>();
    for (String s : StringUtils.split(p.getCmdarray().trim(), "\r\n")) {
      if (s.equals(Config.LaunchProgram.REPLACE_SELECTED_FILES)) {
        int count = 0;
        for (int sel : lp.getSelection().getSelectedIndices()) {
          count++;
          if (0 <= p.getMaxFileArgs() && p.getMaxFileArgs() < count) break; 
          args.add(lp.getModel().getSortedList().get(sel).getAbsolutePath());
        }
      }
      else if (s.equals(Config.LaunchProgram.REPLACE_CURRENT_DIR)) {
        args.add(lp.getModel().getCurrentDir().getAbsolutePath());
      }
      else {
        args.add(s);
      }
    }
    
    // 確認
    StringBuilder msg = new StringBuilder();
    msg.append("[" + p.getName() + "]\n");
    for (int i = 0; i < args.size(); i++) {
      if (i > 10) {
        msg.append("…… (全" + args.size() + "引数)");
        break;
      }
      msg.append(args.get(i));
      msg.append("\n");
    }
    if (p.isConfirm()) {
      if (!MsgBox.okCancel("起動確認", "起動してもよろしいですか？\n\n" + msg.toString())) return;
    }
    
    // 起動
    try {
      Log.d(p.getName());
      Log.d(p.getCmdarray());
      Log.d("args = ", args);
      Log.d("dir = ", lp.getCurrentDir());
      Runtime.getRuntime().exec(args.toArray(new String[0]), null, new File(lp.getCurrentDir().getAbsolutePath()));
    }
    catch (IOException e) {
      Log.w(e);
      MsgBox.error("起動失敗", "プログラムの起動に失敗しました\n\n" + msg);
      return;
    }

    wndController.setStatusMessageTemporary(p.getName() + "でファイルを開きました");
  }
}
