package cvj2.file.fs;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Comparator;

import cvj2.file.fs.FileInfoComparator.Type;
import cvj2.file.util.FileUtil;
import cvj2.util.SizeCountable;
import cvj2.util.Util;

/**
 * ファイルの絶対パス情報。<br>
 * <pre>
 * 普通のファイルの実体は、FileInfoFs, FileInfoListFs
 * Zipファイルの実体は、FileInfoZip, FileInfoListZip, FileInfoZipRoot
 * PDFファイルの実体は、FileInfoPDF, FileInfoListPDF, FileInfoPDFRoot
 * </pre>
 */
public abstract class FileInfo implements Comparable<FileInfo>, SizeCountable {
  /** ファイル操作を synchronized 内部で行うときに使用する同期オブジェクト */ 
  public static final String FILE_IO_SYNC_OBJ = new String("FILE_IO_SYNC");
  /** ルートオブジェクト (ファイルシステムにはこれに対応するパスは存在しない) */
  public static final FileInfo ROOT = FileInfoRoot.getInstance();
  /** ファイル名ソート用 */
  public static Comparator<FileInfo> cmp = new FileInfoComparator(Type.UNICODE);
  
  /** 絶対パスを持ったFile */
  final protected File file;

  /** キャッシュ (これがあると、ファイルが多いときに少し速くなる) */
  protected String name;
  
  /** ファイル情報が初期化されているか */
  protected boolean isAttributeInitialized = false;
  /** 存在するか */
  protected boolean exists;
  /** ディレクトリか */
  protected boolean isDirectory;
  /** サイズ */
  protected long length;
  /** 更新日時 */
  protected long lastModified;

  /**
   * コンストラクタ
   */
  protected FileInfo(File file) {
    assert file != null;
    // ROOT以外の場合は、絶対パスで正規化されたパスでなければならない
    assert this.getClass() == FileInfoRoot.class || file.isAbsolute();
    this.file = file;
  }
  
  @Override
  public long getObjectBytes() {
    // File.path, File.prefixLength, attributes
    return Util.getObjectBytes(file.getPath()) + 4 + (4 + 4 + 4 + 8 + 8);
  }

  /**
   * 絶対パスを取得する
   */
  @Override
  public String toString() {
    return file.getPath();
  }

  /**
   * 絶対パスを取得する
   */
  public String getAbsolutePath() {
    return file.getPath();
  }
  
  /**
   * 比較
   */
  @Override
  public int compareTo(FileInfo o) {
    return cmp.compare(this, o);
  }

  /**
   * 比較
   */
  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (o == null || !(o instanceof FileInfo)) return false;
    FileInfo other = (FileInfo) o;
    return this.file.equals(other.file);
  }

  /** hashCode */
  @Override
  public int hashCode() {
    return file.hashCode();
  }

  /**
   * File取得
   */
  public File getFile() {
    return file;
  }

  /**
   * ファイルの名前取得
   * @return /usr/bin/aaa なら、"aaa"
   */
  public String getName() {
    if (this.name != null) return this.name;
    
    // Windowsのドライブとか、Unix の / の場合は空文字列になるのでチェックする
    String name = file.getName();
    if (name.length() <= 0) name = file.getAbsolutePath();
    this.name = name.intern();
    return name;
  }
  
  /**
   * ファイル情報が読み込み済みかどうか
   */
  public boolean isAttributeInitialized() {
    return isAttributeInitialized;
  }
  
  /**
   * ディレクトリか？
   * (Zip や PDF ファイルの場合は、true)
   */
  public boolean isDirectory() {
    initAttribute();
    return isDirectory;
  }

  /**
   * 最終更新
   */
  public long lastModified() {
    initAttribute();
    return lastModified;
  }

  /**
   * ファイル長
   */
  public long length() {
    initAttribute();
    return length;
  }

  /**
   * 存在するか？
   */
  public boolean exists() {
    initAttribute();
    return exists;
  }

  /**
   * アーカイブファイル内のファイルか?
   * (ファイル削除の可否を判定する)
   */
  public boolean isArchiveChild() {
    return false;
  }
  
  /**
   * アーカイブファイルか?
   * (isDirectory() が true でもファイルとして削除できる)
   */
  public boolean isArchiveRoot() {
    return false;
  }
  
  /**
   * 親ディレクトリ取得
   */
  public abstract FileInfo getParent();
  
  /**
   * 子リスト取得
   */
  public abstract FileInfoList getList();

  /**
   * ファイルのイメージを直接取得する (ファイルがPDFのページの場合など)
   * @param targetSize 取得したい画像サイズ
   * @return イメージの直接取得をサポートしていないときは、null
   */
  public BufferedImage getImage(Dimension targetSize) throws IOException {
    return null;
  }
  
  /**
   * ファイルのイメージを直接取得する (ファイルがPDFのページの場合など)
   * @param dpi dot per inch
   * @return イメージの直接取得をサポートしていないときは、null
   */
  public BufferedImage getImage(int dpi) throws IOException {
    return null;
  }
  
  /**
   * ファイルのサイズを取得する (ファイルがPDFのページの場合など)
   * @param dpi dot per inch
   * @return イメージの直接取得をサポートしていないときは、null
   */
  public Dimension getImageSize(int dpi) throws IOException {
    return null;
  }
  
  /**
   * 自分がアーカイブファイルの子要素の場合、アーカイブファイルを自身を取得する。<br>
   * {@link FileInfo#isArchiveChild()} が true の場合のみ有効
   */
  public abstract FileInfo getArchiveRoot();

  /**
   * ファイル属性を初期化する
   */
  protected abstract void initAttribute();
  
  /**
   * ファイル読み込み
   */
  public InputStream openStream() throws IOException {
    return null;
  }

  /**
   * 画像ファイルかどうか
   * (拡張子で判別する)
   */
  public abstract boolean isImageFile();
  
  /**
   * サムネイル画像を取得できる可能性があるか
   * (ディレクトリの場合は、内部の画像ファイルがあるかもしれない)
   */
  public boolean mayGetThumbnail() {
    return isDirectory() || isImageFile();
  }
  
  /**
   * 渡されたファイルが親かどうか調べる
   */
  public boolean isAncestor(FileInfo parent) {
    return FileUtil.isAncestor(parent.file, this.file);
  }
}
