package cvj2.file.fs;

import java.io.File;

/**
 * FileInfo の生成を管理する
 */
public class FileInfoFactory {
  /** ZIP機能を有効に */
  static public volatile boolean enableZipFolder = true;
  /** PDF機能を有効に */
  static public volatile boolean enablePdfFolder = true;
  
  private FileInfoFactory() {}
  
  /**
   * FileInfo から派生したオブジェクトの生成
   */
  static public FileInfo create(File f) {
    // ROOT は FileInfoFs の getParent() 内でしか取得できない 
    assert f != null;
    
    f = new File(f.getAbsolutePath());
    FileInfo fi = null;
    
    if (enableZipFolder) {
      // Zip ルート
      if (FileInfoZipManager.isZipFilename(f) && f.isFile()) return new FileInfoZipRoot(f);
      // Zip の子
      fi = FileInfoZip.newInstance(f);
      if (fi != null) return fi;
    }

    if (enablePdfFolder) {
      // PDF ルート
      if (FileInfoPDFManager.isPDFFilename(f) && f.isFile()) return new FileInfoPDFRoot(f);
      // PDF の子
      fi = FileInfoPDF.newInstance(f);
      if (fi != null) return fi;
    }
    
    // 通常ファイル
    return new FileInfoFs(f);  
  }
  
  /**
   * 既存のディレクトリの下に実在するファイルから FileInfo を生成する
   * 実在検査をしないので、{@link FileInfoFactory#create(File)} で生成するよりも速い
   * 
   * @param exisitFileOrDir ファイルシステム内に実在が確認されているファイルかディレクトリ
   * @return アーカイブファイルのルート(FileInfoZipRootなど)か、普通の FileInfoFs のどちらか
   */
  static FileInfoFs createChild(File exisitFileOrDir) {
    if (enableZipFolder) {
      // "*.zip" という名前の、ファイルかディレクトリの場合
      if (FileInfoZipManager.isZipFilename(exisitFileOrDir)) return new FileInfoZipRoot(exisitFileOrDir);
    }

    if (enablePdfFolder) {
      // "*.pdf" という名前の、ファイルかディレクトリの場合
      if (FileInfoPDFManager.isPDFFilename(exisitFileOrDir)) return new FileInfoPDFRoot(exisitFileOrDir);
    }
    
    // それ以外の場合
    return new FileInfoFs(exisitFileOrDir); 
  }
  
  /**
   * アーカイブファイルの名前かどうか調べる
   * アーカイブでない場合 (ディレクトリの場合や、存在しないファイルなど) も true になる場合がある
   * 
   * @param f ファイルかディレクトリ
   * @return 名前がアーカイブファイルの特徴に一致すれば、true
   */
  static boolean isArchiveFilename(File f) {
    if (FileInfoZipManager.isZipFilename(f)) return true;
    return false;
  }
}
