package cvj2.file.fs;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.List;

import org.apache.pdfbox.pdmodel.PDDocumentCatalog;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.common.PDRectangle;

import cvj2.file.fs.FileInfoPDFManager.PDFRef;
import cvj2.img.ImageUtil;

/**
 * PDFファイルのページ
 */
public class FileInfoPDF extends FileInfo {
  FileInfoPDFRoot pdfRoot;
  
  /**
   * インスタンス生成
   * @param f アーカイブ内のファイル (アーカイブファイルそのものはNG)
   * @return PDFでないなら、null
   */
  static FileInfoPDF newInstance(File f) {
    // 指定されたファイルがPdfファイルとして存在するかどうか確認する
    File parentPdf = f.getParentFile();
    if (parentPdf == null) return null;
    if (!parentPdf.isFile() || !FileInfoPDFManager.isPDFFilename(parentPdf)) return null;
    assert !parentPdf.equals(f);
    
    // Pdfファイル情報を取得する
    FileInfoPDFRoot PdfRoot = FileInfoPDFManager.getFileInfoPDFRoot(parentPdf);
    if (PdfRoot == null) return null;

    // Pdfファイル内に存在するか探す
    return FileInfoPDFManager.findFileInfo(PdfRoot, f);
  }

  /**
   * PDFファイルの解析中に、PDFManagerからのみ呼ばれる
   */
  FileInfoPDF(File f) {
    super(f);
  }

  @Override
  public boolean isArchiveChild() {
    return true;
  }
  
  @Override
  public FileInfo getParent() {
    assert pdfRoot != null;
    return pdfRoot;
  }

  @Override
  public FileInfoList getList() {
    throw new RuntimeException(this.toString());
  }

  @Override
  protected void initAttribute() {
    /* PDFManagerでセットされる */
  }

  @Override
  public boolean isImageFile() {
    return true;
  }
  
  @Override
  public FileInfo getArchiveRoot() {
    return getParent();
  }

  @Override
  public BufferedImage getImage(Dimension targetSize) throws IOException {
    BufferedImage[] retImg = { null };
    Dimension[] retSize = { null };
    if (!getImage(null, targetSize, retImg, retSize)) return null;
    return retImg[0];
  }
  
  @Override
  public BufferedImage getImage(int dpi) throws IOException {
    BufferedImage[] retImg = { null };
    Dimension[] retSize = { null };
    if (!getImage(dpi, null, retImg, retSize)) return null;
    return retImg[0];
  }

  @Override
  public Dimension getImageSize(int dpi) throws IOException {
    Dimension[] retSize = { null };
    if (!getImage(dpi, null, null, retSize)) return null;
    return retSize[0];
  }
  
  /**
   * 画像サイズと画像の取得
   * dpi か targetSize のどちらかを指定しなければならない
   *  
   * @param dpi 取得したい画像のDPI
   * @param targetSize 取得したい画像サイズ
   * @param retImg null でない値が渡されると、画像が戻る
   * @param retSize 出力される画像サイズ
   * @return retImg, retSize がセットできたら、true
   */
  private boolean getImage(Integer dpi, Dimension targetSize, BufferedImage[] retImg, Dimension[] retSize) throws IOException {
    assert dpi != null || targetSize != null;
    assert retSize != null;
    
    synchronized (FILE_IO_SYNC_OBJ) {
      int i = Integer.parseInt(getName());
      PDFRef pdf = null;
      try {
        // PDFファイル取得
        pdf = FileInfoPDFManager.getPdfAndRef(super.file.getParentFile());
        assert pdf.getRefCount() > 0;
        PDDocumentCatalog catalog = pdf.get().getDocumentCatalog();
        List<?> pages = catalog.getAllPages();
        // PDFページ取得
        PDPage page = (PDPage) pages.get(i);
        // ページのサイズ 1/72 inch 単位
        PDRectangle rect = page.findCropBox();
        // ページのサイズ 1/1000 inch 単位
        Dimension sizeInch1000 = new Dimension((int)(rect.getWidth() / 72.0 * 1000), (int)(rect.getHeight() / 72.0 * 1000));
        
        if (dpi != null) {
          retSize[0] = new Dimension(dpi * sizeInch1000.width / 1000, dpi * sizeInch1000.height / 1000);
        }
        else {
          Dimension targetSize100 = new Dimension(targetSize.width * 100, targetSize.height * 100);
          retSize[0] = ImageUtil.calcImageFitSize(sizeInch1000, targetSize);
          Dimension size100 = ImageUtil.calcImageFitSize(sizeInch1000, targetSize100);
          dpi = (int)(((double)size100.width / 100) / ((double)sizeInch1000.width / 1000));
        }
        if (retImg == null) return true;
        
        retImg[0] = page.convertToImage(BufferedImage.TYPE_INT_RGB, dpi);
        return true;
      }
      finally {
        if (pdf != null) pdf.unref();
      }
    }
  }
}
