package cvj2.file.fs;

import java.io.File;
import java.io.IOException;
import java.util.List;

import org.apache.pdfbox.exceptions.CryptographyException;
import org.apache.pdfbox.exceptions.InvalidPasswordException;
import org.apache.pdfbox.pdmodel.PDDocument;

import cvj2.file.util.FileUtil;
import cvj2.util.CacheMap;
import cvj2.util.CacheMap.RemoveCallback;
import cvj2.util.Log;

/**
 * FileInfoPDF の構築などを行う管理クラス
 */
public class FileInfoPDFManager {
  private static final String PAGE_NAME_FORMAT = "%08d";
  
  // PDFファイル情報ツリーの解析結果のキャッシュ
  public static final CacheMap<File, FileInfoPDFRoot> infoCache = new CacheMap<File, FileInfoPDFRoot>(32);

  // オープンしたPDFファイルのキャッシュ
  public static final CacheMap<File, PDFRef> pdfCache = new CacheMap<File, PDFRef>(3, new PDDocRemoveCallback(), true);
  private static class PDDocRemoveCallback implements RemoveCallback<File, PDFRef> {
    @Override public void removed(File file, PDFRef doc) {
      Log.start(file, doc);
      try {
        doc.unref();
      }
      catch (IOException e) {
        throw new RuntimeException(e);
      }
    }
  }
  
  /**
   * ロード済みPDFの参照オブジェクト。<br>
   * 参照が 0 になると、close() が呼ばれる
   */
  static class PDFRef {
    private PDDocument pdf;
    private int refCount = 0;
    
    /**
     * コンストラクタ
     */
    public PDFRef(PDDocument pdf) {
      this.pdf = pdf;
    }

    /**
     * 参照を増やす
     */
    synchronized public boolean ref() {
      if (pdf == null) return false;
      refCount++;
      return true;
    }
    
    /**
     * 参照を減らす
     */
    synchronized public void unref() throws IOException {
      assert refCount > 0;
      refCount--;
      if (refCount > 0) return;
      pdf.close();
      pdf = null;
    }
    
    /**
     * 参照数を取得
     */
    synchronized int getRefCount() {
      return refCount;
    }
    
    /**
     * オブジェクトを取得 (必ず ref() を読んでから取得すること)
     */
    synchronized PDDocument get() {
      assert refCount > 0;
      return pdf;
    }
  }
  
  /**
   * PDFファイル情報を作成する
   * @file pdfファイル 
   * @return pdfファイルのルートのFileInfo
   */
  static FileInfoPDFRoot getFileInfoPDFRoot(File file) {
    Log.start(file);
    assert file != null;

    // キャッシュにあるなら、それを返す
    FileInfoPDFRoot fi = infoCache.get(file);
    if (fi != null) return fi;
    
    PDFRef pdf = null;
    try {
      pdf = getPdfAndRef(file);
      assert pdf.getRefCount() > 0;
      List<?> pages = pdf.get().getDocumentCatalog().getAllPages();
      FileInfoPDF[] list = new FileInfoPDF[pages.size()];
      FileInfoPDFRoot root = new FileInfoPDFRoot(file);
      for (int i = 0; i < pages.size(); i++) {
        list[i] = new FileInfoPDF(new File(file, String.format(PAGE_NAME_FORMAT, i)));
        list[i].pdfRoot = root;
        list[i].isAttributeInitialized = true;
      }
      root.pdfList = new FileInfoPDFList(list);
      return root;
    }
    catch (IOException e) {
      Log.w(e);
      // 壊れたPDF？
      return null;
    }
    finally {
      try {
        if (pdf != null) pdf.unref();
      }
      catch (IOException e) {
        throw new RuntimeException(e);
      }
    }
  }

  /**
   * PDFを取得する
   * @param file 読み込み対象PDFファイル
   * @return 参照が +1 されたオブジェクト (unref() するのは呼出側の責任)
   */
  static PDFRef getPdfAndRef(File file) throws IOException {
    // キャッシュにあるオブジェクトが有効なら、それをかえす
    PDFRef ref = pdfCache.get(file);
    if (ref != null && ref.ref()) return ref;
    
    synchronized (FileInfo.FILE_IO_SYNC_OBJ) {
      // ファイルをロードする
      PDDocument pdf = PDDocument.load(file);
      if (pdf.isEncrypted()) {
        try {
          pdf.decrypt("");
        } 
        catch (CryptographyException e) {
          Log.e(e);
        } 
        catch (InvalidPasswordException e) {
          Log.e(e);
        }
      }
      
      // 参照オブジェクトを作成する
      ref = new PDFRef(pdf);
      if (!ref.ref()) throw new RuntimeException("bug"); // キャッシュに入れるときは参照を増やす
      if (!ref.ref()) throw new RuntimeException("bug"); // 参照が有効なオブジェクトを return する
      pdfCache.put(file, ref);
      return ref;
    }
  }
  
  /**
   * PDF ファイル内の子ファイルを探す
   */
  static FileInfoPDF findFileInfo(FileInfoPDFRoot pdfRoot, File f) {
    FileInfoPDFList list = (FileInfoPDFList) pdfRoot.getList();
    int i = list.find(new FileInfoPDF(f));
    if (i < 0) return null;
    return list.get(i);
  }

  /**
   * PDF のファイル名かどうか
   */
  static boolean isPDFFilename(File f) {
    String ext = FileUtil.splitBasenameExt(f)[1];
    if (ext.length() == 0) return false;
    return ext.equalsIgnoreCase(".pdf");
  }
}
