package cvj2.file.fs;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;

import org.apache.commons.io.IOUtils;

import cvj2.file.util.FileUtil;
import cvj2.file.util.ZipFile;
import cvj2.img.ImageUtil;

/**
 * Zip中のファイルやディレクトリ。
 * 1つのZipファイル内のFileInfoZipは、全て互いに参照しあっている。
 * よって、GCで回収されるタイミングは一致している。
 * 作成された直後に回収されないように、ZipManager内にキャッシュされる。
 */
public class FileInfoZip extends FileInfo {
  // 自分の属するZipファイル
  File zipRoot;
  
  // 親ファイル (親がルートZIPの時は、FileInfoZipRoot)
  FileInfo parent;

  // Zipファイル内のパス
  String zipEntryName;
  
  // 子リスト
  FileInfoZipList list;
  
  /**
   * インスタンス生成
   * @param f アーカイブ内のファイル (アーカイブファイルそのものはNG)
   * @return Zipでないなら、null
   */
  static FileInfoZip newInstance(File f) {
    // 指定されたファイルがZIPファイルとして存在するかどうか確認する
    File parentZip = FileUtil.getExisitParentInFs(f);
    if (parentZip == null) return null;
    if (!parentZip.isFile() || !FileInfoZipManager.isZipFilename(parentZip)) return null;
    assert !parentZip.equals(f);
    
    // ZIPファイル情報を取得する
    FileInfoZipRoot zipRoot = FileInfoZipManager.getFileInfoZipRoot(parentZip);
    if (zipRoot == null) return null;

    // ZIPファイル内に存在するか探す
    return FileInfoZipManager.findFileInfo(zipRoot, f);
  }

  /**
   * Zipファイルの解析中に、ZipManagerからのみ呼ばれる
   */
  FileInfoZip(File f) {
    super(f);
  }
  
  @Override
  public boolean isArchiveChild() {
    return true;
  }

  @Override
  public FileInfo getParent() {
    assert parent != null;
    return parent;
  }

  @Override
  public FileInfoZipList getList() {
    assert list != null;
    return list;
  }

  @Override
  protected void initAttribute() {
    /* ZipManagerでセットされる */
  }

  @Override
  public boolean isImageFile() {
    return !isDirectory() && ImageUtil.isImgFileName(getName());
  }

  @Override
  public InputStream openStream() throws IOException {
    synchronized (FileInfo.FILE_IO_SYNC_OBJ) {
      ZipFile zip = ZipFile.newInstance(zipRoot, FileInfoZipManager.getEncoding());
      ZipEntry ze = zip.getEntry(zipEntryName);
      if (ze == null) throw new ZipException(this.toString() + ": " + zipRoot.getAbsolutePath() + "#" + zipEntryName + " is not found");
      return new ZipEntryInputStream(new BufferedInputStream(zip.getInputStream(ze)), zip);
    }
  }
  
  /*
   * ZipEntry のストリームを閉じるときに、ZipFile を閉じるためのクラス
   */
  static private class ZipEntryInputStream extends FilterInputStream {
    private ZipFile zip;
    protected ZipEntryInputStream(InputStream in, ZipFile zip) {
      super(in);
      this.zip = zip;
    }
    @Override
    public void close() throws IOException {
      try {
        super.close();
      }
      finally {
        IOUtils.closeQuietly(zip);
      }
    }
  }

  @Override
  public FileInfo getArchiveRoot() {
    return FileInfoFactory.create(zipRoot);
  }
}
