package cvj2.file.fs;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.zip.ZipEntry;

import org.apache.commons.io.FilenameUtils;

import cvj2.file.util.FileUtil;
import cvj2.file.util.ZipFile;
import cvj2.util.CacheMap;
import cvj2.util.Log;

/**
 * FileInfoZip の構築などを行う管理クラス
 */
public class FileInfoZipManager {
  // Zipファイル内のファイル名エンコード
  public static String encoding = "MS932";
  // Zipファイル情報ツリーの解析結果のキャッシュ
  public static final CacheMap<File, FileInfoZipRoot> cache = new CacheMap<File, FileInfoZipRoot>(8);
  
  /**
   * Zipファイル情報を作成する
   * @file zipファイル 
   * @return zipファイルのルートのFileInfo
   */
  static FileInfoZipRoot getFileInfoZipRoot(File file) {
    Log.start(file);
    assert file != null;
    
    // キャッシュにあるなら、それを返す
    FileInfoZipRoot fi = cache.get(file);
    if (fi != null) return fi;
    
    // ファイル情報ツリーを構築する
    synchronized (FileInfo.FILE_IO_SYNC_OBJ) {
      try {
        fi = createFileInfoZipRoot(file);
        cache.put(file, fi);
        return fi;
      } 
      catch (IOException e) {
        Log.w(file, ": ", e.getMessage());
        // 壊れたZIP？
        return null;
      }
    }
  }
  
  /*
   * ファイル情報ツリーを構築する
   */
  private static FileInfoZipRoot createFileInfoZipRoot(File file) throws IOException {
    // Zipファイルから全ZipEntryを取得する
    Map<String, ZipEntry> entryMap = getZipEntries(file);
    // 子 => 親 マップ
    Map<String, String> parentMap = new TreeMap<String, String>();
    // 親 => 子 マップ
    Map<String, List<String>> childsMap = new TreeMap<String, List<String>>();

    // 親子関係マップ構築
    for (String name : entryMap.keySet()) {
      if (name.length() == 0) continue;
      String[] parentChild = splitParentChild(name);
      parentMap.put(name, parentChild[0]);

      List<String> childs = childsMap.get(parentChild[0]);
      if (childs == null) {
        childs = new ArrayList<String>();
        childsMap.put(parentChild[0], childs);
      }
      childs.add(name);
    }

    // ツリー構造作成 (Zipルートから再帰的に)
    FileInfoZipRoot zipRoot = new FileInfoZipRoot(file);
    zipRoot.zipList = createFileInfoZipList("", zipRoot, file, entryMap, parentMap, childsMap);
    return zipRoot;
  }
  
  /*
   * Zipファイルから全ZipEntryを取得する
   * @return Zipファイル内の全ファイル => ZipEntry
   *         Zipファイル内の全ディレクトリ => null or ZipEntry
   */
  private static Map<String, ZipEntry> getZipEntries(File file) throws IOException {
    Map<String, ZipEntry> map = new HashMap<String, ZipEntry>();
    ZipFile zip = ZipFile.newInstance(file, encoding);
    try {
      Enumeration<? extends ZipEntry> zeEnum = zip.entries();
      while (zeEnum.hasMoreElements()) {
        ZipEntry ze = zeEnum.nextElement();
        String name = toSimpleName(ze.getName());
        map.put(name, ze);
        
        // ZipEntryの存在しないディレクトリもnullで登録する
        while (true) {
          name = splitParentChild(name)[0];
          if (map.get(name) == null) map.put(name, null);
          if (name.length() == 0) break;
        }
      }
      return map;
    }
    finally {
      zip.close();
    }
  }

  /**
   * FileInfoZipList を生成する
   * @param parentName 生成リストの親の zipEntry 名
   * @param parent 生成リストの親 (Zipルートの場合は、null)
   * @param rootFile Zipファイル本体
   */
  private static FileInfoZipList createFileInfoZipList(
      String parentName, FileInfo parent, File rootFile,
      Map<String, ZipEntry> entryMap, Map<String, String> parentMap, Map<String, List<String>> childsMap) 
  {
    if (!parent.isDirectory()) return null;
    List<String> childNames = childsMap.get(parentName);
    if (childNames == null) childNames = new ArrayList<String>();
    ArrayList<FileInfoZip> childInfos = new ArrayList<FileInfoZip>();
    for (String childName : childNames) {
      ZipEntry ze = entryMap.get(childName);
      File childFile = new File(FilenameUtils.normalize(new File(rootFile, childName).getPath()));
      FileInfoZip child = new FileInfoZip(childFile);
      child.isAttributeInitialized = true;
      child.exists = true;
      child.isDirectory = ze == null ? true : ze.isDirectory() || childsMap.containsKey(childName);
      child.length = ze == null ? 0 : ze.getSize();
      child.lastModified = ze == null ? 0 : ze.getTime();
      child.zipRoot = rootFile;
      child.parent = parent;
      child.zipEntryName = childName;
      child.list = createFileInfoZipList(childName, child, rootFile, entryMap, parentMap, childsMap);
      childInfos.add(child);
    }
    return new FileInfoZipList(childInfos.toArray(new FileInfoZip[0]));
  }
  
  // 先頭と末尾の '/' を取り除く
  static private String toSimpleName(String name) {
    if (name.endsWith("/")) name = name.substring(0, name.length() - 1);
    if (name.startsWith("/")) name = name.substring(1, name.length() - 1);
    return name;
  }

  /**
   * 自分と親ディレクトリのパスを分離する。
   * @return [親, 子] で返す (親がないときは、"")
   */
  private static String[] splitParentChild(String name) {
    assert name.length() > 0;
    name = toSimpleName(name);
    int index = name.lastIndexOf('/');
    if (index < 0) return new String[] { "", name };
    return new String[] { name.substring(0, index), name.substring(index + 1, name.length()) };
  }

  /**
   * ZIP のファイル名かどうか
   */
  static boolean isZipFilename(File f) {
    String ext = FileUtil.splitBasenameExt(f)[1];
    if (ext.length() == 0) return false;
    return ext.equalsIgnoreCase(".zip");
  }

  /**
   * Zip ファイル内の子ファイルを探す
   */
  static FileInfoZip findFileInfo(FileInfoZipRoot fileInfo, File childFile) {
    assert !fileInfo.getFile().equals(childFile);
    FileInfoList list = fileInfo.getList();
    assert list instanceof FileInfoZipList;
    return findFileInfo((FileInfoZipList)list, childFile);
  }
  static private FileInfoZip findFileInfo(FileInfoZipList list, File f) {
    for (int i = 0; i < list.size(); i++) {
      FileInfoZip child = list.get(i);
      if (!FileUtil.isAncestor(child.getFile(), f)) continue;
      child = findFileInfo(child, f);
      if (child != null) return child;
    }
    return null;
  }
  static private FileInfoZip findFileInfo(FileInfoZip fileInfo, File f) {
    if (fileInfo.getFile().equals(f)) return fileInfo;
    if (!fileInfo.isDirectory()) return null;
    return findFileInfo(fileInfo.getList(), f);
  }

  /**
   * エンコードの設定
   */
  public static void setEncoding(String encoding) {
    FileInfoZipManager.encoding = encoding;
  }
  
  /**
   * エンコードの取得
   */
  static String getEncoding() {
    return encoding;
  }
}
