package cvj2.file.util;

import java.util.Date;
import java.util.LinkedList;

import org.apache.commons.lang3.time.StopWatch;

/**
 * データ転送量 (Bytes / sec) を計算するクラス
 */
public class BPSCout {
  private StopWatch sw = new StopWatch();
  private int splitMilliSec;
  private int recordMax;
  private LinkedList<Record> records = new LinkedList<BPSCout.Record>();
  long currentBytes = 0;

  private static class Record {
    long bytes;
    long time = new Date().getTime();
    public Record(long bytes) {
      this.bytes = bytes;
    }
  }
  
  
  /**
   * コンストラクタ
   * 転送量の計算は、解像度 × 蓄積長 の範囲で行われる。
   * 例) 100(ms) × 10 = 1(sec) なので、過去1秒間の平均転送量が計算される
   * 
   * @param splitMilliSec データ解像度 (ミリ秒)
   * @param recordMax データを蓄積する長さ
   */
  public BPSCout(int splitMilliSec, int recordMax) {
    this.splitMilliSec = splitMilliSec;
    this.recordMax = recordMax;
    sw.start();
  }
  
  /**
   * 転送量を加える
   */
  public void add(long bytes) {
    currentBytes += bytes;
    
    // 時間が経っていなかったら、戻る
    if (sw.getTime() < splitMilliSec) return;

    // 記録を更新する
    records.add(new Record(currentBytes));
    while (records.size() > recordMax) records.removeFirst();
    currentBytes = 0;
    sw.reset();
    sw.start();
  }
  
  /**
   * 秒間転送量を取得する
   */
  public long getBPS() {
    long total = 0;
    long start = -1;
    long end = -1;
    for (Record r : records) {
      if (start < 0) start = r.time;
      end = r.time;
      total += r.bytes;
    }
    if (start == end) return 0;
    return total * 1000 / (end - start);
  }
}
