package cvj2.file.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;

import cvj2.file.util.FileOperationCopy.Notify.DirAnswer;
import cvj2.file.util.FileOperationCopy.Notify.FileAnswer;
import cvj2.util.Log;

/**
 * ファイルコピー。<br>
 * (再帰的に処理する)<br>
 * (シンボリックリンクはコピーできない)
 */
public class FileOperationCopy {
  /** 重複ファイルのリネーム規則 */
  public static String RENAME_PREFIX_1 = "コピー ";
  public static String RENAME_PREFIX_2_FORMAT = "(%d) ";
  public static String RENAME_PREFIX_3 = "～ ";

  private Notify notify;
  private FileOperationCanceller cancel;
  
  /**
   * 処理の通知と問い合わせ
   */
  public interface Notify {
    /** 
     * 処理通知 (ファイルコピーの途中) 
     */
    public void fileProgress(File f, long copySize);

    /** 
     * 処理通知 (ファイルコピー完了) 
     */
    public void fileFinish(File f, long size);

    /** ファイル処理確認への応答 */
    public enum FileAnswer {
      Cancel,
      Rename,
      OverWrite,
      Ignore,
    }
    /** ディレクトリ処理確認への応答 */
    public enum DirAnswer {
      Cancel,
      Rename,
      Merge,
      Ignore,
    }
    
    /**
     * 既存ファイル確認
     */
    public FileAnswer confirmExistFile(File file);
    
    /**
     * 既存ディレクトリ確認
     */
    public DirAnswer confirmExistDir(File dir);
  }
  
  private static class NullNotify implements Notify {
    long count;
    @Override public void fileProgress(File f, long copySize) { count++; };
    @Override public void fileFinish(File f, long size) { count += size; }
    @Override public FileAnswer confirmExistFile(File file) { return FileAnswer.Cancel; }
    @Override public DirAnswer confirmExistDir(File dir) { return DirAnswer.Cancel; }
  }
  
  /**
   * コンストラクタ
   */
  public FileOperationCopy(Notify notify, FileOperationCanceller cancel) {
    assert notify != null;
    this.notify = notify;
    this.cancel = cancel;
  }

  /**
   * ファイルをコピーする
   * @return 途中でキャンセルされたら、false;
   */
  public boolean copy(File[] files, File destDir) throws IOException {
    Log.start(files.length, destDir);
    
    if (!destDir.isDirectory()) throw new RuntimeException(destDir + " はディレクトリではありません。");
    
    for (File f : files) {
      if (!exec(f, destDir)) return false;
    }
    return true;
  }

  /**
   * コピーできるかどうか試してみる (実際にコピーはしない)
   * コピーできないファイルがあれば、例外を投げる
   * @return コピーするバイト数 + ファイル数
   *          途中でキャンセルされたら、-1;
   */
  public long tryCopy(File[] files, File destDir) throws IOException {
    Log.start(files.length, destDir);
    
    Notify orgNotify = notify;
    notify = new NullNotify();
    try {
      if (!copy(files, destDir)) return -1;
      return ((NullNotify)notify).count;
    }
    finally {
      notify = orgNotify;
    }
  }

  /*            ┌── (再帰)───┐
   *            ↓                 │
   * copy ─→ exec ─┬→ execDir ┘
   *                  │
   *                  └→ execFile ─→ copyFile  
   */

  /*
   * ファイル or ディレクトリの処理
   */
  private boolean exec(File f, File destDir) throws IOException {
    if (FileUtils.isSymlink(f)) {
      throw new FileOperationException(f + " はシンボリックリンクです");
    }
    else if (f.isDirectory()) {
      return execDir(f, destDir);
    }
    else if (f.isFile()) {
      return execFile(f, destDir);
    }
    else {
      throw new FileOperationException(f + " をコピーできません");
    }
  }

  /*
   * ディレクトリの処理
   */
  private boolean execDir(File dir, File destDir) throws IOException {
    if (cancel.isCancel()) return false;

    // 子リスト取得
    File[] list = dir.listFiles();
    Arrays.sort(list);
    if (list == null) throw new FileOperationException(dir + " の中を見ることができません");
    
    File newDir = new File(destDir, dir.getName());
    if (!(notify instanceof NullNotify)) {
      // コピー先ディレクトリ作成
      if (newDir.exists()) {
        newDir = execExistDir(dir, destDir, newDir);
        if (newDir == null) return false;
        if (newDir == IGNORE) return true;
      }
      else {
        if (!newDir.mkdir()) throw new FileOperationException(newDir + " を作成できません");
      }
    }
    
    // 再帰処理
    for (File f : list) {
      if (!exec(f, newDir)) return false;
    }
    
    // タイムスタンプ更新
    if (!(notify instanceof NullNotify)) {
      if (!newDir.setLastModified(dir.lastModified())) throw new FileOperationException(newDir + " の日付更新に失敗しました");
    }
    
    return true;
  }

  private static final File IGNORE = new File("/_IGNORE_");
  /**
   * 重複ディレクトリの扱い
   * @return キャンセルの場合: null
   *          マージの場合: マージ対象ディレクトリ
   *          リネームの場合: 新しい名前で作成されたディレクトリ
   *          無視の場合: IGNORE
   */
  private File execExistDir(File dir, File destDir, File newDir) throws IOException {
    DirAnswer ans = notify.confirmExistDir(newDir);
    if (ans == DirAnswer.Cancel) {
      // キャンセル
      return null;
    }
    else if (ans == DirAnswer.Merge) {
      // マージ
      return newDir;
    }
    else if (ans == DirAnswer.Rename) {
      // リネーム
      newDir = getNewName(destDir, dir.getName());
      if (!newDir.mkdir()) throw new FileOperationException(newDir + " を作成できません");
      return newDir;
    }
    else if (ans == DirAnswer.Ignore) {
      // 無視
      return IGNORE;
    }
    throw new RuntimeException("" + ans);
  }

  /*
   * ファイルの処理
   */
  private boolean execFile(File file, File destDir) throws IOException {
    // 通知
    if (notify != null) notify.fileProgress(file, 0);

    if (cancel.isCancel()) return false;

    File newFile = new File(destDir, file.getName());
    if (!(notify instanceof NullNotify)) {
      // 存在チェック
      if (newFile.exists()) {
        newFile = execExistFile(file, destDir, newFile);
        if (newFile == null) return false;
        if (newFile == IGNORE) {
          notify.fileProgress(file, file.length());
          notify.fileFinish(file, file.length());
          return true;
        }
      }

      // コピーする
      if (!copyFile(file, newFile)) return false;
    }
    
    notify.fileFinish(file, file.length());
    
    return true;
  }

  /**
   * 重複ファイルの扱い
   * @return キャンセルの場合: null
   *          上書きの場合: 上書き対象のファイル
   *          リネームの場合: 新しい名前でコピーする先のファイル
   *          無視の場合: IGNORE
   */
  private File execExistFile(File file, File destDir, File newFile) {
    FileAnswer ans = notify.confirmExistFile(newFile);
    if (ans == FileAnswer.Cancel) {
      // キャンセル
      return null;
    }
    else if (ans == FileAnswer.OverWrite) {
      // 上書き
      return newFile;
    }
    else if (ans == FileAnswer.Rename) {
      // リネーム
      return getNewName(destDir, file.getName());
    }
    else if (ans == FileAnswer.Ignore) {
      // 無視
      return IGNORE;
    }
    
    throw new RuntimeException("" + ans);
  }
  
  /**
   * 重複しない新しい名前を取得する
   */
  private File getNewName(File dir, String filename) {
    File newFile = new File(dir, filename);
    int x = 0;
    while (newFile.exists()) {
      x++;
      if (x == 1) {
        newFile = new File(dir, RENAME_PREFIX_1 + RENAME_PREFIX_3 + filename);
      }
      else {
        String f = RENAME_PREFIX_1 + RENAME_PREFIX_2_FORMAT + RENAME_PREFIX_3;
        newFile = new File(dir, String.format(f, x) + filename);
      }
    }
    return newFile;
  }

  /*
   * ファイルのコピー
   */
  private boolean copyFile(File orgFile, File newFile) throws IOException {
    // 既に存在する場合は、一時的に別の名前でコピーする
    File renameOnExit = null;
    if (newFile.exists()) {
      renameOnExit = newFile;
      newFile = File.createTempFile(newFile.getName() + "_tmp_", null, newFile.getParentFile());
    }
    
    byte[] buff = new byte[64 * 1024];
    int read;
    BufferedInputStream in = new BufferedInputStream(new FileInputStream(orgFile));
    BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(newFile));
    try {
      while (true) {
        if (cancel.isCancel()) break;

        // 読み書き
        read = in.read(buff);
        if (notify != null) notify.fileProgress(orgFile, Math.max(0, read));
        if (read < 0) break;
        out.write(buff, 0, read);
      }
    }
    finally {
      if (in != null) IOUtils.closeQuietly(in);
      if (out != null) out.close();
      
      if (cancel.isCancel()) {
        if (!newFile.delete()) throw new FileOperationException("コピー中のファイルの削除に失敗しました: " + newFile);
        return false;
      }
      else if (renameOnExit != null) {
        renameOnExit.delete();
        if (!newFile.renameTo(renameOnExit)) throw new FileOperationException(newFile + " のリネームに失敗しました");
        newFile = renameOnExit;
      }
    }
    
    // タイムスタンプ更新
    if (!newFile.setLastModified(orgFile.lastModified())) throw new FileOperationException(newFile + " の日付更新に失敗しました");
    
    return true;
  }
}
