package cvj2.file.util;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.io.FileUtils;

import cvj2.util.Log;

/**
 * ファイル削除。<br>
 *  (通知と書き込み禁止時の問い合わせあり)
 */
public class FileOperationDelete {
  private FileOperationCanceller cancel;
  private Notify notify;
  
  /**
   * 処理の通知と問い合わせ
   */
  public interface Notify {
    /** 
     * ディレクトリの処理開始通知 
     */
    public void progressDir(File dir);
    
    /** 
     * ファイルの処理通知 
     */
    public void progressFile(File f);

    /**
     * 読み込み専用ファイルorディレクトリを削除してもいいかどうかの確認
     * @param f 削除対象
     * @return false を返すと処理を中断する
     */
    public boolean askDelete(File f);
  }
  
  /*
   * ファイル数計測モード用の通知クラス
   * (このクラスのオブジェクトがセットされていると、実際の削除は行わない)
   */
  private class CountNotify implements Notify {
    int count = 0;
    @Override public void progressDir(File dir) { count++; }
    @Override public void progressFile(File f) { count++; }
    @Override public boolean askDelete(File f) { return false; }
  }
  
  /**
   * コンストラクタ
   */
  public FileOperationDelete(Notify notify, FileOperationCanceller cancel) {
    assert notify != null;
    this.notify = notify;
    this.cancel = cancel;
  }
  
  /**
   * ファイルを削除する (ディレクトリがあれば再帰的に)
   * @return 途中でキャンセルされたら、false;
   */
  public boolean delete(File[] files) throws IOException {
    Log.start(files.length);
    
    return delete(null, files);
  }
  
  /**
   * ファイル処理数を数える
   * @return delete() を呼んだときに Notify で通知される数
   */
  public int count(File[] files) throws IOException {
    Log.start(files.length);

    Notify oldNotify = notify;
    CountNotify countNotify = new CountNotify();
    notify = countNotify;
    try {
      if (!delete(null, files)) return -1;
      return countNotify.count;
    }
    finally {
      notify = oldNotify;
    }
  }
  
  /*
   * ディレクトリ削除
   * @param null あるいは、実在するディレクトリ(シンボリックリンクではない)
   * @param parent が null でないときは、その子要素
   */
  private boolean delete(File parent, File[] childs) throws IOException {
    // 権限のないディレクトリの場合は、null になる
    if (childs == null) childs = new File[0];
    
    Arrays.sort(childs);
    Set<File> finished = new HashSet<File>();
    
    // 親ディレクトリが書き込み禁止だと、子ファイルが削除できないのであらかじめ確認する (Linuxの場合)
    if (parent != null && !parent.canWrite()) {
      // 計測モードのときは何もしない
      if (!(notify instanceof CountNotify)) {
        // 問い合わせる
        if (!notify.askDelete(parent)) return false;
        parent.setWritable(true);
      }
    }
    
    // ファイルを削除する
    for (File f : childs) {
      if (!f.isFile()) continue;
      notify.progressFile(f);
      if (!deleteOneObject(f, true)) return false;
      finished.add(f);
    }

    // 子ディレクトリを削除する
    for (File f : childs) {
      if (!f.isDirectory()) continue;
      notify.progressDir(f);

      if (FileUtils.isSymlink(f)) {
        // シンボリックリンクの場合は、再帰削除でなく普通の削除
        if (!deleteOneObject(f, false)) return false;
      }
      else {
        // 再帰削除
        if (!delete(f, f.listFiles())) return false;
      }
      
      finished.add(f);
    }

    // 壊れたシンボリックリンクも削除する
    for (File f : childs) {
      if (finished.contains(f)) continue;
      if (!deleteOneObject(f, false)) return false;
      finished.add(f);
    }
    assert finished.size() == childs.length;
    
    // 親ディレクトリを削除する
    if (parent == null) return true;
    return deleteOneObject(parent, true);
  }

  /*
   * 空のディレクトリ or シンボリックリンク or ファイル を削除する
   * @param askReadOnlyDelete シンボリックリンクは書き込み可にならないので、問い合わせない
   */
  private boolean deleteOneObject(File obj, boolean askReadOnlyDelete) throws IOException {
    if (cancel.isCancel()) return false;

    // 計測モードのときは何もしない
    if (notify instanceof CountNotify) return true;
    
    // 読み取り専用になっているときは、問い合わせる
    if (askReadOnlyDelete && !obj.canWrite()) {
      if (!notify.askDelete(obj)) return false;
      obj.setWritable(true);
    }
    
    // 削除する
    if (!obj.delete()) throw new FileOperationException(obj + ": 削除できません");
    
    return true;
  }
}
