package cvj2.file.util;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import cvj2.util.Log;

/**
 * ファイル移動。<br>
 * (再帰的には処理しない)<br>
 * (別パーティションには移動できない)
 */
public class FileOperationMove {
  private FileOperationCanceller cancel;
  private Notify notify;
  
  /**
   * 処理の通知と問い合わせ
   */
  public interface Notify {
    /** 
     * 処理通知 
     */
    public void progress(File f);
    
    /**
     * 上書き確認
     */
    public boolean confirmOverwrite(File f);
  }
  
  private static class NullNotify implements Notify {
    @Override public void progress(File f) { }
    @Override public boolean confirmOverwrite(File f) { return true; }
  }
  
  /**
   * コンストラクタ
   */
  public FileOperationMove(Notify notify, FileOperationCanceller cancel) {
    assert notify != null;
    this.notify = notify;
    this.cancel = cancel;
  }
  
  /**
   * ファイルを移動する
   * @return 途中でキャンセルされたら、false;
   */
  public boolean move(File[] files, File destDir) throws IOException {
    Log.start(files.length, destDir);

    if (!destDir.isDirectory()) throw new FileOperationException(destDir + " はディレクトリではありません。");
    
    for (File f : files) {
      if (!moveOneObject(f, destDir)) return false;
    }
    return true;
  }

  /**
   * 移動できるかどうか試してみる (実際に移動はしない)
   * 移動できないファイルがあれば、例外を投げる
   * @return 途中でキャンセルされたら、false;
   */
  public boolean tryMove(File[] files, File destDir) throws IOException {
    Log.start(files.length, destDir);
    
    Notify orgNotify = notify;
    notify = new NullNotify();
    try {
      return checkSamePartition(files, destDir);
    }
    finally {
      notify = orgNotify;
    }
  }
  
  /*
   * 同じパーティションかどうか確かめる
   */
  private boolean checkSamePartition(File[] files, File destDir) throws IOException {
    // 移動元ファイルの親ディレクトリをまとめる
    Set<File> set = new HashSet<File>();
    for (File f : files) {
      set.add(f.getParentFile());
    }
    // それぞれに対してチェックする
    for (File f : set) {
      if (cancel.isCancel()) return false;
      if (f != null) checkSamePartition(f, destDir);
    }
    return true;
  }
  private void checkSamePartition(File srcDir, File destDir) throws IOException  {
    // 後始末する必要のあるファイル
    File tmp = null;
    
    try {
      // テストの移動元ファイルを作成する
      File tmpSrc;
      try {
        tmpSrc = File.createTempFile("_cbj2_move_test_", null, srcDir);
      }
      catch (IOException e) {
        throw new FileOperationException(srcDir.getAbsolutePath() + " は変更できません");
      }
      tmp = tmpSrc;
      
      // テストの移動先ファイル
      File tmpDest = new File(destDir, tmpSrc.getName());
      
      // 同じファイルかどうか
      if (tmpSrc.equals(tmpDest) || tmpSrc.getCanonicalFile().equals(tmpDest.getCanonicalFile())) {
        throw new FileOperationException("移動元と移動先が同じディレクトリです");
      }

      // 移動先が存在していたら、新しい名前にする
      int x = 2;
      while (tmpDest.exists()) {
        tmpDest = new File(destDir, tmpSrc.getName() + "_" + (x++));
      }
      
      // 移動してみる
      if (!tmpSrc.renameTo(tmpDest)) {
        throw new FileOperationException(destDir.getAbsolutePath() + " に移動できません。\n書き込み禁止になっていないか、同じドライブかどうかを確認してください。"); 
      }
      tmp = tmpDest;
    }
    finally {
      if (tmp != null) tmp.delete();
    }
  }

  /*
   * 1つ移動する
   */
  private boolean moveOneObject(File f, File destDir) throws IOException {
    // 通知
    notify.progress(f);
    
    if (cancel.isCancel()) return false;
    
    // 移動先存在チェック
    File newFile = new File(destDir, f.getName());
    File tmp = null;
    if (newFile.exists()) {
      if (f.isFile() && newFile.isFile()) {
        if (!notify.confirmOverwrite(newFile)) return false;
      }
      else {
        throw new FileOperationException("移動で上書きできるのはファイルだけです: " + f);
      }
      
      // 移動先のファイルの名前を変えておく (tmp に憶えておく)
      tmp = newFile;
      int x = 2;
      while (tmp.exists()) {
        tmp = new File(destDir, f.getName() + ".tmp_" + (x++));
      }
      if (!newFile.renameTo(tmp)) throw new FileOperationException("移動先のファイルが変更できません: " + newFile);
    }

    try {
      // 移動する (newFile は存在しないので成功するはず)
      if (!f.renameTo(newFile)) throw new FileOperationException(f.getAbsolutePath() + " を " + destDir + " に移動できません。");
    }
    finally {
      if (tmp != null) tmp.delete();
    }
    
    return true;
  }
}
