package cvj2.gui.swt.util;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.wb.swt.SWTResourceManager;

import cvj2.util.Log;

/**
 * 長時間の処理をする間、画面を無効にしてカーソルを変える
 */
public class WaitCursor {
  private WaitCursor() {}
  
  /**
   * 開始
   * @param work 処理内容 (※ 別スレッドで実行される)
   * @return 処理結果
   * @throws 処理で何か例外が起きたら、RuntimeException を投げる
   */
  static public <T> T start(Shell shell, final Work<T> work) {
    // 画面を無効にする
    Cursor oldCursor = shell.getCursor();
    shell.setCursor(SWTResourceManager.getCursor(SWT.CURSOR_WAIT));
    shell.setEnabled(false);
    
    try {
      // 処理開始
      BgThread<T> t = new BgThread<T>(work);
      t.start();
      
      // 画面が止まらないように、イベント処理をする
      while (!t.finish && !shell.isDisposed()) {
        if (!shell.getDisplay().readAndDispatch())
          shell.getDisplay().sleep();
      }
      
      if (t.exception != null) throw new RuntimeException(t.exception);
      return t.ret;
    }
    finally {
      // 画面を戻す
      shell.setCursor(oldCursor);
      shell.setEnabled(true);
    }
  }

  /**
   * 処理内容
   */
  static public interface Work <T> {
    public T run();
  }
  
  private static class BgThread<T> extends Thread {
    volatile T ret;
    volatile boolean finish = false;
    volatile Throwable exception;
    private Work<T> work;
    
    public BgThread(Work<T> work) {
      this.work = work;
    }
    
    @Override
    public void run() {
      try {
        ret = work.run();
      }
      catch (Throwable e) {
        Log.e(e);
        exception = e;
      }
      finally {
        finish = true;
      }
    }
  }
}
