package cvj2.gui.swt.widgets;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.wb.swt.SWTResourceManager;

import cvj2.gui.swt.util.GCX;
import cvj2.gui.swt.util.SWTUtil;
import cvj2.util.DelimString;
import cvj2.util.LateBindString;
import cvj2.util.Log;

/**
 * スクロールバー
 */
public class CustomScrollbar extends DoubleBufferedCanvas {
  private static final int MIN_THUMB_LEN = 12;
  private static final int BUTTON_DOWN_REPEAT_TIME = 150;

  private Color thumbColor_border = SWTResourceManager.getColor(SWT.COLOR_WIDGET_BORDER);
  private Color thumbColor_fg = SWTResourceManager.getColor(SWT.COLOR_WIDGET_HIGHLIGHT_SHADOW);
  
  // スクロールの方向
  private boolean horizontal;
  // 最小値
  private int min = 0;
  // 最大値
  private int max = 100;
  // スクロール位置
  private int pos = 0;
  // スクロールのページ移動値
  private int page = 10;
  // ボタンが押されたときのスクロール移動値
  private int increment = 1;
  
  // マウスボタンが押された座標
  private Point grabPoint;
  // マウスボタンが押されている間、繰り替えされる処理
  private MouseDownRepeat mouseDownRepeat;
  // つかむ箇所でボタンが押された時の初期スクロール位置
  private Integer thumbGrabStartPos;

  private boolean ignoreEvent = false;
  
  // リスナー
  private List<SelectionListener> listeners = new ArrayList<SelectionListener>();
  private PaintListener bgPaintListener;
  
  /**
   * コンストラクタ
   * @param parent 親コンポーネント
   * @param style SWT.VERTICAL, SWT.HORIZONTAL, SWT.BORDER
   */
  public CustomScrollbar(Composite parent, int style) {
    super(parent, style & ~(SWT.HORIZONTAL | SWT.VERTICAL));
    assert (style & SWT.HORIZONTAL) != 0 || (style & SWT.VERTICAL) != 0;
    this.horizontal = (style & SWT.HORIZONTAL) != 0;
    
    addPaintListener(new PaintHandler());
    addMouseListener(new MouseHandler());
    addMouseMoveListener(new MouseHandler());
  }
  
  /**
   * 状態設定
   */
  public void setup(int min, int max, int page, int increment) {
    Log.ut_start(min, max, page, increment);
    assert max >= min;
    this.min = min;
    this.max = max;
    this.page = Math.min(page, max - min);
    this.increment = increment;
    this.pos = min;
    updateControl(false);
  }

  /**
   * イベントを無視するか (表示のみになる)
   */
  public void setIgnoreEvent(boolean ignore) {
    this.ignoreEvent = ignore;
  }
  
  /**
   * 位置設定
   */
  public void setSelection(int pos) {
    Log.ut_start(pos);
    this.pos = pos;
    normalizePos();
    updateControl(false);
  }
  
  /**
   * 位置取得
   * @return 返される値の範囲は、最小: min から、最大: max - page まで
   */
  public int getSelection() {
    return pos;
  }
  
  /**
   * リスナー登録
   */
  public void addSelectionListener(SelectionListener listener) {
    listeners.add(listener);
  }
  
  /**
   * リスナー解除
   */
  public void removeSelectionListener(SelectionListener listener) {
    listeners.remove(listener);
  }
  
  /**
   * 背景描画リスナー登録
   * 通知されたイベントの e.data が描画する領域の Rectangle
   */
  public void setBgPaintListener(PaintListener listener) {
    bgPaintListener = listener;
  }
  
  // リスナーに通知する
  private void notifyListeners() {
    Event ev = new Event();
    ev.widget = this;
    for (SelectionListener listener : listeners) {
      listener.widgetSelected(new SelectionEvent(ev));
    }
  }
  
  // 状態が更新された
  private void updateControl(boolean notify) {
    redraw();
    if (notify) notifyListeners();
  }
  
  // スクロール位置の正規化
  private void normalizePos() {
    if (pos < min) pos = 0;
    if (pos > max - page) pos = Math.max(min, max - page);
  }
  
  // スクロールバーの幅
  private int getWidth() {
    return horizontal ? getClientArea().height : getClientArea().width;
  }
  // スクロールバーの長さ
  private int getLength() {
    return horizontal ? getClientArea().width : getClientArea().height;
  }
  // スクロールバーの長さ (ボタン領域を除いた値)
  private int getBgLength() {
    return getLength() - (isButtonVisible() ? getWidth() * 2 : 0);
  }
  // ボタンを描画するか
  private boolean isButtonVisible() {
    return getLength() > getWidth() * 2.5 + MIN_THUMB_LEN;
  }
  // スクロールバーのボタンを除いた領域
  private Rectangle getBackgroundRectangle() {
    Rectangle rect = getClientArea();
    int w = getWidth();
    if (!isButtonVisible()) return rect;
    if (horizontal) {
      rect.x += w;
      rect.width -= w * 2;
    }
    else {
      rect.y += w;
      rect.height -= w * 2;
    }
    return rect;
  }
  // 左側(上側)ボタンの座標
  private Rectangle getButton0Rectangle() {
    if (!isButtonVisible()) return null;
    return new Rectangle(0, 0, getWidth(), getWidth());
  }
  // 右側(下側)ボタンの座標
  private Rectangle getButton1Rectangle() {
    if (!isButtonVisible()) return null;
    if (horizontal) {
      return new Rectangle(getLength() - getWidth(), 0, getWidth(), getWidth());
    }
    else {
      return new Rectangle(0, getLength() - getWidth(), getWidth(), getWidth());
    }
  }
  // つかむ箇所の座標
  private Rectangle getThumbRectangle() {
    int bgLen = getBgLength();
    if (max - min <= 0) return getBackgroundRectangle();
    int thumbLen = (int)(bgLen * (double)page / (max - min) + 0.5);
    if (thumbLen < MIN_THUMB_LEN) {
      bgLen -= MIN_THUMB_LEN - thumbLen;
      thumbLen = MIN_THUMB_LEN;
    }
    int thumbPos = (int)(bgLen * (double)pos / (max - min) + 0.5);
    Rectangle bg = getBackgroundRectangle();
    if (horizontal) {
      bg.x += thumbPos;
      bg.width = thumbLen;
    }
    else {
      bg.y += thumbPos;
      bg.height = thumbLen;
    }
    return bg;
  }
  
  /**
   * データダンプ
   */
  public LateBindString getDump() {
    return new LateBindString() {
      @Override
      public String create() {
        DelimString s = new DelimString(", ", "=");
        s.append("h?", horizontal);
        s.append("min", min);
        s.append("max", max);
        s.append("pos", pos);
        s.append("page", page);
        s.append("inc", increment);
        s.append("GP", grabPoint);
        s.append("MDR", mouseDownRepeat == null ? "" : "o");
        s.append("TGSP", thumbGrabStartPos == null ? "" : thumbGrabStartPos);
        s.append("w", getWidth());
        s.append("l", getLength());
        s.append("bgL", getBgLength());
        s.append("bv", isButtonVisible());
        s.append("bgR", SWTUtil.toString(getBackgroundRectangle()));
        s.append("b0R", SWTUtil.toString(getButton0Rectangle()));
        s.append("b1R", SWTUtil.toString(getButton1Rectangle()));
        s.append("tR", SWTUtil.toString(getThumbRectangle()));
        return s.toString();
      }
    };
  }
  
  /**
   * 描画処理クラス
   */
  private class PaintHandler implements PaintListener {
    Image arrowU = SWTResourceManager.getImage(this.getClass(), "arrow-up.png");
    Image arrowD = SWTResourceManager.getImage(this.getClass(), "arrow-down.png");
    Image arrowL = SWTResourceManager.getImage(this.getClass(), "arrow-left.png");
    Image arrowR = SWTResourceManager.getImage(this.getClass(), "arrow-right.png");
    
    
    /**
     * 描画要求
     */
    @Override
    public void paintControl(PaintEvent e) {
      Log.ut(getDump());
      if (bgPaintListener != null) {
        Object old = e.data;
        e.data = getBackgroundRectangle();
        bgPaintListener.paintControl(e);
        e.data = old;
      }
      boolean grab = grabPoint != null;
      GCX gcx = new GCX(e.gc);
      if (isButtonVisible()) {
        drawButton(gcx, getButton0Rectangle(), grab && getButton0Rectangle().contains(grabPoint), 0);
        drawButton(gcx, getButton1Rectangle(), grab && getButton1Rectangle().contains(grabPoint), 1);
      }
      drawThumb(gcx, getThumbRectangle(), grab && getThumbRectangle().contains(grabPoint));
    }

    /*
     * つかむ箇所描画
     */
    private void drawThumb(GCX gcx, Rectangle rect, boolean push) {
      drawThumb(gcx, rect.x, rect.y, rect.width, rect.height, push);
    }
    private void drawThumb(GCX gcx, int x, int y, int width, int height, boolean push) {
      gcx.setForeground(thumbColor_border);
      gcx.setBackground(thumbColor_fg);
      gcx.fillRectangle(x, y, width, height);
      gcx.drawRectangle(x, y, width, height);
    }

    /*
     * ボタン描画
     */
    private void drawButton(GCX gcx, Rectangle rect, boolean push, int buttonIndex) {
      WidgetPaintUtil.drawButtonBg(gcx, rect.x, rect.y, rect.width, rect.height, push);
      WidgetPaintUtil.drawButtonFg(gcx, rect.x, rect.y, rect.width, rect.height, push);
      drawImage(gcx, rect.x, rect.y, rect.width, rect.height, push, buttonIndex);
    }
    
    /*
     * ボタン画像描画
     */
    private void drawImage(GCX gcx, int x, int y, int width, int height, boolean push, int buttonIndex) {
      Image image = horizontal ? (buttonIndex == 0 ? arrowL : arrowR) : (buttonIndex == 0 ? arrowU : arrowD);
      Rectangle r = image.getBounds();
      int margin = width / 4;
      int p = push ? 1 : 0;
      gcx.drawImage(image, 
          r.x, r.y, r.width, r.height, 
          x + margin + p, y + margin + p, width - margin * 2, height - margin * 2);
    }
  }
  
  /**
   * マウス処理クラス
   */
  private class MouseHandler extends MouseAdapter implements MouseMoveListener {
    /**
     * ボタンが押された
     */
    @Override
    public void mouseDown(MouseEvent e) {
      if (ignoreEvent) return;
      Log.ut_start(e);
      if (e.button != 1) return;

      // これ以降のマウスイベントを全て取得する
      assert grabPoint == null;
      grabPoint = new Point(e.x, e.y);
      thumbGrabStartPos = null;
      setCapture(true);
      
      // 押された位置によって処理を分岐
      Point pt = grabPoint;
      if (getButton0Rectangle().contains(pt)){
        // 減少ボタン
        buttonDown(-1);
      }
      else if (getButton1Rectangle().contains(pt)) {
        // 増加ボタン
        buttonDown(1);
      }
      else if (getThumbRectangle().contains(pt)) {
        // つかむ箇所
        thumbGrabStartPos = pos;
        /* thumb の処理は、mouseMove でする */
      }
      else if (getBackgroundRectangle().contains(pt)) {
        // ページスクロール
        Rectangle thumb = getThumbRectangle();
        int direction = 1;
        if (horizontal) {
          if (e.x <= thumb.x) direction = -1;
        }
        else {
          if (e.y <= thumb.y) direction = -1; 
        }
        pageDown(direction);
      }
    }

    /**
     * ボタンが離された
     */
    @Override
    public void mouseUp(MouseEvent e) {
      if (ignoreEvent) return;
      Log.ut_start(e);
      if (e.button != 1) return;
      assert grabPoint != null;
      grabPoint = null;
      mouseDownRepeat = null;
      setCapture(false);
      updateControl(false);
    }

    /**
     * カーソルが動いた
     */
    @Override
    public void mouseMove(MouseEvent e) {
      if (ignoreEvent) return;
      if (grabPoint == null || thumbGrabStartPos == null) return;
      Log.ut_start(e);
      
      int oldPos = pos;
      int move;
      if (horizontal) {
        move = e.x - grabPoint.x; 
      }
      else {
        move = e.y - grabPoint.y;
      }
      double rate = (double)move / getBgLength();
      pos = thumbGrabStartPos + (int)((max - min) * rate);
      normalizePos();
      updateControl(oldPos != pos);
    }
    
    // スクロールのボタンのが押された
    private void buttonDown(int direction) {
      int oldPos = pos;
      pos += direction * increment;
      normalizePos();
      updateControl(oldPos != pos);

      if (mouseDownRepeat != null) return;
      mouseDownRepeat = new MouseDownRepeat();
      mouseDownRepeat.incrementDirection = direction;
      Display.getCurrent().timerExec(BUTTON_DOWN_REPEAT_TIME, mouseDownRepeat);
    }

    // ページスクロール領域が押された
    private void pageDown(int direction) {
      int oldPos = pos;
      pos += direction * page;
      normalizePos();
      updateControl(oldPos != pos);

      if (mouseDownRepeat != null) return;
      mouseDownRepeat = new MouseDownRepeat();
      mouseDownRepeat.pageDirection = direction;
      Display.getCurrent().timerExec(BUTTON_DOWN_REPEAT_TIME, mouseDownRepeat);
    }
  }
  
  /*
   * ボタンが押されっぱなしになっている間、
   * ボタンが連続して押されたことにする。
   */
  private class MouseDownRepeat implements Runnable {
    Integer incrementDirection;
    Integer pageDirection;
    @Override
    public void run() {
      if (mouseDownRepeat != this) return;
      if (pageDirection != null) {
        new MouseHandler().pageDown(pageDirection);
      }
      else {
        new MouseHandler().buttonDown(incrementDirection);
      }
      Display.getCurrent().timerExec(BUTTON_DOWN_REPEAT_TIME, mouseDownRepeat);
    }
  }
}
