package cvj2.gui.swt.widgets.gdidpanel;

import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DragSource;
import org.eclipse.swt.dnd.DragSourceEvent;
import org.eclipse.swt.dnd.DragSourceListener;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseWheelListener;
import org.eclipse.swt.widgets.Canvas;

import cvj2.gui.swt.ui.input.KeyCombination;
import cvj2.util.Log;

/**
 * GridPanel のキーとマウスの処理
 */
public class GridInputHandler {
  private GridPanel parent;
  private GridSelection selection;
  private boolean dndEnabled;

  /**
   * コンストラクタ
   */
  public GridInputHandler(GridPanel parent, Canvas canvas, DragSource dragSource) {
    this.parent = parent;
    
    GridMouseListener ml = new GridMouseListener();
    GridKeyListener kl = new GridKeyListener();
    
    canvas.addKeyListener(kl);
    canvas.addMouseListener(ml);
    canvas.addMouseWheelListener(ml);
    
    if (dragSource != null) {
      dndEnabled = true;
      dragSource.addDragListener(ml);
    }
  }

  /**
   * データ設定
   */
  public void setup(GridSelection selection) {
    this.selection = selection;
  }
  
  // ################################################################################################################
  
  /*
   * キーイベント処理
   */
  private class GridKeyListener extends KeyAdapter {
    @Override
    public void keyPressed(KeyEvent e) {
      KeyCombination input = new KeyCombination(e);
//      Log.d(input);
      
      int mask = input.getKeyData() & SWT.MODIFIER_MASK;
      int key = input.getKeyData() & ~SWT.MODIFIER_MASK;
      boolean shift = (mask & SWT.SHIFT) != 0;
      boolean ctrl = (mask & SWT.CTRL) != 0;
      boolean singleColumn = selection.getColumn() == 1;
      
      if ((mask & ~(SWT.SHIFT | SWT.CTRL)) != 0) {
        return;
      }
      else if (key == SWT.ARROW_LEFT) {
        if (!singleColumn) selection.moveCursor(-1, 0, 0, 0, shift, ctrl);
      }
      else if (key == SWT.ARROW_RIGHT) {
        if (!singleColumn) selection.moveCursor(1, 0, 0, 0, shift, ctrl);
      }
      else if (key == SWT.ARROW_UP) {
        selection.moveCursor(0, -1, 0, 0, shift, ctrl);
      }
      else if (key == SWT.ARROW_DOWN) {
        selection.moveCursor(0, 1, 0, 0, shift, ctrl);
      }
      else if (key == SWT.PAGE_UP) {
        selection.moveCursor(0, 0, -1, 0, shift, ctrl);
      }
      else if (key == SWT.PAGE_DOWN) {
        selection.moveCursor(0, 0, 1, 0, shift, ctrl);
      }
      else if (key == SWT.HOME) {
        selection.moveCursor(0, 0, 0, -1, shift, ctrl);
      }
      else if (key == SWT.END) {
        selection.moveCursor(0, 0, 0, 1, shift, ctrl);
      }
      else if (key == ' ') {
        if (ctrl) selection.moveCursor(0, 0, 0, 0, false, ctrl);
      }
      else if (key == 'a' && ctrl) {
        selection.selAll();
      }
    }
  }
  
  /*
   * マウスイベント処理
   */
  private class GridMouseListener extends MouseAdapter implements MouseWheelListener, DragSourceListener {
    // D&Dが有効で、選択されているセルが押された瞬間に 非null になる
    private MouseEvent downEvent;
    private int downIndex;
    // ボタン押下後、ドラッグを開始したら 非null になる
    private DragSourceEvent dragStart;
    
    @Override
    public void mouseDown(MouseEvent e) {
      downEvent = null;
      dragStart = null;
      if ((e.stateMask & ~(SWT.SHIFT | SWT.CTRL)) != 0) return;
      if (e.button != 1) return;
      if (e.count != 1) return;

      Log.ut_start(e);

      boolean shift = (e.stateMask & SWT.SHIFT) != 0;
      boolean ctrl = (e.stateMask & SWT.CTRL) != 0;
      
      int pushIndex = selection.findCell(e);
      if (pushIndex < 0) return;
      int focusIndex = selection.getFocusIndex();
      
      // 選択されているセルでボタンが押されたら、ボタンが離されたときまで処理を先延ばしする
      // (ドラッグする時に、ボタン押下 -> ドラッグ開始 の順でイベントが来るため、
      //  複数選択のドラッグ時に選択が外れて困るため)
      if (dndEnabled && !shift && !ctrl && selection.getSelectedIndices().contains(pushIndex)) {
        downEvent = e;
        downIndex = pushIndex;
        return;
      }
      
      selection.moveCursor(pushIndex - focusIndex, 0, 0, 0, shift, ctrl);
      if (ctrl && !shift) {
        if (pushIndex != focusIndex) selection.moveCursor(0, 0, 0, 0, false, true);
      }
    }
    
    @Override
    public void mouseUp(MouseEvent e) {
      try {
        Log.ut_start(e);
        if (downEvent == null) return;
        
        // ドラッグが開始されていたら、ボタン押下イベントは無視する
        if (dragStart != null) return;
        
        // ドラッグが開始されていなかったら、普通のボタン押下イベントとして扱う
        selection.moveCursor(downIndex - selection.getFocusIndex(), 0, 0, 0, false, false);
      }
      finally {
        downEvent = null;
        dragStart = null;
      }
    }

    @Override
    public void mouseScrolled(MouseEvent e) {
      Log.ut_start(e);
      downEvent = null;
      dragStart = null;
      
      int scroll = selection.getScrollIndex();
      int page = selection.getVisibleRow();
      
      switch (parent.getMouseScrollType()) {
      case HALF_PAGE: 
        page = Math.max(1, page / 2);
        break;
      case LINE:
        page = 1;
        break;
      case PAGE: 
        page = Math.max(1, page - 1);
        break;
      }
      
      if (e.count > 0) selection.setScrollIndex(scroll - page);
      if (e.count < 0) selection.setScrollIndex(scroll + page);
    }

    @Override
    public void dragStart(DragSourceEvent event) {
      Log.ut_start(event);
      dragStart = event;
    }

    @Override 
    public void dragSetData(DragSourceEvent event) { 
      Log.ut_start(event);
    }
    
    @Override 
    public void dragFinished(DragSourceEvent event) { 
      Log.ut_start(event);
    }
  }
}
