package cvj2.gui.swt.widgets.gdidpanel;

import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DragSource;
import org.eclipse.swt.dnd.DropTarget;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;

import cvj2.gui.swt.widgets.CustomScrollbar;
import cvj2.gui.swt.widgets.DoubleBufferedCanvas;
import cvj2.util.Log;


/**
 * 縦横にセルを並べるパネル (縦スクロールバー付き)
 */
public class GridPanel extends Composite {
  private DoubleBufferedCanvas canvas;
  private CustomScrollbar scroll;
  private GridListHeader header;
  
  private GridSelection selection;
  private GridInputHandler inputHandler;
  private GridPaintHandler paintHandler;
  
  private GridSelectionListener selListener;
  
  private DragSource dragSource;
  private DropTarget dropTarget;

  /**
   * データ設定
   */
  public void setData(Model model, CellPainter painter, Point indent, ColumnPainter columnData) {
    Log.start();
    
    // 選択ハンドラー設定
    selection = new GridSelection(canvas, scroll, model, painter.getCellSize());
    selection.setSelectionListener(selListener);
    selection.setCanvasArea(canvas.getClientArea());
    // 入力ハンドラー設定
    inputHandler.setup(selection);
    // 描画設定
    paintHandler.setup(model, painter, selection, indent);
    // ヘッダー描画設定
    header.setColumnData(columnData);
    
    // ヘッダー高さが変わるかもしれないので、再レイアウト
    layout();
  }

  /**
   * DragSource の取得 (コンストラクタで生成を指示しておくこと)
   */
  public DragSource getDragSource() {
    return dragSource;
  }
  
  /**
   * DropTarget の取得 (コンストラクタで生成を指示しておくこと)
   */
  public DropTarget getDropTarget() {
    return dropTarget;
  }
  
  /**
   * 選択状態取得
   */
  public GridSelection getSelection() {
    return selection;
  }

  /**
   * マウスホイールによるスクロール種別
   */
  public enum MouseScrollType {
    /** 1行 */
    LINE("1行"),
    /** 半ページ */
    HALF_PAGE("1/2ページ"),
    /** 1ページ */
    PAGE("1ページ");
    
    private String name;
    private MouseScrollType(String name) {
      this.name = name;
    }
    @Override public String toString() {
      return name;
    }
  }
  private MouseScrollType mouseScrollType = MouseScrollType.HALF_PAGE;
  
  /**
   * マウスホイールによるスクロール種別設定
   */
  public void setMouseScrollType(MouseScrollType mouseScrollType) {
    this.mouseScrollType = mouseScrollType;
  }
  
  /**
   * マウスホイールによるスクロール取得
   */
  public MouseScrollType getMouseScrollType() {
    return mouseScrollType;
  }
  
  /**
   * 選択のリスナーを登録する
   */
  public void setSelectionListener(GridSelectionListener listener) {
    this.selListener = listener;
    selection.setSelectionListener(listener);
  }
  
  /**
   * 列選択リスナーの登録
   */
  public void setColumnSelectionListener(ColumnSelectionListener listener) {
    header.setColumnSelectionListener(listener);
  }
  
  /**
   * スクロールバーの背景描画リスナー登録
   */
  public void setScrollBgPaintListener(PaintListener listener) {
    scroll.setBgPaintListener(listener);
  }
  
  private void postCreateComponents(int dndStyle) {
    if (dndStyle != 0) {
      dragSource = new DragSource(getCanvas(), dndStyle);
      dropTarget = new DropTarget(getCanvas(), dndStyle);
    }
    
    inputHandler = new GridInputHandler(this, canvas, dragSource);
    paintHandler = new GridPaintHandler();
    setData(new NullModel(), new NullCellPainter(), new Point(0, 0), new NullColumnPainter());
  }
  
  /* スクロール位置更新 -> セル選択状態設定 */
  private void scrollUpdated() {
    selection.setScrollIndex(scroll.getSelection());
  }
  
  @Override
  public boolean setFocus() {
    return canvas.setFocus();
  }
  
  @Override
  public void setFont(Font font) {
    super.setFont(font);
    getCanvas().setFont(font);
    getHeader().setFont(font);
  }

  /**
   * コンストラクタ
   */
  public GridPanel(Composite parent, int style) {
    this(parent, style, 0);
  }

  /**
   * コンストラクタ
   * @param dndStyle D&Dを使う場合は、指定する
   */
  public GridPanel(Composite parent, int style, int dndStyle) {
    super(parent, style);
    GridLayout gridLayout = new GridLayout(2, false);
    gridLayout.verticalSpacing = 0;
    gridLayout.marginWidth = 0;
    gridLayout.marginHeight = 0;
    gridLayout.horizontalSpacing = 0;
    setLayout(gridLayout);
    
    header = new GridListHeader(this, SWT.NONE);
    header.setLayoutData(new GridData(SWT.FILL, SWT.TOP, true, false, 1, 1));
    
    scroll = new CustomScrollbar(this, SWT.BORDER | SWT.V_SCROLL);
    scroll.addControlListener(new ControlAdapter() {
      @Override
      public void controlResized(ControlEvent e) {
        Log.ut(scroll.getBounds());
      }
    });
    scroll.addSelectionListener(new SelectionAdapter() {
      @Override
      public void widgetSelected(SelectionEvent e) {
        scrollUpdated();
      }
    });
    GridData gd_scroll = new GridData(SWT.LEFT, SWT.FILL, false, true, 1, 2);
    gd_scroll.widthHint = 16;
    scroll.setLayoutData(gd_scroll);
    
    canvas = new DoubleBufferedCanvas(this, SWT.NONE);
    canvas.addControlListener(new ControlAdapter() {
      @Override
      public void controlResized(ControlEvent e) {
        selection.setCanvasArea(canvas.getClientArea());
      }
    });
    canvas.addPaintListener(new PaintListener() {
      public void paintControl(PaintEvent e) {
        paintHandler.redrawAllCells(e);
      }
    });
    canvas.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false, 1, 1));
    
    postCreateComponents(dndStyle);
  }

  public Canvas getCanvas() {
    return canvas;
  }
  protected CustomScrollbar getScroll() {
    return scroll;
  }
  public GridListHeader getHeader() {
    return header;
  }
  
  // ################################################################################################################

  /**
   * 選択状態の変更通知
   */
  static public interface GridSelectionListener {
    public void selectionUpdated();
  }
  
  /**
   * 列ヘッダー選択通知
   */
  static public interface ColumnSelectionListener {
    public void columnSelected(int columnIndex);
  }
  
  /**
   * 列ヘッダー描画
   */
  static public interface ColumnPainter {
    /** 列幅 */
    public int[] getColumnWidths(int headerWidth);
    /** 列高 */
    public int getColumnHeight(Control control);
    /** 列データ取得 (デフォルト描画で使用する) */
    public Object getColumnData(int column);
    /** 列描画コールバック (falseを返すと、デフォルトの描画になる) */
    public boolean paintControl(GC gc, Rectangle rect, Object data, int column, boolean isSelected);
  }
  /** 何もしないダミークラス */
  static public class NullColumnPainter implements ColumnPainter {
    @Override public int[] getColumnWidths(int headerWidth) { return new int[] { headerWidth }; }
    @Override public int getColumnHeight(Control control) { return 24; }
    @Override public Object getColumnData(int column) { return "DUMMY HEADER"; }
    @Override public boolean paintControl(GC gc, Rectangle rect, Object data, int column, boolean isSelected) { return false; }
  }
  
  /**
   * セルに対する描画要求
   */
  static public interface CellPainter {
    /** セル描画 */
    public void paintControl(GC gc, Rectangle rect, Object data, int dataIndex, boolean isSelected, boolean hasFocus);
    /** セルサイズ取得 */
    public Point getCellSize();
  }
  
  static private class NullCellPainter implements CellPainter {
    @Override public void paintControl(GC gc, Rectangle rect, Object data, int dataIndex, boolean isSelected, boolean hasFocus) { }
    @Override public Point getCellSize() { return new Point(1, 1); }
  }
  
  /**
   * セルデータのリスト
   */
  static public interface Model {
    /** リスナー登録 */
    public void addModelListener(ModelListener listener);
    /** リスナー解除 */
    public void removeModelListener(ModelListener listener);
    /** データ取得 */
    public Object get(int index);
    /** データ数取得 */
    public int size();
  }
  /** セルデータのリストの変更通知 */
  static public interface ModelListener {
    public void dataChanged(int updateIndex);
  }
  
  static private class NullModel implements Model {
    @Override public void addModelListener(ModelListener listener) { }
    @Override public void removeModelListener(ModelListener listener) { }
    @Override public Object get(int index) { return null; }
    @Override public int size() { return 0; }
  }
}
