package cvj2.gui.swt.widgets.gdidpanel;

import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Canvas;

import cvj2.gui.swt.util.SWTUtil;
import cvj2.gui.swt.widgets.CustomScrollbar;
import cvj2.gui.swt.widgets.gdidpanel.GridPanel.GridSelectionListener;
import cvj2.gui.swt.widgets.gdidpanel.GridPanel.Model;
import cvj2.gui.swt.widgets.gdidpanel.GridPanel.ModelListener;
import cvj2.util.DelimString;
import cvj2.util.LateBindString;
import cvj2.util.Log;
import cvj2.util.RangeIterator;

/**
 * GridScrollPanel でのキーボード、マウスによる選択処理
 */
public class GridSelection {
  private CustomScrollbar scroll;
  private Canvas canvas;
  private GridSelectionListener listener;
  
  // ***** 最初に設定されるデータ ***** 
  private Point orgCellSize;
  private Point cellSize;
  private int dataSize;
  private int column;
  private int totalRow;
  private int visibleRow;

  // ***** 変化するデータ *****
  private int focusIndex;
  private int selStartIndex;
  private int selEndIndex;
  private Set<Integer> selIndicesExceptedRange = new HashSet<Integer>();
  private int scrollIndex;
  
  // ***** 再描画するために前回の値を覚えておく *****
  private int old_focusIndex;
  private Set<Integer> old_selectedIndices = new HashSet<Integer>();
  private int old_scrollIndex;
  
  private int lastMoveDirection = 1;
  
  /**
   * コンストラクタ
   */
  public GridSelection(Canvas canvas, CustomScrollbar scroll, Model model, Point cellSize) {
    Log.start("...", model, cellSize);
    
    this.canvas = canvas;
    this.scroll = scroll;
    this.dataSize = model.size();
    this.orgCellSize = cellSize;
    this.cellSize = SWTUtil.clone(cellSize);
    model.addModelListener(new DataUpdated());

    init(new Rectangle(0, 0, 1, 1), true);
  }
  
  /**
   * 描画エリアのサイズが変更されたときに呼ばれる
   */
  public void setCanvasArea(Rectangle rect) {
    Log.start(rect);
    init(rect, false);
  }
  
  /**
   * セル選択のリスナー登録
   */
  void setSelectionListener(GridSelectionListener listener) {
    this.listener = listener;
  }

  /* 
   * 内部情報再設定
   */
  private void init(Rectangle canvasRect, boolean focusInitialize) {
    cellSize = SWTUtil.clone(orgCellSize);
    if (cellSize.x <= 0) cellSize.x = Math.max(1, canvasRect.width); 
    column = Math.max(1, canvasRect.width / cellSize.x);
    totalRow = (dataSize + column - 1) / column;
    visibleRow = Math.max(1, canvasRect.height / cellSize.y);
    
    if (focusInitialize) {
      focusIndex = 0;
      selStartIndex = -1;
      selEndIndex = -1;
      selIndicesExceptedRange.clear();
      if (dataSize > 0) selIndicesExceptedRange.add(0);
      scrollIndex = focusIndex / column + visibleRow + 1;
    }

    scroll.setup(0, totalRow, visibleRow, 1);
    
    normalizeData(focusInitialize);
    Log.d("DUMP: ", getDump());
    dataUpdated(true);
  }
  
  private class DataUpdated implements ModelListener {
    @Override
    public void dataChanged(int updateIndex) {
      if (canvas.isDisposed()) return;
      Log.start(updateIndex);
      redrawCell(updateIndex);
    }
  }
  
  /**
   * データダンプ
   */
  public LateBindString getDump() {
    return new LateBindString() {
      @Override
      public String create() {
        DelimString s = new DelimString(", ", " = ");
        s.append("cellSize", cellSize);
        s.append("orgCellSize", orgCellSize);
        s.append("dataSize", dataSize);
        s.append("column", column);
        s.append("totalRow", totalRow);
        s.append("visibleRow", visibleRow);
        s.append("focusIndex", focusIndex);
        s.append("selStartIndex", selStartIndex);
        s.append("selEndIndex", selEndIndex);
        s.append("selIndicesExceptedRange.size", selIndicesExceptedRange.size());
        s.append("scrollIndex", scrollIndex);
        return s.toString();
      }
    };
  }
  
  /*
   * データの正規化
   */
  private void normalizeData(boolean moveScrollToFocus) {
    // フォーカスをデータサイズ内に収める
    normalizeFocusIndex();
    normalizeScrollIndex();

    // 選択状態
    for (Integer i : selIndicesExceptedRange.toArray(new Integer[0])) {
      if (i < 0 || dataSize <= i) selIndicesExceptedRange.remove(i);
    }
    
    if (!moveScrollToFocus) return;
    // スクロール位置をフォーカスが収まるように移動する
    int focusRow = focusIndex / column;
    while (scrollIndex + visibleRow - 1 < focusRow) scrollIndex++;
    while (focusRow < scrollIndex) scrollIndex--;
    normalizeScrollIndex();
  }
  private void normalizeFocusIndex() {
    // フォーカスをデータサイズ内に収める
    if (focusIndex < 0) focusIndex = 0;
    if (focusIndex >= dataSize) focusIndex = dataSize - 1;
    if (dataSize <= 0) focusIndex = -1;
  }
  private void normalizeScrollIndex() {
    if (scrollIndex > totalRow - visibleRow) scrollIndex = totalRow - visibleRow;
    if (scrollIndex < 0) scrollIndex = 0;
  }
  
  /*
   * データが更新されたときに呼ばれる
   */
  public void dataUpdated(boolean allRedraw) {
    // スクロールコントロールを更新
    if (scroll.getSelection() != scrollIndex) {
      scroll.setSelection(scrollIndex);
    }
    
    Set<Integer> selectedIndices = getSelectedIndices();

    // スクロール位置が変わっていたら、全部再描画
    if (allRedraw || old_scrollIndex != scrollIndex) {
      Log.ut("all redraw");
      canvas.redraw();
    }
    else {
      Log.ut("partial redraw");
      int startCell = scrollIndex * column;
      int endCell = startCell + ((visibleRow + 1) * column);
      for (int i = startCell; i < endCell && i < dataSize; i++) {
        boolean redraw = false;
        if ((i == focusIndex || i == old_focusIndex) && old_focusIndex != focusIndex) {
          redraw = true;
        }
        else if (selectedIndices.contains(i) != old_selectedIndices.contains(i)) {
          redraw = true;
        }
        
        if (redraw) redrawCell(i);
      }
    }

    // 現在の値を保存する
    old_focusIndex = focusIndex;
    old_scrollIndex = scrollIndex;
    old_selectedIndices = selectedIndices;

    // リスナーに通知
    if (listener != null) listener.selectionUpdated();
  }

  /**
   * セル再描画
   */
  public void redrawCell(int index) {
    Rectangle rect = getCellRect(index);
    canvas.redraw(rect.x, rect.y, rect.width, rect.height, false);
  }
  
  /**
   * フォーカスのあるインデックス
   * @return データがないときは、-1
   */
  public int getFocusIndex() {
    return focusIndex;
  }
  
  /**
   * 最後にカーソルで移動した方向
   * @return -1:前, 0:?, 1:先
   */
  public int getLastMoveDirection() {
    return lastMoveDirection == 0 ? 0 : (lastMoveDirection < 0 ? -1 : 1);
  }
  
  /**
   * データ数
   */
  public int getDataSize() {
    return dataSize;
  }

  /**
   * 列数
   */
  public int getColumn() {
    return column;
  }

  /**
   * データを表示するのに必要な行数
   */
  public int getTotalRow() {
    return totalRow;
  }

  /**
   * 画面上で完全に見えている行数 (最低でも1)
   */
  public int getVisibleRow() {
    return visibleRow;
  }
  
  /**
   * 選択されているセルのインデックスを返す Iterator
   */
  public Set<Integer> getSelectedIndices() {
    // 範囲以外での選択状態
    TreeSet<Integer> sel = new TreeSet<Integer>(selIndicesExceptedRange);
    
    // 範囲選択を加える
    RangeIterator range = new RangeIterator(0, 0);
    if (selStartIndex < 0) {
      range = new RangeIterator(0, 0);
    }
    else if (selStartIndex > selEndIndex) {
      range = new RangeIterator(selEndIndex, selStartIndex + 1);
    }
    else {
      range = new RangeIterator(selStartIndex, selEndIndex + 1);
    }
    for (Integer i : range) sel.add(i);

    return sel;
  }

  /**
   * セルの選択状態を設定する
   */
  public void setSelectedIndices(Set<Integer> sels) {
    selIndicesExceptedRange.clear();
    selIndicesExceptedRange.addAll(sels);
    selStartIndex = -1;
    selEndIndex = -1;
    if (focusIndex >= 0) selIndicesExceptedRange.add(focusIndex);
    normalizeData(false);
    dataUpdated(false);
  }
  
  /**
   * スクロール位置を取得
   */
  public int getScrollIndex() {
    return scrollIndex;
  }
  
  /**
   * セルの描画領域取得
   */
  public Rectangle getCellRect(int index) {
    int y = (toRow(index) - scrollIndex) * cellSize.y;
    int x = toColumn(index) * cellSize.x;
    return new Rectangle(x, y, cellSize.x, cellSize.y);
  }
  
  private int toRow(int index) {
    return index / column;
  }
  private int toColumn(int index) {
    return index % column;
  }
  
  /**
   * フォーカス位置をセットする (スクロール位置も合わせて動かす)
   */
  public void setFocusIndex(int i) {
    focusIndex = i;
    selStartIndex = -1;
    selIndicesExceptedRange.clear();
    
    normalizeData(true);
    if (focusIndex >= 0) selIndicesExceptedRange.add(focusIndex);
    dataUpdated(false);
  }
  
  /**
   * スクロール位置を設定する
   */
  public void setScrollIndex(int row) {
    Log.ut_start(row);

    scrollIndex = row;
    
    normalizeData(false);
    dataUpdated(false);
  }
  
  /**
   * フォーカスが見えるようにスクロールさせる
   */
  public void scrollToFocus() {
    normalizeData(true);
    dataUpdated(false);
  }
  
  /**
   * カーソル移動
   * @param moveCol 左右
   * @param moveRow 上下
   * @param movePage ページ移動
   * @param headTailDirection 先頭: -1, 末尾: 1
   * @param shiftPress 範囲選択
   * @param ctrlPress 選択を変更しない移動
   * @return カーソルが移動したら、true
   */
  public boolean moveCursor(int moveCol, int moveRow, int movePage, int headTailDirection, boolean shiftPress, boolean ctrlPress) {
    Log.ut_start(moveCol, moveRow, movePage, headTailDirection, shiftPress, ctrlPress);
    
    int lastFocusIndex = focusIndex;
    
    // 移動方向を憶えておく
    lastMoveDirection = moveCol;
    if (lastMoveDirection == 0) lastMoveDirection = moveRow;
    if (lastMoveDirection == 0) lastMoveDirection = movePage;

    // 新しいフォーカス位置の計算
    int oldFocusIndex = focusIndex;
    focusIndex += moveCol + (moveRow * column) + (movePage * visibleRow * column);
    if (headTailDirection < 0) focusIndex = 0;
    if (headTailDirection > 0) focusIndex = dataSize - 1;
    normalizeFocusIndex();
    
    if (shiftPress) {
      if (selStartIndex < 0) {
        // 範囲選択の開始
        selStartIndex = oldFocusIndex;
        selEndIndex = focusIndex;
      }
      else {
        // 範囲選択のつづき
        selEndIndex = focusIndex;
      }
    }
    else if (!shiftPress && !ctrlPress) {
      // 普通のカーソル移動
      selStartIndex = -1;
      selIndicesExceptedRange.clear();
      if (focusIndex >= 0) selIndicesExceptedRange.add(focusIndex);
    }

    // 選択状態の切り替え
    if (oldFocusIndex == focusIndex && ctrlPress) {
      selIndicesExceptedRange = getSelectedIndices();
      selStartIndex = -1;
      
      if (selIndicesExceptedRange.contains(focusIndex)) {
        selIndicesExceptedRange.remove(focusIndex);
      }
      else {
        if (focusIndex >= 0) selIndicesExceptedRange.add(focusIndex);
      }
    }
    
    normalizeData(true);
    dataUpdated(false);
    
    return lastFocusIndex != focusIndex;
  }
  
  /**
   * 全選択
   */
  public void selAll() {
    Log.ut_start();
    
    if (getSelectedIndices().size() == dataSize) {
      selStartIndex = -1;
    }
    else {
      selStartIndex = 0;
      selEndIndex = dataSize - 1;
    }
    selIndicesExceptedRange.clear();

    normalizeData(false);
    dataUpdated(false);
  }
    
  /**
   * マウスイベントの座標に対応するセルインデックスを取得
   * @param ev canvas が受け取ったマウスイベント
   * @return 見つからなかったら、-1
   */
  public int findCell(MouseEvent ev) {
    if (ev.x < 0 || cellSize.x * column < ev.x) return -1;
    if (ev.y < 0 || cellSize.y * totalRow < ev.y) return -1;
    int index = scroll.getSelection() * column;
    index += (ev.y / cellSize.y) * column;
    index += ev.x / cellSize.x;
    if (dataSize - 1 < index) return -1;
    return index;
  }
}
