package cvj2.img;

import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.awt.geom.Point2D.Double;

import cvj2.util.Log;

/**
 * 画面の中での画像の表示位置。<br>
 * {@link ImageUtil#getImageDispLocation(Dimension, Dimension, java.awt.geom.Point2D.Double)} のコメントを参照
 */
public class ImageLocation implements Cloneable {
  /** 移動単位 (画面サイズに対する相対指定) */
  static public double moveRatio = 0.5;
  
  private Point2D.Double location = new Point2D.Double();

  /**
   * 表示位置取得 (-1.0 から 1.0 の値)
   */
  public Point2D.Double getLocation() {
    return new Point2D.Double(location.x, location.y);
  }
  
  /**
   * 位置リセット
   */
  public void reset() {
    location.x = 0.0;
    location.y = 0.0;
  }
  
  /**
   * 画面位置を移動する
   * 
   * @param directionH -1 or 1
   * @param directionV -1 or 1
   * @param imageSize 画像サイズ
   * @param displaySize 画面サイズ
   * @return 移動できたかどうか
   */
  public boolean move(int directionH, int directionV, Dimension imageSize, Dimension displaySize) {
    assert -1 <= directionH && directionH <= 1;
    assert -1 <= directionV && directionV <= 1;
    Point2D.Double old = getLocation();
    location.x = move(location.x, directionH, imageSize.width, displaySize.width);
    location.y = move(location.y, directionV, imageSize.height, displaySize.height);
    return !old.equals(location);
  }
  private double move(double orgLocation, int direction, int imageLength, int displayLength) {
    if (direction == 0) return orgLocation;
    Log.ut_start(orgLocation, direction, imageLength, displayLength);
    
    // 画像サイズが画面サイズより小さければ、位置は変わらない
    if (imageLength <= displayLength) return orgLocation;
    
    // 移動するピクセルサイズ (左に移動するときは、負の値になる)
    int moveLength = direction * (int)(displayLength * moveRatio);
    // 画面外にある画像のサイズ
    int outLength = imageLength - displayLength;
    
    // 右に移動する画面外の割合
    double moveRatio = (double)moveLength / (outLength / 2);
    
    double ret = orgLocation + moveRatio;
    ret = Math.min(1.0, ret);
    ret = Math.max(-1.0, ret);
    return ret;
  }

  @Override
  public String toString() {
    return "ImageLocation[" + location + "]";
  }
  
  @Override
  public ImageLocation clone() {
    try {
      ImageLocation other = (ImageLocation) super.clone();
      other.location = (Double) this.location.clone();
      return other;
    }
    catch (CloneNotSupportedException e) {
      throw new RuntimeException(e);
    }
  }
}
