package cvj2.img;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.stream.ImageInputStream;

import cvj2.file.util.FileUtil;
import cvj2.img.loader.ImageLoader;
import cvj2.img.loader.ImageLoaderPlugin;
import cvj2.util.Log;

/**
 * 画像のツール
 */
public class ImageUtil {
  private static Set<String> exts = new HashSet<String>();
  static {
    for (ImageLoaderPlugin p : ImageLoader.getAllPlugins().values()) {
      for (String ext : p.getSupportExt()) {
        Log.d(p.getName(), ", ", ext);
        exts.add(ext);
      }
    }
  }

  private ImageUtil() {}
  
  /**
   * 画像ファイルかどうかを調べる
   * @return 画像ファイルの拡張子を持つファイルなら、true
   */
  public static boolean isImgFileName(String name) {
    String ext = FileUtil.splitBasenameExt(name)[1];
    if (ext.length() == 0) return false;
    return exts.contains(ext.substring(1).toLowerCase());
  }

  /**
   * BufferedImage の type の名前を取得
   */
  public static  String getBufferedImageTypeString(int type) {
    switch (type) {
    case BufferedImage.TYPE_CUSTOM:         return "CUSTOM";
    case BufferedImage.TYPE_INT_RGB:        return "INT_RGB";
    case BufferedImage.TYPE_INT_ARGB:       return "INT_ARGB";
    case BufferedImage.TYPE_INT_ARGB_PRE:   return "INT_ARGB_PRE";
    case BufferedImage.TYPE_INT_BGR:        return "INT_BGR";
    case BufferedImage.TYPE_3BYTE_BGR:      return "3BYTE_BGR";
    case BufferedImage.TYPE_4BYTE_ABGR:     return "4BYTE_ABGR";
    case BufferedImage.TYPE_4BYTE_ABGR_PRE: return "4BYTE_ABGR_PRE";
    case BufferedImage.TYPE_USHORT_565_RGB: return "USHORT_565_RGB";
    case BufferedImage.TYPE_USHORT_555_RGB: return "USHORT_555_RGB";
    case BufferedImage.TYPE_BYTE_GRAY:      return "BYTE_GRAY";
    case BufferedImage.TYPE_USHORT_GRAY:    return "USHORT_GRAY";
    case BufferedImage.TYPE_BYTE_BINARY:    return "BYTE_BINARY";
    case BufferedImage.TYPE_BYTE_INDEXED:   return "BYTE_INDEXED";
    default: return "???";
    }
  }

  /**
   * ファイルから画像サイズを取得する。<br>
   * <pre>
   * ※) ConcurrentModificationException が発生したので、synchronized にする
   * 
   * java.util.ConcurrentModificationException: null
   *   at java.util.Vector$Itr.checkForComodification(Vector.java:1156)
   *   at java.util.Vector$Itr.next(Vector.java:1133)
   *   at sun.java2d.cmm.ProfileDeferralMgr.activateProfiles(ProfileDeferralMgr.java:93)
   *   at java.awt.color.ICC_Profile.getInstance(ICC_Profile.java:775)
   *   at com.sun.imageio.plugins.jpeg.JPEGImageReader.setImageData(JPEGImageReader.java:598)
   *   at com.sun.imageio.plugins.jpeg.JPEGImageReader.readImageHeader(Native Method)
   *   at com.sun.imageio.plugins.jpeg.JPEGImageReader.readNativeHeader(JPEGImageReader.java:550)
   *   at com.sun.imageio.plugins.jpeg.JPEGImageReader.checkTablesOnly(JPEGImageReader.java:295)
   *   at com.sun.imageio.plugins.jpeg.JPEGImageReader.gotoImage(JPEGImageReader.java:427)
   *   at com.sun.imageio.plugins.jpeg.JPEGImageReader.readHeader(JPEGImageReader.java:543)
   *   at com.sun.imageio.plugins.jpeg.JPEGImageReader.getWidth(JPEGImageReader.java:651)
   *   at cvj2.img.ImageUtil.readImageInfo(ImageUtil.java:83)
   *   at cvj2.app.file.imgload.BackgroundImageLoaderTask.getImageUseMemory(BackgroundImageLoaderTask.java:313)
   *   at cvj2.app.file.imgload.BackgroundImageLoaderTask.getNecessaryWorkMemory(BackgroundImageLoaderTask.java:266)
   *   at cvj2.app.file.imgload.BackgroundImageLoaderTask.load(BackgroundImageLoaderTask.java:127)
   *   at cvj2.app.file.imgload.BackgroundImageLoaderTask.run(BackgroundImageLoaderTask.java:91)
   *   at cvj2.app.file.imgload.BackgroundImageLoaderThread.run(BackgroundImageLoaderThread.java:39)
   *   </pre>
   */
  synchronized static public Dimension readImageInfo(InputStream in) {
    ImageInputStream iis = null;
    try {
      iis = ImageIO.createImageInputStream(in);
      Iterator<ImageReader> ite = ImageIO.getImageReaders(iis);
      while (ite.hasNext()) {
        ImageReader ir = ite.next();
        ir.setInput(iis);
        Dimension size = new Dimension(ir.getWidth(0), ir.getHeight(0));
        in.reset();
        return size;
      }
      return null;
    }
    catch (IOException e) {
      return null;
    }
    finally {
      try {
        if (iis != null) iis.close();
      }
      catch (IOException e) {
        Log.e(e);
      }
    }
  }
  
  /**
   * 画面の中心に画像を表示する時の場所を計算
   * @param imgSize 画像サイズ
   * @param targetRect 画面位置
   */
  public static Rectangle calcImageCenterRect(Dimension imgSize, Rectangle targetRect) {
    Point center = new Point(targetRect.x + targetRect.width / 2, targetRect.y + targetRect.height / 2);
    Rectangle rect = new Rectangle();
    rect.x = center.x - imgSize.width / 2;
    rect.width = imgSize.width;
    rect.y = center.y - imgSize.height / 2;
    rect.height = imgSize.height;
    return rect;
  }
  
  /**
   * アスペクト比を保った画像変換サイズ計算
   * @param imgSize 画像サイズ
   * @param targetRect 画面位置
   */
  public static Rectangle calcImageFitRect(Dimension imgSize, Rectangle targetRect) {
    Dimension targetSize = calcImageFitSize(imgSize, new Dimension(targetRect.width, targetRect.height));
    Rectangle ret = new Rectangle(targetRect);
    if (targetRect.width == targetSize.width) {
      ret.y += (targetRect.height - targetSize.height) / 2;
      ret.height = targetSize.height;
    }
    else if (targetRect.height == targetSize.height) {
      ret.x += (targetRect.width - targetSize.width) / 2;
      ret.width = targetSize.width;
    }
    else {
      throw new RuntimeException("imgSize = " + imgSize + ", targetRect = " + targetRect + ", targetSize = " + targetSize);
    }
    return ret;
  }
  
  /**
   * アスペクト比を保った画像変換サイズ計算
   */
  public static Dimension calcImageFitSize(int imgX, int imgY, int targetX, int targetY) {
    return calcImageFitSize(new Dimension(imgX, imgY), new Dimension(targetX, targetY));
  }
  /**
   * アスペクト比を保った画像変換サイズ計算
   * @param imgSize 画像サイズ
   * @param targetSize 画面サイズ
   */
  public static Dimension calcImageFitSize(Dimension imgSize, Dimension targetSize) {
    // 縦横比 (大きいほど縦長)
    double imgRate = ((double)imgSize.height) / imgSize.width;
    double targetRate = ((double)targetSize.height) / targetSize.width;

    // 縮小率
    double d = Math.min(
        ((double)targetSize.width) / imgSize.width,
        ((double)targetSize.height) / imgSize.height);

    if (imgRate < targetRate) {
      // 横いっぱいに表示される (画面のほうが縦長)
      return new Dimension(targetSize.width, (int)(imgSize.height * d));
    }
    else {
      // 縦いっぱいに表示される
      return new Dimension((int)(imgSize.width * d), targetSize.height);
    }
  }

  /**
   * 画像描画位置の計算
   * 
   * @param imgLocation -1.0 から 1.0 以下の値<br>
   * <pre><code>
   *   ※) 以下のような関係
   *                  |<---------------画面--------------->|
   *                              |<--(0.0)-->|                                                小さい画像の場合
   *                              |<--(1.0)-->|                                                小さい画像の場合
   *                              |<-(-1.0)-->|                                                小さい画像の場合
   *         |<---------------------- (0.0) ------------------------>|                         大きい画像の中央が表示されている場合
   *                  |<------------------------------ (1.0) ------------------------------>|  大きい画像の左端が表示されている場合
   * |<----------------------- (-1.0) -------------------->|                                   大きい画像の右端が表示されている場合
   * </code></pre>
   * @return 画像の左上の表示位置
   */
  public static Point getImageDispLocation(Dimension imgSize, Dimension displaySize, Point2D.Double imgLocation) {
    return new Point(
        getImageDispLocation(imgSize.width, displaySize.width, imgLocation.x),
        getImageDispLocation(imgSize.height, displaySize.height, imgLocation.y));
  }
  private static int getImageDispLocation(int imgLen, int displayLen, double location) {
    // ディスプレイの方が大きいときは、中央に表示する
    if (imgLen <= displayLen) {
      return (displayLen - imgLen) / 2;
    }

    // 左側にはみ出している割合
    double leftOutRatio = 1.0 - (1.0 + location) / 2.0;
    // 左側にはみ出しているピクセル数
    return -1 * (int)((imgLen - displayLen) * leftOutRatio);
  }
}
