package cvj2.img;

import java.awt.image.BufferedImage;
import java.awt.image.ComponentColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.IndexColorModel;
import java.awt.image.WritableRaster;

import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.graphics.RGB;

/**
 * AWTのBufferedImageを、SWTのImgeDataに変換する。<br>
 * ※) グレースケールの画像は扱えないので、あらかじめカラーに変換しておくこと
 */
public class SWTImageConverter {
  static public ImageData cnv(BufferedImage awtImage) {
    if (awtImage.getColorModel() instanceof DirectColorModel) {
      return cnvDirect(awtImage);
    }
    else if (awtImage.getColorModel() instanceof IndexColorModel) {
      return cnvIndex(awtImage);
    }
    else if (awtImage.getColorModel() instanceof ComponentColorModel) {
      return cnvComponent(awtImage);
    }
    throw new RuntimeException("no support ColorModel: " + awtImage.getColorModel().getClass());
  }

  private static ImageData cnvDirect(BufferedImage awtImage) {
    DirectColorModel awtColorModel = (DirectColorModel) awtImage.getColorModel();
    PaletteData swtPalette = new PaletteData(
        0xFF0000,
        0x00FF00, 
        0x0000FF); 
    ImageData swtImg = new ImageData(
        awtImage.getWidth(), 
        awtImage.getHeight(), 
        awtColorModel.getPixelSize(), 
        swtPalette);
    WritableRaster awtRaster = awtImage.getRaster();
    int bands = awtRaster.getSampleModel().getNumBands();
    int[] awtPixelRGBs = new int[bands * awtRaster.getWidth()];
    int[] swtPixels = new int[awtRaster.getWidth()];
    byte[] swtAlphas = new byte[awtRaster.getWidth()];
    for (int y = 0; y < swtImg.height; y++) {
      awtRaster.getPixels(0, y, awtRaster.getWidth(), 1, awtPixelRGBs);
      for (int x = 0; x < swtImg.width; x++) {
        swtPixels[x] = (awtPixelRGBs[x * bands + 0] & 0xFF) << 16 | (awtPixelRGBs[x * bands + 1] & 0xFF) << 8 | (awtPixelRGBs[x * bands + 2] & 0xFF);
      }
      swtImg.setPixels(0, y, swtImg.width, swtPixels, 0);
      if (bands == 4) {
        for (int x = 0; x < swtImg.width; x++) {
          swtAlphas[x] = (byte)(awtPixelRGBs[x * bands + 3] & 0xFF);
        }
        swtImg.setAlphas(0, y, swtImg.width, swtAlphas, 0);
      }
    }
    return swtImg;
  }

  private static ImageData cnvIndex(BufferedImage awtImage) {
    IndexColorModel awtColorModel = (IndexColorModel) awtImage.getColorModel();
    int paletteSize = awtColorModel.getMapSize();
    byte[] paletteR = new byte[paletteSize];
    byte[] paletteG = new byte[paletteSize];
    byte[] paletteB = new byte[paletteSize];
    awtColorModel.getReds(paletteR);
    awtColorModel.getGreens(paletteG);
    awtColorModel.getBlues(paletteB);
    RGB[] paletteRGBs = new RGB[paletteSize];
    for (int i = 0; i < paletteRGBs.length; i++) {
      paletteRGBs[i] = new RGB(paletteR[i] & 0xFF, paletteG[i] & 0xFF, paletteB[i] & 0xFF);
    }
    PaletteData swtPalette = new PaletteData(paletteRGBs);
    ImageData swtImage = new ImageData(
        awtImage.getWidth(), 
        awtImage.getHeight(), 
        awtColorModel.getPixelSize(),
        swtPalette);
    swtImage.transparentPixel = awtColorModel.getTransparentPixel();
    WritableRaster awtRaster = awtImage.getRaster();
    int[] pixelIndexs = new int[awtRaster.getWidth()];
    for (int y = 0; y < swtImage.height; y++) {
      awtRaster.getPixels(0, y, awtRaster.getWidth(), 1, pixelIndexs);
      swtImage.setPixels(0, y, awtRaster.getWidth(), pixelIndexs, 0);
    }
    return swtImage;
  }

  private static ImageData cnvComponent(BufferedImage awtImage) {
    ComponentColorModel awtColorModel = (ComponentColorModel)awtImage.getColorModel();
    WritableRaster awtRaster = awtImage.getRaster();
    int bands = awtRaster.getSampleModel().getNumBands();
    PaletteData swtPalette;
    if (bands >= 3) {
      swtPalette = new PaletteData(0xFF0000, 0x00FF00, 0x0000FF);
    }
    else  {
      swtPalette = new PaletteData(0xFF, 0xFF, 0xFF);
    }
    ImageData swtImg = new ImageData(
        awtImage.getWidth(), 
        awtImage.getHeight(), 
        awtColorModel.getPixelSize(), 
        swtPalette);

    int[] awtPixelRGBs = new int[bands * awtRaster.getWidth()];
    int[] swtPixels = new int[awtRaster.getWidth()];
    byte[] swtAlphas = new byte[awtRaster.getWidth()];
    for (int y = 0; y < swtImg.height; y++) {
      awtRaster.getPixels(0, y, awtRaster.getWidth(), 1, awtPixelRGBs);
      for (int x = 0; x < swtImg.width; x++) {
        swtPixels[x] = (awtPixelRGBs[x * bands + 0] & 0xFF) << 16 | (awtPixelRGBs[x * bands + 1] & 0xFF) << 8 | (awtPixelRGBs[x * bands + 2] & 0xFF);
      }
      swtImg.setPixels(0, y, swtImg.width, swtPixels, 0);
      if (bands == 4) {
        for (int x = 0; x < swtImg.width; x++) {
          swtAlphas[x] = (byte)(awtPixelRGBs[x * bands + 3] & 0xFF);
        }
        swtImg.setAlphas(0, y, swtImg.width, swtAlphas, 0);
      }
    }
    return swtImg;
  }
}
