package cvj2.img.loader;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.commons.io.IOUtils;

import cvj2.file.fs.FileInfo;
import cvj2.file.util.FileUtil;
import cvj2.util.Log;

/**
 * 画像読み込み
 */
public class ImageLoader {
  // システム中の全てのプラグイン
  private static Map<String, ImageLoaderPlugin> allPlugins = new TreeMap<String, ImageLoaderPlugin>();

  // 他のプラグインで読めなかったときに、拡張子に関係なく試す最後のプラグイン
  public static String defaultPlugin = new ImageLoaderPluginImageIO().getName();
  
  // 拡張子(小文字)とプラグイン名の対応
  private static Map<String, List<String>> extPluginNameMap = new HashMap<String, List<String>>();

  static {
    // プラグイン登録
    addPlugin(new ImageLoaderPluginAWTToolkit());
    addPlugin(new ImageLoaderPluginImageIO());
    // 拡張子の対応登録
    addExtPluginName(new ImageLoaderPluginAWTToolkit());
    addExtPluginName(new ImageLoaderPluginImageIO());
  }
  static private void addPlugin(ImageLoaderPlugin p) {
    assert !allPlugins.containsKey(p.getName());
    allPlugins.put(p.getName(), p);
  }
  static private void addExtPluginName(ImageLoaderPlugin p) {
    for (String ext : p.getSupportExt()) {
      List<String> list = extPluginNameMap.get(ext);
      if (list == null) {
        list = new ArrayList<String>();
        extPluginNameMap.put(ext, list);
      }
      list.add(p.getName());
    }
  }

  /**
   * システム中の全てのプラグインを取得
   */
  public static Map<String, ImageLoaderPlugin> getAllPlugins() {
    return Collections.unmodifiableMap(allPlugins);
  }
  
  /**
   * 拡張子に対応するプラグインを設定する
   * @param pluginMap { { "ext", "pluginName" }, ... }
   */
  public static void setExtPluginNameMap(String[][] pluginMap) {
    HashMap<String, List<String>> tmp = new HashMap<String, List<String>>();
    for (int i = 0; i < pluginMap.length; i++) {
      String[] elm = pluginMap[i];
      assert elm.length == 2;
      // 拡張子
      String ext = elm[0];
      // プラグイン名
      String pluginName = elm[1];
      
      // 存在確認
      ImageLoaderPlugin plugin = allPlugins.get(pluginName);
      assert plugin != null;
      assert plugin.getSupportExt().contains(ext);
      
      // 登録
      List<String> list = tmp.get(ext);
      if (list == null) {
        list = new ArrayList<String>();
        tmp.put(ext, list);
      }
      assert !list.contains(pluginName);
      list.add(pluginName);
    }
    ImageLoader.extPluginNameMap = tmp;
  }
  
  /**
   * 画像をロードする
   * PDF ファイルを読む場合があるので、dpi か targetSize のどちらかを指定しなければならない
   * 
   * @param dpi null でないときは この解像度で画像を生成する
   * @param targetSize null でないときは このサイズで画像を生成する
   */
  public BufferedImage load(FileInfo file, Integer dpi, Dimension targetSize) throws IOException {
    Log.start(file);
    
    // FileInfo から直接取得可能なら、それを返す
    BufferedImage img = null;
    if (dpi != null) img = file.getImage(dpi);
    else             img = file.getImage(targetSize);
    if (img != null) return img;
   
    // 読み込むプラグインを探す
    List<String> pluginNames = findPlugin(file);
    if (pluginNames.size() == 0) throw new IOException("このファイルは現在の設定では読み込めません: " + file);

    // ファイルから読み込む
    IOException ex = null;
    for (String p : pluginNames) {
      InputStream in = file.openStream();
      try {
        img = allPlugins.get(p).load(in);
        if (img.getWidth() <= 0 || img.getHeight() <= 0) {
          throw new IOException("invalid image: width = " + img.getWidth() + " height = " + img.getHeight());
        }
        return img;
      }
      catch (IOException e) {
        if (ex == null) ex = e;
      }
      finally {
        IOUtils.closeQuietly(in);
      }
    }
    throw ex;
  }
  
  private List<String> findPlugin(FileInfo file) {
    // 拡張子からさがす
    String ext = FileUtil.splitBasenameExt(file.getName())[1];
    List<String> plugins = new ArrayList<String>();
    if (ext.length() > 0) {
      List<String> find = extPluginNameMap.get(ext.substring(1).toLowerCase());
      if (find != null) plugins.addAll(find);
    }
    
    // デフォルトを最後に加える
    if (!plugins.contains(defaultPlugin)) {
      plugins.add(defaultPlugin);
    }
    
    return plugins;
  }
}
