package cvj2.migemo;

import java.util.List;

import org.apache.commons.lang3.ObjectUtils;

import cvj2.util.Log;

/**
 * リストをMigemoでサーチする
 */
public class ListSearch<E> {
  private Migemo migemo;
  private Sen sen;
  private StringConverter<E> cnv;
  
  private boolean isEnableLastSearchCache = false;
  private LastSearch lastSearch;
  
  private SenBackgroundAnalyzer senBackgroundAnalyzer = new SenBackgroundAnalyzer();

  /**
   * リスト要素を文字列に変換する
   */
  public interface StringConverter<E> {
    String cnv(E element);
  }
  
  /**
   * コンストラクタ
   * @param cnv 文字列変換クラス
   */
  public ListSearch(Migemo migemo, Sen sen, StringConverter<E> cnv) {
    this.migemo = migemo;
    this.sen = sen;
    this.cnv = cnv;
  }

  /**
   * 前回の結果をキャッシュするかどうか。<br>
   * ※) ただし、CnvRomaji2Kata.setIgnoreTrailingAlphabet() を true にセットしておかなければならない<br>
   * なぜかは、LastSearch.willGetSameResult() の中のコメントを見ること。
   */
  public void setEnableLastSearchCache(boolean isEnableLastSearchCache) {
    this.isEnableLastSearchCache = isEnableLastSearchCache;
  }
  
  /**
   * バックグラウンドで文字列の解析をする
   */
  public void analyzeBackground(List<E> list) {
    senBackgroundAnalyzer.analyze(list, cnv);
  }
  
  /**
   * リスト中の文字列の検索
   * @param inputRomaji キーボードから入力されたローマ字文字列
   * @param list 検索対象リスト
   * @param startIndex 検索開始位置 
   * @param skipFirst 最初の要素はスキップして次の要素から検索を開始する
   * @param direction 検索方向 (1: 前, -1:後ろ)
   * @return { 見つかったリストのインデックス, 見つかった文字列の開始位置, 見つかった文字列の終了位置 + 1 }
   *            見つからなかったら、null
   */
  public int[] find(String inputRomaji, List<E> list, int startIndex, boolean skipFirst, int direction) {
    Log.ut_start(inputRomaji, "...", startIndex, skipFirst, direction);
    assert Sen.analyzeCache.getLimit() > list.size();    
    
    if (migemo != null) {
      migemo.setInputRomaji(inputRomaji);
    }
    if (sen != null) {
      sen.setInputRomaji(inputRomaji);
    }
    
    LastSearch currentSearch = new LastSearch(inputRomaji, list, startIndex, skipFirst, direction);
    if (isEnableLastSearchCache) {
      if (currentSearch.willGetSameResult(lastSearch)) {
//        Log.ut("cache hit: result=", Arrays.toString(lastSearch.result), ", old=", lastSearch.inputRomaji, ", new=", currentSearch.inputRomaji);
        return lastSearch.result;
      }
    }
    
    ListIterator ite = new ListIterator(list.size(), startIndex, skipFirst, direction);
    int[] find = null;
    while (ite.hasNext()) {
      int i = ite.next();
      find = foundIn(cnv.cnv(list.get(i)));
      if (find == null) continue;
      find = new int[] { i, find[0], find[1] };
      break;
    }
    lastSearch = currentSearch;
    lastSearch.result = find;
    return find;
  }

  /**
   * 1項目の検索
   */
  private int[] foundIn(String fileText) {
    int[] find;
    if (migemo != null) {
      find = migemo.foundIn(fileText);
      if (find != null) return find;
    }
    if (sen != null) {
      find = sen.foundIn(fileText);
      if (find != null) return find;
    }
    return null;
  }

  // ==================================================================================================
  
  /**
   * 前回の検索結果
   */
  static private class LastSearch {
    String inputRomaji;
    String listObjectId;
    int listSize;
    int startIndex;
    boolean skipFirst;
    int direction;
    int[] result;
    
    public LastSearch(String inputRomaji, List<?> list, int startIndex, boolean skipFirst, int direction) 
    {
      this.inputRomaji = inputRomaji;
      this.listObjectId = ObjectUtils.identityToString(list);
      this.listSize = list.size();
      this.startIndex = startIndex;
      this.skipFirst = skipFirst;
      this.direction = direction;
    }

    /**
     * 以前と同じ検索結果になるか？
     */
    public boolean willGetSameResult(LastSearch old) {
      if (old == null) return false;
      if (direction != old.direction) return false;
      if (!listObjectId.equals(old.listObjectId)) return false;
      if (listSize != old.listSize) return false;
      if (skipFirst != old.skipFirst) return false;
      if (startIndex != old.startIndex) return false;
      
      // 完全に同一の検索条件なら、結果も同じになる
      if (inputRomaji.equals(old.inputRomaji)) {
        return true;
      }
      
      // 新しい検索文字が古い検索文字を先頭に含む
      // 例) 
      //  old: "kensak" で見つからない new: "kensaku" => 前回ダメなら、今回も見つからない
      //  old: "kensak" で見つかった   new: "kensaku" => 今回はどうなるか分からない
      // ※) ただし、CnvRomaji2Kata.setIgnoreTrailingAlphabet() を true にセットしておかなければならない
      if (inputRomaji.startsWith(old.inputRomaji)) {
        // ただし、2文字以上の場合のみ
        // ※) 1文字だと、先頭一致の '/' をキャッシュしてしまうため
        if (old.inputRomaji.length() > 1) {
          // 前回見つかっていなければ、今回も同様に見つからない
          // ※) 前回見つかっていても今回見つかるかわからない
          if (old.result == null) return true;
        }
      }
      
      return false;
    }
  }
  
  /**
   * リストのループインデックス
   */
  private static class ListIterator {
    private int size;
    private int startIndex;
    private int direction;
    
    private int i;
    private int loopCount = 0;

    // コンストラクタ
    public ListIterator(int size, int startIndex, boolean skipFirst, int direction) {
      assert 0 <= size;
      assert direction == -1 || direction == 1;

      startIndex = Math.max(0, startIndex);
      startIndex = Math.min(size, startIndex);
      if (skipFirst) startIndex += direction;
      startIndex = correctListIndex(size, startIndex);

      this.size = size;
      this.startIndex = startIndex;
      this.direction = direction;
      
      i = startIndex;
      loopCount = 0;
    }

    // まだあるか
    public boolean hasNext() {
      if (size == 0) return false;
      return loopCount == 0 || i != startIndex;
    }
    
    // 次のインデックス
    public int next() {
      if (!hasNext()) throw new IllegalStateException("not hasNext");
      int ret = i;
      loopCount++;
      i += direction;
      i = correctListIndex(size, i);
      return ret;
    }

    /* リストから外れたインデックスを修正する */
    private int correctListIndex(int listSize, int i) {
      if (i >= listSize) return 0;
      if (i < 0) return listSize - 1;
      return i;
    }
  }
}
