package cvj2.migemo;

import java.io.File;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import cvj2.util.CnvHira2Kata;


/**
 * Migemo による検索クラス
 */
public class Migemo implements Searcher {
  // 初期化フラグ
  private static boolean initialized = false; 

  private static String prefixMatchString = "/";
  
  // ひらがなカタカナを区別しない
  private boolean ignoreHiraKata;
  // 前方一致モードを有功にするか
  private boolean enablePrefixMatch;

  private String inputRomaji;
  private Pattern regex = Pattern.compile("^$");
  
  /**
   * 初期化
   */
  static public void initialize() throws IOException {
    if (initialized) return;
    loadDic(new File(System.getProperty("user.dir"), "lib/migemo-dict"), null);
    loadDic(new File(System.getProperty("user.dir"), "lib/migemo-dict-add"), "MS932");
    initialized = true;
  }
  
  static private void loadDic(File file, String encode) throws IOException {
    if (encode != null) {
      if (!org.monazilla.migemo.Migemo.loadDictionary(file, encode)) throw new IOException("Migemo dictionary load error: " + file);
    }
    else {
      if (!org.monazilla.migemo.Migemo.loadDictionary(file)) throw new IOException("Migemo dictionary load error: " + file);
    }
  }

  /**
   * コンストラクタ
   * @param ignoreHiraKata Migemoで検索する時に、ひらがなとカタカナを区別しない
   * @param enablePrefixMatch 前方一致モードを有功にするか 
   *         (検索文字を "/" で開始すると前方一致検索になる)
   */
  public Migemo(boolean ignoreHiraKata, boolean enablePrefixMatch) {
    assert initialized; 
    this.ignoreHiraKata = ignoreHiraKata;
    this.enablePrefixMatch = enablePrefixMatch;
  }

  /**
   * キーボードから入力されたローマ字文字列の設定
   */
  @Override
  public void setInputRomaji(String inputRomaji) {
    this.inputRomaji = inputRomaji;
    
    // 正規表現コンパイル
    regex = Pattern.compile(
        getMigemoStr(),
        Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE);
  }
  
  /**
   * Migemoコンパイルされた正規表現を取得
   */
  public String getMigemoStr() {
    if (inputRomaji.isEmpty()) return "^$";
    if (enablePrefixMatch && inputRomaji.equals(prefixMatchString)) return "^$";
    
    // 前方一致の判定
    if (enablePrefixMatch && inputRomaji.startsWith(prefixMatchString)) {
      return cnvSerchStr("^(" + org.monazilla.migemo.Migemo.lookup(inputRomaji.substring(prefixMatchString.length())) + ")");
    }
    
    return cnvSerchStr(org.monazilla.migemo.Migemo.lookup(inputRomaji));
  }
  
  private String cnvSerchStr(String str) {
    if (ignoreHiraKata) str = CnvHira2Kata.cnv(str);
    return str;
  }
  
  /**
   * 検索対象文字列から一致する箇所を探す
   * あらかじめ、setInputStr() を呼んでおく
   * 
   * @param fileText 検索対象文字列 (grep における、ファイルの1行の役目)
   * @return 一致したら、{ 一致した箇所の開始位置、終了位置+1 }
   *          一致しなければ、null。
   */
  @Override
  public int[] foundIn(String fileText) {
    // 解析するまでもないとき
    if (fileText.length() == 0) return null;
    if (enablePrefixMatch && inputRomaji.equals(prefixMatchString))return null;
    
    Matcher m = regex.matcher(cnvSerchStr(fileText));
    if (!m.find()) return null;
    return new int[] { m.start(), m.end() };
  }
}
