package cvj2.migemo;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.java.sen.SenFactory;
import net.java.sen.StringTagger;
import net.java.sen.dictionary.Token;
import cvj2.util.CacheMap;
import cvj2.util.CnvRomaji2Kata;
import cvj2.util.CnvZenHan;
import cvj2.util.Log;
import cvj2.util.Util;

/**
 * lucense-gosen を使用して形態素解析をする
 */
public class Sen implements Searcher {
  /**
   * 解析結果キャッシュ
   * アプリケーション内でグローバルに使用される
   * スレッドセーフ
   */
  static CacheMap<String, SenString> analyzeCache = new CacheMap<String, SenString>(1000);
  
  private StringTagger tagger;
  private String inputKatakana;

  /** Migemoの辞書使用種別 */
  static public MigemoDictUseType useMigemoDictType = MigemoDictUseType.ALL;
  /** Migemoの辞書使用種別 */
  public enum MigemoDictUseType { 
    NONE("使用しない"), 
    ONLY_FAIL("解析できなかった漢字のみ"),
    ALL("一致する漢字全て"); 
    
    private String name;
    private MigemoDictUseType(String name) {
      this.name = name;
    }
    @Override public String toString() {
      return name;
    }
  };
  
  private static String prefixMatchString = "/";

  // 前方一致モードを有功にするか
  private boolean enablePrefixMatch;

  /**
   * コンストラクタ
   * @param enablePrefixMatch 前方一致モードを有功にするか 
   *         (検索文字を "/" で開始すると前方一致検索になる)
   */
  public Sen(boolean enablePrefixMatch) {
    this.enablePrefixMatch = enablePrefixMatch;
    this.tagger = SenFactory.getStringTagger(null);
  }
  
  /**
   * キーボードから入力されたローマ字文字列の設定
   */
  @Override
  public void setInputRomaji(String inputRomaji) {
    this.inputKatakana = CnvRomaji2Kata.cnv(inputRomaji);
  }
  
  /**
   * 解析結果のキャッシュ個数を指定する
   */
  public static void updateAnalyzeCacheSize(int size) {
    size *= 1.5;
    if (analyzeCache.getLimit() >= size) return;
    analyzeCache.setLimit(size);
  }
  
  /**
   * 解析結果のキャッシュをクリア
   */
  public static void clearAnalyzeCache() {
    analyzeCache.clear();
  }
  
  /**
   * キャッシュの使用しているメモリサイズを取得
   */
  public static long getCacheObjectsBytes() {
    ArrayList<String> keys = new ArrayList<String>();
    ArrayList<SenString> vals = new ArrayList<SenString>();
    analyzeCache.getAll(keys, vals);
    return Util.getObjectBytesStringCollection(keys) + Util.getObjectBytes(vals);
  }
  
  /**
   * 形態素解析をする
   */
  public SenString analyze(String string) {
    SenString ss = analyzeCache.get(string);
    if (ss != null) return ss;
    
    List<Token> list;
    try {
      list = tagger.analyze(CnvZenHan.cnvZen2HanAlnum(string), new ArrayList<Token>());
    } 
    catch (IOException e) {
      Log.e(e);
      throw new RuntimeException(e.getMessage());
    }
    
    ss = new SenString(string);
    for (Token token : list) {
      if (token.getMorpheme().getReadings().size() > 0) {
        ss.append(token.getStart(), token.end(), tryAnotherYomi(token.getSurface(), token.getMorpheme().getReadings()));
      }
      else {
        ss.append(token.getStart(), token.end(), cnvByMigemo(token.getSurface()));
      }
    }
    analyzeCache.put(string, ss);
    return ss;
  }
  
  /*
   * Migemo辞書で漢字を読みに変換する
   */
  private List<String> cnvByMigemo(String kanji) {
    ArrayList<String> ret = new ArrayList<String>();
    if (useMigemoDictType == MigemoDictUseType.NONE) {
      ret.add(kanji);
      return ret;
    }
    
    // Migemoの逆辞書から探す
    Set<String> yomi = MigemoReverseDict.instance().find(kanji);
    if (yomi == null) {
      ret.add(kanji);
      return ret;
    }
    ret.addAll(yomi);
    return ret;
  }
  
  /*
   * Migemo辞書で他の読み方も試す
   */
  private List<String> tryAnotherYomi(String kanji, List<String> kaiseki) {
    if (useMigemoDictType != MigemoDictUseType.ALL) return kaiseki;
    
    // Migemoの逆辞書から探す
    Set<String> yomi = MigemoReverseDict.instance().find(kanji);
    if (yomi == null) return kaiseki;
    
    // 全て追加
    HashSet<String> ret = new HashSet<String>();
    ret.addAll(kaiseki);
    ret.addAll(yomi);
    return new ArrayList<String>(ret);
  }

  /**
   * 検索対象文字列から一致する箇所を探す
   * あらかじめ、setInputStr() を呼んでおく
   * 
   * @param fileText 検索対象文字列 (grep における、ファイルの1行の役目)
   * @return 一致したら、{ 一致した箇所の開始位置、終了位置+1 }
   *          一致しなければ、null。
   */
  @Override
  public int[] foundIn(String fileText) {
    // 解析するまでもないとき
    if (fileText.length() == 0) return null;
    if (enablePrefixMatch && inputKatakana.equals(prefixMatchString))return null;
    
    SenString ss = analyze(fileText);
    if (enablePrefixMatch && inputKatakana.startsWith(prefixMatchString)) {
      return ss.startsWith(inputKatakana.substring(prefixMatchString.length()));
    }
    else {
      return ss.find(inputKatakana);
    }
  }
}
