package cvj2.migemo;

import java.util.ArrayList;
import java.util.List;

import cvj2.util.CnvHira2Kata;
import cvj2.util.DelimString;
import cvj2.util.Log;
import cvj2.util.SizeCountable;
import cvj2.util.Util;

/**
 * Sen で解析された文字列。<br>
 * <pre><code>
 *
 * オリジナル文字列 = 会社更生手続き中の半導体大手
 *                    0 1 2 3 4 5 6 7 8 9 10111213
 *
 * 解析された文字列 = [カイシャ, ガイシャ][コウセイ][テツヅキ][チュウ, ジュウ, ナカ][ノ][ハンドウタイ][オオテ, オオデ]
 * 
 * 内部データ =       [0,2]カイシャ [2,4]コウセイ [4,7]テツヅキ [7,8]チュウ [8,9]ノ [9,12]ハンドウタイ [12,14]オオテ
 *                        Lガイシャ                                 Lジュウ                                  Lオオデ
 *                                                                  Lナカ
 * これを検索すると、
 * find tetudukityuu(テツヅキチュウ) -> int[4,8]
 * find syako(シャコ) -> int[0,4]
 * find nakano(ナカノ) -> int[7,9]
 * starts kaisyakousei(カイシャコウ) -> int[0,4]
 * 
 * 単語の途中は認識できないので、検索の戻り値は単語単位になる
 * </code></pre>
 */
public class SenString implements SizeCountable {
  private final boolean DEBUG = false;
  private String orgString;
  private List<Element> list = new ArrayList<Element>();
  
  private class Element implements SizeCountable {
    public int start, end;
    public List<String> strs;

    @Override
    public String toString() {
      return new StringBuilder().append(start).append(',').append(end).append(strs).toString();
    }
    public String toSimpleString() {
      if (strs.size() == 1) return strs.get(0);
      return "" + strs;
    }
    @Override public long getObjectBytes() {
      return 4 + 4 + Util.getObjectBytesStringCollection(strs); // start, end, strs
    }
  }
  
  /** 使用メモリ */
  @Override public long getObjectBytes() {
    return Util.getObjectBytes(orgString) + Util.getObjectBytes(list); // orgString, list,
  }
  
  /**
   * コンストラクタ
   */
  public SenString(String orgString) {
    this.orgString = orgString;
  }
  
  /**
   * 解析前の文字列を取得する
   */
  public String getOrgString() {
    return orgString;
  }
  
  /**
   * 解析された単語の数を取得
   */
  public int getElementSize() {
    return list.size();
  }
  
  /**
   * 文字列を追加する
   * @param start オリジナル文中での開始インデックス
   * @param end オリジナル文中での終了インデックス + 1
   * @param strs Senで変換された文字列
   */
  public void append(int start, int end, List<String> strs) {
    Element elm = new Element();
    elm.start = start;
    elm.end = end;
    ArrayList<String> strs2 = new ArrayList<String>();
    for (String s : strs) {
      strs2.add(CnvHira2Kata.cnv(s));
    }
    elm.strs = strs2;
    list.add(elm);
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    for (Element elm : list) sb.append(elm);
    return sb.toString();
  }

  /**
   * 単語を指定の区切りで結合して文字列にする
   */
  public String toSimpleString(String delim) {
    DelimString str = new DelimString(delim);
    for (Element elm : list) str.append(elm.toSimpleString());
    return str.toString();
  }
  
  /**
   * 指定された単語から検索開始
   */
  public int[] find(String katakana, int startElm) {
    if (DEBUG) Log.ut_start(katakana);
    if (katakana.isEmpty()) return null;

    // リストの先頭から順番に検索していく
    for (int i = startElm; i < list.size(); i++) {
      int[] find = findElm(katakana, i, list.get(i));
      if (find != null) return find;
    }
    return null;
  }
  
  /**
   * 文字列内を検索する
   * @param katakana 全角カタカナ
   * @return 一致したら、{ 一致した箇所の開始位置、終了位置+1 }
   *          一致しなければ、null。
   */
  public int[] find(String katakana) {
    if (DEBUG) Log.ut_start(katakana);
    if (katakana.isEmpty()) return null;

    // リストの先頭から順番に検索していく
    for (int i = 0; i < list.size(); i++) {
      int[] find = findElm(katakana, i, list.get(i));
      if (find != null) return find;
    }
    return null;
  }
  
  /**
   * 指定された文字列で始まるかどうか
   * @param katakana 全角カタカナ
   * @return 一致したら、{ 一致した箇所の開始位置、終了位置+1 }
   *          一致しなければ、null。
   */
  public int[] startsWith(String katakana) {
    if (DEBUG) Log.ut_start(katakana);
    if (katakana.isEmpty()) return null;

    // リストの先頭のみ検索
    if (list.size() <= 0) return null;
    return startsElm(katakana, 0, list.get(0));
  }

  /*
   *                                  ┌─── (再帰) ─────┐
   *                                  ↓                        │
   * find → findElm → FindElmStr → startsElm → startsElmStr ┘
   *                        └→ 発見                  └→ 発見
   * 
   *                            ┌─ (再帰)─┐
   *                            ↓           │
   * startsWith → startsElm → startsElmStr ┘
   *                                └→ 発見
   */
  
  // このElementの中から一致文字列が開始しているか調べる
  private int[] findElm(String katakana, int searchElm, Element matchStartElm) {
    if (DEBUG) Log.ut_start(katakana, searchElm, matchStartElm);
    assert katakana.length() > 0;
    Element elm = list.get(searchElm);

    // Element 中の文字列リストを順番に検索する
    for (int i = 0; i < elm.strs.size(); i++) {
      int[] find = findElmStr(katakana, elm.strs.get(i), searchElm, matchStartElm);
      if (find != null) return find;
    }
    return null;
  }

  // このElementの先頭から一致文字列が開始しているか調べる
  private int[] startsElm(String katakana, int searchElm, Element matchStartElm) {
    if (DEBUG) Log.ut_start(katakana, searchElm, matchStartElm);
    assert katakana.length() > 0;
    Element elm = list.get(searchElm);

    // Element 中の文字列リストを順番に検索する
    for (int i = 0; i < elm.strs.size(); i++) {
      int[] find = startsElmStr(katakana, elm.strs.get(i), searchElm, matchStartElm);
      if (find != null) return find;
    }
    return null;
  }

  // このElementのこの文字列の中から一致文字列が開始しているか調べる
  private int[] findElmStr(String katakana, String elmStr, int searchElm, Element matchStartElm) {
    if (DEBUG) Log.ut_start(katakana, elmStr, searchElm, matchStartElm);
    assert katakana.length() > 0;
    Element elm = list.get(searchElm);

    // このElementのどこかで見つかるか調べる (例: katakana="アイウ" elmStr="---アイウ---")
    if (elmStr.contains(katakana)) {
      return new int[] { matchStartElm.start, elm.end };
    }

    // 検索文字の最初の1文字が見つかるか調べる
    if (elmStr.indexOf(katakana.charAt(0)) < 0) return null;
    
    // このElementの末尾で一部が見つかるか調べる (例: katakana="アイウ" elmStr="---アイ")
    // 検索文字列をすこしずつ短くしていく
    if (searchElm >= list.size() - 1) return null;
    for (int i = 0; i < katakana.length(); i++) {
      String searchStr = katakana.substring(0, katakana.length() - i);
      assert searchStr.length() > 0;
      if (elmStr.endsWith(searchStr)) {
        // 次のElementを検索する
        int[] find = startsElm(katakana.substring(searchStr.length()), searchElm + 1, matchStartElm);
        if (find != null) return find;
      }
    }
    return null;
  }
  
  // このElementの先頭から一致文字列が開始しているか調べる
  private int[] startsElmStr(String katakana, String elmStr, int searchElm, Element matchStartElm) {
    if (DEBUG) Log.ut_start(katakana, elmStr, searchElm, matchStartElm);
    assert katakana.length() > 0;
    Element elm = list.get(searchElm);

    // このElementの先頭で見つかるか調べる (例: katakana="アイウ" elmStr="アイウ---")
    if (elmStr.startsWith(katakana)) {
      return new int[] { matchStartElm.start, elm.end };
    }

    // このElementで一部が見つかるか調べる (例: katakana="アイウ", elmStr="アイ")
    if (katakana.startsWith(elmStr)) {
      // 次のElementを検索する
      if (searchElm >= list.size() - 1) return null;
      int[] find = startsElm(katakana.substring(elmStr.length()), searchElm + 1, matchStartElm);
      if (find != null) return find;
    }
    return null;
  }
}
