package cvj2.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.TreeMap;

/**
 * キーに対応する値をキャッシュするMap。
 * 容量の制限値があり、古いデータは自動的に消す。
 * スレッドセーフに使用できる。
 */
public class CacheMap<K, V> {
  private int limit;
  private RemoveCallback<K, V> callback;
  private boolean callbackOnOverwrite;
  
  // IDの大小でキャッシュの新旧を判別する (get()される度にIDが更新される)
  private long curId = 0;
  
  // 値を格納するためのコンテナ
  private class Container {
    K key;
    V val;
    Container(K k, V v) { this.key = k; this.val = v; }
  }
  
  // キー => ID のマップ
  private HashMap<K, Long> keyIdMap = new HashMap<K, Long>();
  // ID => 値 のマップ
  private TreeMap<Long, Container> idValMap = new TreeMap<Long, Container>();
  
  /**
   * 削除された時のコールバック
   */
  public interface RemoveCallback<K, V> {
    public void removed(K key, V val);
  }
  
  /**
   * コンストラクタ
   */
  public CacheMap(int limit, RemoveCallback<K, V> callback, boolean callbackOnOverwrite) {
    assert limit >= 1;
    this.limit = limit;
    this.callback = callback;
    this.callbackOnOverwrite = callbackOnOverwrite;
    assert (callback == null && !callbackOnOverwrite) || (callback != null);
  }
  public CacheMap(int limit) {
    this(limit, null, false);
  }

  /**
   * 全て取得 (先頭に古いデータが来る)
   */
  synchronized public void getAll(ArrayList<K> keys, ArrayList<V> vals) {
    for (Long id : idValMap.keySet()) {
      Container c = idValMap.get(id);
      if (vals != null) vals.add(c.val);
      if (keys != null) keys.add(c.key);
    }
  }
  
  /**
   * デバッグダンプ
   */
  synchronized public void dump() {
    Log.start();
    Log.d("keyIdMap.size = ", keyIdMap.size());
    for (K key : keyIdMap.keySet()) {
      Log.d("key = ", key, ", id = ", keyIdMap.get(key));
    }
    
    Log.d("idValMap.size = ", idValMap.size());
    for (Long id : idValMap.keySet()) {
      Container c = idValMap.get(id);
      Log.d("id = ", id, ", key = ", c.key, ", val = ", c.val);
    }
  }
  
  /**
   * 存在チェック (データの新旧を更新しない)
   */
  synchronized public boolean exists(K key) {
    assert key != null;
    return keyIdMap.containsKey(key);
  }
  
  /**
   * 削除
   * @param key 削除対象のキー
   * @return 削除された値
   */
  synchronized public V remove(K key) {
    if (!exists(key)) return null;
    Long id = keyIdMap.remove(key);
    Container c = idValMap.remove(id);
    assert keyIdMap.size() == idValMap.size();
    assert c.key.equals(key);
    if (callback != null) callback.removed(key, c.val);
    return c.val;
  }
  
  
  /**
   * データを取得する
   * @return 存在しないなら、null
   */
  synchronized public V get(K key) {
    assert key != null;
    Container c = findContainer(key);
    if (c != null) return c.val;
    return null;
  }
  
  /**
   * データをキャッシュにセットする
   * 
   * @param key データのキー
   * @param val データ
   */
  synchronized public void put(K key, V val) {
    assert key != null;
    assert val != null;

    // 既にある
    Container c = findContainer(key);
    if (c != null) {
      if (callbackOnOverwrite && callback != null) callback.removed(key, c.val);
      c.val = val;
      return;
    }
    
    // 追加
    Long id = newId();
    keyIdMap.put(key, id);
    idValMap.put(id, new Container(key, val));
    
    checkLimit();
    return;
  }

  /* コンテナーを探す */
  private Container findContainer(K key) {
    assert keyIdMap.size() == idValMap.size();
    Long id = keyIdMap.get(key);
    if (id != null) {
      if (!idValMap.lastKey().equals(id)) id = idUpdate(key, id);
      assert idValMap.containsKey(id);
      return idValMap.get(id);
    }
    return null;
  }
  
  /* 新しいID */
  private Long newId() {
    return curId++;
  }

  /* ID の更新 */
  private Long idUpdate(K key, Long id) {
    Container cont = idValMap.remove(id);
    assert cont.key.equals(key);
    Long id2 = keyIdMap.remove(key);
    assert id.equals(id2);
    id = newId();
    keyIdMap.put(key, id);
    idValMap.put(id, cont);
    return id;
  }
  
  /* 容量制限の確認 */
  private void checkLimit() {
    while (keyIdMap.size() > limit) {
      Long id = idValMap.firstKey();
      Container cont = idValMap.remove(id);
      Long id2 = keyIdMap.remove(cont.key);
      assert id.equals(id2);
      assert keyIdMap.size() == idValMap.size();
      if (callback != null) callback.removed(cont.key, cont.val);
    }
  }

  /**
   * 容量制限の設定
   */
  synchronized public void setLimit(int limit) {
    this.limit = limit;
    checkLimit();
    assert (limit > 0) || (keyIdMap.size() == 0 && idValMap.size() == 0);
  }
  
  /**
   * 容量制限の取得
   */
  synchronized public int getLimit() {
    return limit;
  }
  
  /**
   * キャッシュクリア
   */
  synchronized public void clear() {
    int orgLimit = limit;
    limit = 0;
    checkLimit();
    limit = orgLimit;
  }
}
