package cvj2.util;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;

import org.apache.commons.lang3.ArrayUtils;

import cvj2.app.gui.util.MsgBox;

/**
 * ログ管理クラス
 */
public class Log {
  private static ThreadLocal<SimpleDateFormat> dateFormat = new ThreadLocal<SimpleDateFormat>() {
    @Override
    protected SimpleDateFormat initialValue() {
      return new SimpleDateFormat("yyyy/MM/dd HH:mm:ss.SSS");
    }
  };
  
  public static Level printLevel = Level.DEBUG;
  public static boolean printClassFullname = true;
  public static boolean stdoutput = true;
  public static boolean stdoutputSync = false;
  public static boolean errWarnToStderr = false;
  
  private static PrintStream output;

  /** 出力レベル */
  public enum Level {
    ERROR   ("ERR "),
    WARNING ("WARN"),
    DEBUG   ("DBG "),
    UT      ("UT  ");
    String s;
    Level(String s) { this.s = s; }
    @Override public String toString() { return s; }
  };

  public synchronized static void fileOpen(File f) throws IOException {
    assert output == null; 
    FileOutputStream fos = new FileOutputStream(f);
    if (null == fos.getChannel().tryLock()) {
      MsgBox.info("ログファイルに書き込めません", "他のアプリケーションがログファイルを開いているので、書き込みを行いません\n" + f);
      fos.close();
    }
    output = new PrintStream(new BufferedOutputStream(fos));
  }

  synchronized private static void p(int depth, Level level, String str) {
    StackTraceElement st = Thread.currentThread().getStackTrace()[depth];
    String outstr = dateFormat.get().format(new Date()) + " " + level + " " +
        Thread.currentThread().getName() + " ■ " +
        getClassName(st) + "." + st.getMethodName() +
        str;
    
    if (stdoutput) {
      if (level.ordinal() <= Level.WARNING.ordinal() && errWarnToStderr) {
        System.err.println(outstr);
        if (stdoutputSync) System.err.flush();
      }
      else {
        System.out.println(outstr);
        if (stdoutputSync) System.out.flush();
      }
    }
    
    if (output != null) output.println(outstr);
  }
  private static String getClassName(StackTraceElement st) {
    String c = st.getClassName();
    if (printClassFullname) return c;
    int i = c.lastIndexOf('.');
    if (i < 0) return c;
    return c.substring(i + 1);
  }

  public static void start() {
    if (printLevel.ordinal() < Level.DEBUG.ordinal()) return;
    p(3, Level.DEBUG, "(): start");
  }

  public static void start(Object ... param) {
    if (printLevel.ordinal() < Level.DEBUG.ordinal()) return;
    p(3, Level.DEBUG, "(" + getObjString(param, ", ") + "): start");
  }

  public static void ut_start(Object ... param) {
    if (printLevel.ordinal() < Level.UT.ordinal()) return;
    p(3, Level.UT, "(" + getObjString(param, ", ") + "): start");
  }

  public static void ut(Object ... obj) {
    if (printLevel.ordinal() < Level.UT.ordinal()) return;
    p(3, Level.UT, "(): " + getObjString(obj, ""));
  }

  public static void d(Object ... obj) {
    if (printLevel.ordinal() < Level.DEBUG.ordinal()) return;
    p(3, Level.DEBUG, "(): " + getObjString(obj, ""));
  }

  public static void df(String format, Object ... obj) {
    if (printLevel.ordinal() < Level.DEBUG.ordinal()) return;
    p(3, Level.DEBUG, "(): " + String.format(format, obj));
  }
  
  public static void d(Throwable e) {
    if (printLevel.ordinal() < Level.DEBUG.ordinal()) return;
    p(3, Level.DEBUG, "() " + getErrString(e));
  }

  public static void w(Object ... obj) {
    if (printLevel.ordinal() < Level.WARNING.ordinal()) return;
    p(3, Level.WARNING, "(): " + getObjString(obj, ""));
  }

  public static void wf(String format, Object ... obj) {
    if (printLevel.ordinal() < Level.WARNING.ordinal()) return;
    p(3, Level.WARNING, "(): " + String.format(format, obj));
  }
  public static void w(Throwable e) {
    if (printLevel.ordinal() < Level.WARNING.ordinal()) return;
    p(3, Level.WARNING, "() " + getErrString(e));
  }

  public static void e(Object ... obj) {
    if (printLevel.ordinal() < Level.ERROR.ordinal()) return;
    p(3, Level.ERROR, "(): " + getObjString(obj, ""));
  }

  public static void ef(String format, Object ... obj) {
    if (printLevel.ordinal() < Level.ERROR.ordinal()) return;
    p(3, Level.ERROR, "(): " + String.format(format, obj));
  }

  public static void e(Throwable e) {
    if (printLevel.ordinal() < Level.ERROR.ordinal()) return;
    p(3, Level.ERROR, "() " + getErrString(e));
  }

  public static String getErrString(Throwable e) {
    StringBuilder s = new StringBuilder();
    while (e != null) {
      if (s.length() > 0) {
        s.append("\ncause .....\n");
      }
      s.append(e.getClass().getName() + ": " + e.getMessage() + "\n  at " + getObjString(e.getStackTrace(), "\n  at "));
      e = e.getCause();
    }
    return s.toString();
  }
  
  private static String getObjString(Object[] obj, String delim) {
    StringBuilder s = new StringBuilder();
    for (int i = 0; i < obj.length; i++) {
      if (i != 0) s.append(delim);
      s.append(toString(obj[i]));
    }
    return s.toString();
  }

  public static String toString(Object obj) {
    if (obj == null) return "" + obj;

    if (obj.getClass().isArray()) {
      return ArrayUtils.toString(obj);
    }
    return obj.toString();
  }

  public static String toString(Object[] array) {
    return Arrays.toString(array);
  }

  public static synchronized void close() {
    Log.start();
    System.err.flush();
    System.out.flush();
    if (output != null) {
      output.close();
      output = null;
    }
  }
}
