package cvj2.util;

import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ConcurrentSkipListMap;


/**
 * 同時使用できるメモリー量を制限するロッククラス
 */
public class MemorySizeMutex {
  private final Queue<Thread> waiters = new ConcurrentLinkedQueue<Thread>();
  private final Map<Long, Long> getterIds = new ConcurrentSkipListMap<Long, Long>();
  
  private Object LOCK_OBJ = new String("LOCK_OBJ");
  volatile private long limit;
  volatile private long remainMemory;
  
  /**
   * コンストラクタ
   * @param limit 制限値
   */
  public MemorySizeMutex(long limit) {
    assert limit >= 0;
    this.limit = limit;
    this.remainMemory = limit;
  }
  
  /**
   * リミット値の再設定
   */
  public void setLimit(long newLimit) {
    assert newLimit >= 0;
    synchronized (LOCK_OBJ) {
      // メモリー増加量
      long plus = newLimit - this.limit;
      // 残りメモリーを増やす
      remainMemory += plus;
      // 制限値更新
      this.limit = newLimit;
    }
  }
  
  /**
   * 制限値
   */
  public long getLimit() {
    return limit;
  }
  
  /**
   * 使用中のメモリー量
   */
  public long getUsedMemory() {
    synchronized (LOCK_OBJ) {
      return limit - remainMemory;
    }
  }
  
  /**
   * メモリー使用許可を得る
   * @param size メモリーサイズ (limitをオーバーしていても、自分が待ち行列の先頭ならOK)
   * @throws InterruptedException 割り込みされたとき
   */
  public void get(long size) throws InterruptedException {
    assert size >= 0;
    synchronized (LOCK_OBJ) {
      assert !getterIds.containsKey(Thread.currentThread().getId());
      assert !waiters.contains(Thread.currentThread());
      
      // 待ち行列の末尾に並ぶ
      waiters.add(Thread.currentThread());
      
      boolean waited = false;
      while (true) {
        boolean iamFirst = waiters.peek() == Thread.currentThread();
        boolean enoughMemory = remainMemory - size >= 0;
        boolean nobodyGot = getterIds.size() == 0;
  
        // 自分が行列の先頭であること
        if (iamFirst) {
          // 誰もメモリーを使用していないなら、容量制限に関係なく使用できる
          if (nobodyGot) break;
          // 誰かがメモリーを使用しているなら、空きがなければいけない
          if (enoughMemory) break;
        }
        
        // メモリー不足なので待つ
        if (!waited) {
          Log.d("wait for memory: ", size, ", limit = ", limit, ", ramainMemory = ", remainMemory);
          waited = true;
        }
        LOCK_OBJ.wait();
      }
      if (waited) Log.d("got memory");
      
      // 待ち行列から離れる
      Thread me = waiters.poll();
      assert me == Thread.currentThread();
  
      // メモリーを取得して去る
      assert !getterIds.containsKey(Thread.currentThread().getId());
      getterIds.put(Thread.currentThread().getId(), size);
      remainMemory -= size;
    }
  }

  /**
   * メモリーをかえす
   */
  public void release() {
    synchronized (LOCK_OBJ) {
      if (waiters.size() > 0) Log.d("release memory");
      
      // メモリーをかえす
      Long size = getterIds.remove(Thread.currentThread().getId());
      assert size != null;
      remainMemory += size;
      
      // 待っている人たちにメモリーが開いたことを知らせる
      LOCK_OBJ.notifyAll();
    }
  }
}
