package cvj2.util;

import java.util.HashMap;
import java.util.TreeMap;


/**
 * キーに対応する値をキャッシュするMap。
 * <pre>
 * オブジェクトのバイトサイズによる容量制限があり、古いデータは自動的に消す。
 * スレッドセーフに使用できる。
 * 
 * ※) サイズの取得は synchronized の内部で行ってはいけない。
 * サイズ取得メソッドの中で何らかのロックを取得していると、
 * 別のスレッドが同じロックを取得してから、このクラスのメソッドを呼ぶとデッドロックする。
 * (よって、サイズ取得は synchronized でない put() メソッドで行ってから synchronized の put() を呼んでいる。)
 * 
 * ※2) SizeCountable を使用せずに、put() の引数に size を入れておけば問題なかった。
 * とりあず、教訓として実装はこのままにしておく。
 * </pre>
 */
public class SizeCacheMap<K, V extends SizeCountable> {
  private long limit;
  private RemoveCallback<K, V> callback;
  private boolean callbackOnOverwrite;
  private long totalBytes;
  
  // IDの大小でキャッシュの新旧を判別する (get()される度にIDが更新される)
  private long curId = 0;
  
  // 値を格納するためのコンテナ
  private class Container {
    K key;
    V val;
    long size;
    Container(K k, V v, long size) { this.key = k; this.val = v; this.size = size; }
  }
  
  // キー => ID のマップ
  private HashMap<K, Long> keyIdMap = new HashMap<K, Long>();
  // ID => 値 のマップ
  private TreeMap<Long, Container> idValMap = new TreeMap<Long, Container>();
  
  /**
   * 削除された時のコールバック
   */
  public interface RemoveCallback<K, V extends SizeCountable> {
    public void removed(K key, V val, long size);
  }
  
  /**
   * コンストラクタ
   */
  public SizeCacheMap(long limit, RemoveCallback<K, V> callback, boolean callbackOnOverwrite) {
    assert limit >= 0;
    this.limit = limit;
    this.callback = callback;
    this.callbackOnOverwrite = callbackOnOverwrite;
    assert (callback == null && !callbackOnOverwrite) || (callback != null);
  }
  public SizeCacheMap(long limit) {
    this(limit, null, false);
  }

  /**
   * 存在チェック (データの新旧を更新しない)
   */
  synchronized public boolean exists(K key) {
    assert key != null;
    return keyIdMap.containsKey(key);
  }
  
  /**
   * データを取得する
   * @return 存在しないなら、null
   */
  synchronized public V get(K key) {
    assert key != null;
    Container c = findContainer(key);
    if (c != null) return c.val;
    return null;
  }
  
  /**
   * データをキャッシュにセットする
   * 
   * @param key データのキー
   * @param val データ
   */
  public void put(K key, V val) {
    assert key != null;
    assert val != null;
    
    // サイズは最低でも1にしないと setLimit(0) でクリアされない
    long size = Math.max(1, val.getObjectBytes());
    put_sync(key, val, size);
  }
  /*
   * データをキャッシュにセットする
   */
  synchronized public void put_sync(K key, V val, long valSize) {
    // 既にある
    Container c = findContainer(key);
    if (c != null) {
      totalBytes -= c.size;
      if (callbackOnOverwrite && callback != null) {
        callback.removed(key, c.val, c.size);
      }
      totalBytes += valSize;
      c.size = valSize;
      c.val = val;
      checkLimit();
      return;
    }
    
    // 追加
    Long id = newId();
    keyIdMap.put(key, id);
    idValMap.put(id, new Container(key, val, valSize));
    totalBytes += valSize;
    
    checkLimit();
    return;
  }

  /* コンテナーを探す */
  private Container findContainer(K key) {
    assert keyIdMap.size() == idValMap.size();
    Long id = keyIdMap.get(key);
    if (id != null) {
      if (!idValMap.lastKey().equals(id)) id = idUpdate(key, id);
      assert idValMap.containsKey(id);
      return idValMap.get(id);
    }
    return null;
  }
  
  /* 新しいID */
  private Long newId() {
    return curId++;
  }

  /* ID の更新 */
  private Long idUpdate(K key, Long id) {
    Container cont = idValMap.remove(id);
    assert cont.key.equals(key);
    Long id2 = keyIdMap.remove(key);
    assert id.equals(id2);
    id = newId();
    keyIdMap.put(key, id);
    idValMap.put(id, cont);
    return id;
  }
  
  /* 容量制限の確認 */
  private void checkLimit() {
    while (totalBytes > limit) {
      assert idValMap.size() > 0;
      Long id = idValMap.firstKey();
      Container cont = idValMap.remove(id);
      Long id2 = keyIdMap.remove(cont.key);
      assert id.equals(id2);
      assert keyIdMap.size() == idValMap.size();
      totalBytes -= cont.size;
      if (callback != null) callback.removed(cont.key, cont.val, cont.size);
    }
  }

  /**
   * 容量制限の設定
   */
  synchronized public void setLimit(long limit) {
    assert limit >= 0;
    this.limit = limit;
    checkLimit();
    assert (limit > 0) || (keyIdMap.size() == 0 && idValMap.size() == 0);
  }
  
  /**
   * 容量制限の取得
   */
  synchronized public long getLimit() {
    return limit;
  }
  
  /**
   * 現在容量の取得
   */
  synchronized public long getTotalBytes() {
    return totalBytes;
  }
  
  /**
   * オブジェクト数の取得
   */
  synchronized public int size() {
    return keyIdMap.size();
  }
  
  /**
   * キャッシュクリア
   */
  synchronized public void clear() {
    long orgLimit = limit;
    limit = 0;
    checkLimit();
    limit = orgLimit;
  }
}
