package cvj2.util;

import java.lang.Character.UnicodeBlock;
import java.nio.charset.Charset;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.lang3.CharUtils;
import org.apache.commons.lang3.StringUtils;

/**
 * 文字列比較クラス。<br>
 * 解析済みのデータは内部にキャッシュされる。<br>
 * キャッシュがあふれたら、ランダムに削除される
 */
public class StringComparator implements Comparator<String>, SizeCountable {
  private volatile ConcurrentHashMap<String, CacheElm> g_cache = new ConcurrentHashMap<String, CacheElm>();
  private final Object CACHE_LOCK = new String("_CACHE_LOCK_");
  
  static private class CacheElm implements SizeCountable {
    byte[] buff;
    @Override public long getObjectBytes() {
      return 4 + 4 + buff.length; // buff, buff.len, buff[]
    }
  }
  static private Charset sjis = Charset.forName("Windows-31J");
  
  private volatile boolean g_ignoreCase = true;
  private volatile boolean g_ignoreHiraKata = true;
  private volatile boolean g_ignoreZenHan = true;
  private volatile boolean g_symbolFirst = true;
  private volatile char[] g_ignoreChars = {};
  private volatile boolean g_cnvSjis = true;
  private volatile int g_cacheLimit = 1000;
  
  /**
   * 比較条件初期化
   */
  public void init(boolean ignoreCase, boolean ignoreHiraKata, boolean ignoreZenHan, boolean symbolFirst, char[] ignoreChars, boolean cnvSjis) {
    g_ignoreCase = ignoreCase;
    g_ignoreHiraKata = ignoreHiraKata;
    g_ignoreChars = ignoreChars == null ? null : ignoreChars.clone();
    g_ignoreZenHan = ignoreZenHan;
    g_symbolFirst = symbolFirst;
    g_cnvSjis = cnvSjis;
    g_cache = new ConcurrentHashMap<String, CacheElm>();
  }
  
  @Override
  public long getObjectBytes() {
    return Util.getObjectBytesStringCollection(g_cache.keySet()) + Util.getObjectBytes(g_cache.values());
  }
  
  /**
   * 最大キャッシュ数を取得
   */
  public int getCacheLimit() {
    return g_cacheLimit;
  }

  /**
   * 最大キャッシュ数を設定する
   */
  public void setCacheLimit(int limit) {
    synchronized (CACHE_LOCK) {
      g_cacheLimit = limit;
      checkLimit();
    }
  }
  
  /**
   * キャッシュクリア
   */
  public void clearCache() {
    synchronized (CACHE_LOCK) {
      int org = g_cacheLimit;
      g_cacheLimit = 0;
      checkLimit();
      g_cacheLimit = org;
    }
  }
  
  

  /*
   * キャッシュ数制限チェック
   */
  private void checkLimit() {
    while (g_cache.size() > g_cacheLimit) {
      synchronized (CACHE_LOCK) {
        int maxRemove = Math.max(100, g_cacheLimit / 10);
        int removeCount = 0;
        Enumeration<String> e = g_cache.keys();
        for (/* */; e.hasMoreElements() && removeCount < maxRemove; removeCount++) {
          String key = e.nextElement();
          g_cache.remove(key);
        }
      }
    }
  }

  /**
   * 文字列比較
   */
  @Override
  public int compare(String s1, String s2) {
    if (s1 == s2) return 0;
    checkLimit();
    ConcurrentHashMap<String, CacheElm> cache = g_cache;
    boolean ignoreCase = g_ignoreCase;
    boolean ignoreHiraKata = g_ignoreHiraKata;
    boolean ignoreZenHan = g_ignoreZenHan;
    boolean symbolFirst = g_symbolFirst;
    char[] ignoreChars = g_ignoreChars;
    boolean cnvSjis = g_cnvSjis;
    CacheElm elm1 = getCacheElm(s1, cache, ignoreCase, ignoreHiraKata, ignoreZenHan, symbolFirst, ignoreChars, cnvSjis);
    CacheElm elm2 = getCacheElm(s2, cache, ignoreCase, ignoreHiraKata, ignoreZenHan, symbolFirst, ignoreChars, cnvSjis);
    return compare(elm1, elm2, s1, s2); 
  }

  /*
   * キャッシュから探す
   */
  private CacheElm getCacheElm(String s, ConcurrentHashMap<String, CacheElm> cache, 
      boolean ignoreCase, boolean ignoreHiraKata, boolean ignoreZenHan, boolean symbolFirst, char[] ignoreChars, boolean cnvSjis) 
  {
    // 既にキャッシュにあればそれを返す
    String sOrg = s;
    CacheElm elm = cache.get(s);
    if (elm != null) return elm;
    
    // 変換
    if (ignoreChars != null && ignoreChars.length != 0) {
      s = removeChars(s, ignoreChars);
    }
    if (ignoreZenHan) {
      s = CnvZenHan.cnvZen2HanAlnum(s);
    }
    if (ignoreHiraKata) {
      s = CnvHira2Kata.cnv(s);
    }
    if (ignoreCase) {
      s = s.toLowerCase();
    }
    if (symbolFirst) {
      s = cnvSymbol(s);
    }
    byte[] buff;
    if (cnvSjis) {
      buff = s.getBytes(sjis);
    }
    else {
      buff = s.getBytes();
    }

    // キャッシュに追加
    elm = new CacheElm();
    elm.buff = buff;
    cache.put(sOrg, elm);
    return elm;
  }

  /*
   * 記号を置き換え
   */
  private String cnvSymbol(String s) {
    StringBuilder sb = new StringBuilder(s.length());
    for (int i = 0; i < s.length(); i++) {
      char c = s.charAt(i);
      if (i == 0) {
        if (CharUtils.isAscii(c)) {
          // ASCII
          if (!CharUtils.isAsciiAlphanumeric(c)) sb.append(' ');
        }
        else {
          // 非ASCII
          UnicodeBlock b = UnicodeBlock.of(c);
          if (b != UnicodeBlock.HIRAGANA && b != UnicodeBlock.KATAKANA && b != UnicodeBlock.CJK_UNIFIED_IDEOGRAPHS) {
            sb.append(' ');
          }
        }
      }
      sb.append(c);
    }
    return sb.toString();
  }

  /*
   * 不要な文字を削除
   */
  private String removeChars(String s, char[] rmChars) {
    if (!StringUtils.containsAny(s, rmChars)) return s;
    StringBuilder sb = new StringBuilder(s.length());
    for (int i = 0; i < s.length(); i++) {
      char c = s.charAt(i);
      for (int j = 0; j < rmChars.length; j++) {
        if (c != rmChars[j]) continue;
        c = '\0';
        break;
      }
      if (c != '\0') sb.append(c);
    }
    return sb.toString();
  }

  /*
   * 比較
   */
  private int compare(CacheElm elm1, CacheElm elm2, String s1, String s2) {
    int x = compare(elm1.buff, elm2.buff);
    if (x != 0) return x;
    return s1.compareTo(s2);
  }
  
  /*
   * 比較 int
   */
  private int compare(int i1, int i2) {
    if (i1 < i2) return -1;
    if (i1 > i2) return 1;
    return 0;
  }
  
  /*
   * 比較 byte[]
   */
  private int compare(byte[] buff1, byte[] buff2) {
    int len = Math.min(buff1.length, buff2.length);
    for (int i = 0; i < len; i++) {
      int b1 = buff1[i] & 0xFF;
      int b2 = buff2[i] & 0xFF;
      int x = compare(b1, b2);
      if (x != 0) return x;
    }
    if (buff1.length == buff2.length) return 0;
    if (buff1.length < buff2.length) return -1;
    if (buff1.length > buff2.length) return 1;
    throw new RuntimeException();
  }
}
